;
; This file is part of MIA.
;
; MIA is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 2 of the License, or
; (at your option) any later version.
;
; MIA is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with MIA; if not, write to the Free Software
; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
;
; $Id: midivisi.pro,v 1.13 2003/11/08 06:48:56 koehler Exp $
;
; midivisi.pro
; Created:     Thu May  8 15:05:02 2003 by Koehler@sun47
; Last change: Sun Oct 31 19:29:14 2004
;
; PURPOSE:
;	save the world by computing a simple estimate for the
;	visibility of a MIDI observation.
;
; CALLING SEQUENCE:
;	midivisi, fringefile, photfile, [skyfile]
;
; INPUT:
;	fringefile	MIDI file with fringe data
;	photfileA	MIDI file with photometric data (chop or nod)
;	photfileB	same with beam B open in case photfileA was
;			beam A open
;
; optional:
;	MASKFILE=	mask to use, if different from mask created
;			from photfileA
;	SKY=		MIDI file with sky for nodding data, used for
;			photfileA and -B
;	ASKY=		sky used with photfileA
;	BSKY=		sky used with photfileB
;	THRESHOLD=	threshold to distinguish good and bad fringes
;	/MOVIE		play fringe movie
;	/NOFOU		use maximum method (?)
;	/NOPOSCHECK	don't ask if position of mask is ok
;
; OUTPUT:
;	OPD=opd		2D-array with opd
;	FRINGE=fringe	2D-array with all amplitudes
;	FOUFRI=foufri	2D-array with fourier-trafos of amplitudes
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; create (if necessary) and read photometry-textfile
; note that fluxA, -B is more like flux1, -2 with A/BOPEN and beamsplitter
;
; IN:	photfile - MIDI data file
;	sky	 - MIDI file with sky during nodding
; OUT:	PhFilt	 - Filter used for photfile
;	fluxA	 - Flux in window 2
;	fluxB	 - Flux in window 1
;	(win2 is beam A and win1 beam B unless the beamspliter is in)
;
pro read_photometry, photfile,PhFilt,fluxA,fluxB,specA,specB,SKY=sky

    first_char    = STRPOS(photfile,'/MIDI.20')+6
    last_char     = STRPOS(photfile,'.fits')
    short_photfile= STRMID(photfile,first_char,last_char-first_char)

    phtxtfile= 'PHOTOMETRY/Phot_'+short_photfile+'.txt'
    print,"Photfile is ",photfile
    print,"Textfile is ",phtxtfile

    im = obj_new('fitsfile',(oirFileList(photfile))[0])
    he = im->prihead()
    ;OBJECT = STRCOMPRESS(he->getpar('OBJECT'),/remove_all)
    ;DIT    = STRCOMPRESS(he->getpar('DET DIT'),/remove_all)
    PhFilt = STRCOMPRESS(he->getpar('INS FILT NAME'),/remove_all)
    Grism  = STRCOMPRESS(he->getpar('INS GRIS NAME'),/remove_all)
    obj_destroy,im

    if not file_test(phtxtfile) then begin
        print, "Textfile does not exist, let's create it!"

        if Grism eq 'OPEN' then begin
            print, "hmmm...smells like undispersed data"
            if keyword_set(sky) then begin
                chop_nod_phot, photfile, sky[0]
            endif else begin
                chop_nod_phot, photfile
            endelse
        endif else begin
            print, "ahh...tastes like dispersed data"
            if keyword_set(sky) then begin
                chop_nod_disp, photfile, sky[0]
            endif else begin
                chop_nod_disp, photfile
            endelse
        endelse
    endif

    got_flux_A = 0
    got_flux_B = 0
    openr,lun,phtxtfile,/get_lun
    while got_flux_A eq 0  or  got_flux_B eq 0 do begin
        line= ""
        readf,lun,line
        ;print,"Got: >",line,"<"
        pos = strpos(line,"Flux A in Photons/s")
        if pos ge 0 then begin
            FluxA = float(strmid(line,0,pos))
            got_flux_A = 1
            print,line
        endif
        pos = strpos(line,"Flux B in Photons/s")
        if pos ge 0 then begin
            FluxB = float(strmid(line,0,pos))
            got_flux_B = 1
            print,line
        endif
    end
    ;
    ; do we have to read the spectra, too?
    if N_params() gt 4 then begin
        print,"   reading spectra..."
        repeat begin
            line = ""
            readf,lun,line
            pos = strpos(line,"----- Table of flux per column in Photons/s -----")
        endrep until (pos ge 0)
        readf,lun,line          ;this should be "Beam A"
        readf,lun,line          ;this is the table head
        specA = fltarr(1000)
        i = 0
        repeat begin
            readf,lun,line
            pos = strpos(line,"Beam B")
            if (pos lt 0) then begin
                reads, line, x,y,flux,cflux
                specA[i] = cflux
                i = i + 1
            endif
        endrep until (pos ge 0)
        specA = specA[0:i-1]

        specB = fltarr(1000)
        i = 0
        while not eof(lun) do begin
            readf,lun, x,y,flux,cflux
            specB[i] = cflux
            i = i + 1
        endwhile
        specB = specB[0:i-1]
    endif
    close,lun & free_lun,lun
end

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
function check_mask_position, fringefile, maskfile

    rmsdata = Cached_MeanRMS(fringefile)
    maskdata= oirgetdata(maskfile)

    screensz= get_screen_size()
    datasz  = size(rmsdata.data1)
    factor  = floor( screensz[0]/datasz[1] < screensz[1]/4/datasz[2] < 5)
    print,"screen size: ", screensz
    print,"data size:   ", datasz[[1,2]]
    print,"enlarging windows by ",factor
    sz= size(rmsdata.data1) * factor
    window,0, xsize=sz[1], ysize=sz[2]*4.2
    loadct,0
    tvscl, rebin(sqrt(rmsdata[1].data2), sz[1], sz[2], /sample), 0,sz[2]*3.2
    tvscl, rebin(sqrt(rmsdata[1].data1), sz[1], sz[2], /sample), 0,sz[2]*1.0

    tvscl, rebin(sqrt(rmsdata[1].data2*maskdata[0].data2),sz[1],sz[2],/sample),0,sz[2]*2.2
    tvscl, rebin(sqrt(rmsdata[1].data1*maskdata[0].data1),sz[1],sz[2],/sample),0,0

    testOk= "string"
    read, '> Is the position of the mask satisfactory? (y/n) ', testOK
    if (strupcase(testOK) eq "N") then begin
        print, '<!> Sorry about that.  Better luck next time'
        return,0
    endif
    return,1
end

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
PRO fringe_movie, opd, fringe
    !P.multi= 0
    nScan = N_elements(fringe[0,*])
    xrange = [ min(opd), max(opd) ]
    yrange = [ min(fringe), max(fringe) ]
    for f=0,nScan-2 do begin
        plot, opd[*,f], fringe[*,f], xrange=xrange, yrange=yrange
        wait, .02
    endfor
end

PRO fourier_movie, foufri
    !P.multi= 0
    nScan = N_elements(foufri[0,*])
    yrange = [ 0, max(foufri) ]
    for f=0,nScan-2 do begin
        plot, foufri[*,f], yrange=yrange
        wait, .02
    endfor
end

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
FUNCTION compute_visi, corrFlux, FluxA, FluxB, VERBOSE=verbose

    if keyword_set(verbose) then begin
        print," correlated Flux =", corrFlux," photons/sec"
        print,"          Flux A =", FluxA," photons/sec"
        print,"          Flux B =", FluxB," photons/sec"
    endif
    fringe_contrast = corrFlux / (FluxA+FluxB)

    if keyword_set(verbose) then $
      print,"Fringe contrast = corr.Flux / (Flux A + Flux B) =", fringe_contrast

    phot_corr = 2d * sqrt(FluxA*FluxB) / (FluxA+FluxB)

    if keyword_set(verbose) then $
      print,"Phot.correction = 2*sqrt(Fl_A*Fl_B)/(Fl_A+Fl_B) =", phot_corr

    visibility = fringe_contrast / phot_corr

    if keyword_set(verbose) then begin
        print,"                                               +---------------------+"
        print,"Fringe contrast corrected (alias visibility) = |",visibility,'     |'
        print,"                                               +---------------------+"
    endif
    return, visibility
END

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
PRO midivisi, fringefile, photfileA, photfileB, MASKFILE=maskfile, $
              SKY=sky, ASKY=skyA, BSKY=skyB, $
              THRESHOLD=threshold, MOVIE=movie, NOFOU=nofou, NOPOSCHECK=noposcheck,$
              OPD=opd, FRINGE=fringe, FOUFRI=foufri

    fringefile= fringefile[0]
    photfileA = photfileA[0]

    !X.style = 1
    !Y.style = 1

    first_char    = STRPOS(fringefile,'/MIDI.20')+6
    last_char     = STRPOS(fringefile,'.fits')
    short_fringefile= STRMID(fringefile,first_char,last_char-first_char)

    if N_Params() gt 2  then begin
	; we have photometry A and B
        if not keyword_set(skyA) and keyword_set(sky) then skyA=sky
        read_photometry, photfileA, PhFilt, FluxA2, FluxA1, sky=skyA
        FluxA = FluxA1 + FluxA2

        if not keyword_set(skyB) and keyword_set(sky) then skyB=sky
        read_photometry, photfileB[0], PhFiltB, FluxB2, FluxB1, sky=skyB
        FluxB = FluxB1 + FluxB2

        if PhFilt ne PhFiltB then begin
            eniki= "any key"
            print, "Photometric filter A is ",PhFilt, ", but"
            print, "Photometric filter B is ",PhFiltB,"."
            read,  "Do you know what you are doing? ",eniki
            if eniki ne "y" then return
        endif
    endif else begin
        read_photometry, photfileA, PhFilt, FluxA, FluxB, sky=sky
    endelse

    print,"Filter: ",PhFilt,", Flux A=",FluxA, ", FluxB=",FluxB

    if NOT keyword_set(maskfile) then maskfile = make_maskname(photfileA)
    print,"Maskfile is ",maskfile
    short_maskfile= STRMID(maskfile, 11) ;rm "PHOTOMETRY/"

    ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

    if not keyword_set(noposcheck) then begin
        if check_mask_position(fringefile,maskfile) eq 0 then  return
    endif

    ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

    comprfile = 'PHOTOMETRY/Compr_'+short_fringefile+'_'+short_maskfile

    if file_test(comprfile) then begin
        print, comprfile, " exists"
    endif else begin
        cmd = getenv("MIAroot") + '/bin/oirCompress "'+ $
          fringefile +'" '+ maskfile +' '+ comprfile
        print, ''
        print, cmd
        spawn, cmd
    endelse
    print, ''
    print, 'getting data...'
    data = oirgetdata(comprfile)

    ; the following was (partly) stolen from midiRawFringe
    dtime = 86400. * (data[1].time[0] - data[0].time[0])
    tZero = 1.
    nZero = (tZero/dtime) > 10
    ;print,"dtime ",dtime,", nzero ",nzero

    ; we have to use the original filenames because oircompress
    ; removes all the interesting things from the header
    filename = oirFileList(fringefile)

    im = obj_new('fitsfile',filename[0])
    he = im->prihead()
    OBJECT = STRCOMPRESS(he->getpar('OBJECT'),/remove_all)
    DIT    = STRCOMPRESS(he->getpar('DET DIT'),/remove_all)
    FrFilt = STRCOMPRESS(he->getpar('INS FILT NAME'),/remove_all)
    obj_destroy,im

    ;   figure out scan size from header
    im = OBJ_NEW('imagedata', filename[0])
    scanLength = (im->head())->getpar('maxstep')

    ;   figure out channel location from OPT1
    beam = (im->prihead())->getpar('INS OPT1')
    OBJ_DESTROY,im

    print, "nZero",nzero,", scanlength",scanlength,", beam is ",beam
    ;    calculate scan size data
    nData = N_ELEMENTS(data)
    nScan = nData/scanLength
    print, nData," frames =>", nScan, " scans"
    if (beam EQ 'HIGH_SENS') then begin
        flux1 = data.data1
        flux2 = data.data2
    endif else if (beam EQ 'SCI_PHOT') then begin
        flux1 = data.data2
        flux2 = data.data3
    endif else begin
        print,"INS OPT1 is",beam
        print,"Neither HIGH_SENS nor SCI_PHOT, don't know where the fringes are!"
        return
    endelse
    fringe = flux1 - flux2
    maxflux= max( [ flux1, flux2 ] )
    minflux= min( [ flux1, flux2 ] )

    opd = data.localOpd[1]-data.localOpd[0] + data.opd[1]-data.opd[0]

    ; 1st two points are garbage
    fringe[0:1] = MEDIAN(fringe[2:200])
    ; last two opd points are garbage
    opd[nData-2:nData-1] = median(opd[nData-scanLength:nData-2])

    ; remove low freq trend
    fringe = fringe - SMOOTH(fringe, nZero,/EDGE_TRUNCATE)

    ; reform into scans, droping last fractional scan if any
    nFringe= LONG(scanLength)*LONG(nScan)
    fringe = REFORM(fringe(0:nFringe-1), scanLength, nScan,/overwrite)
    opd    = REFORM(opd(0:nFringe-1), scanLength, nScan,/overwrite)

    foufri = fltarr(scanLength, nScan)

    amplmax= fltarr(nScan,/nozero)
    amplfou= fltarr(nScan,/nozero)

    case FrFilt of
    ; "OPEN" added by FP, wavelength range just estimated
    'OPEN':  begin print,"filter: open"  & longwl=15.  & shortwl= 7.  & end
    'Nband': begin print,"N-band filter" & longwl=13.  & shortwl= 7.7 & end
    'N8.7':  begin print,"N 8.7 filter"  & longwl=10.3 & shortwl= 7.2 & end
    'SiC':   begin print,"SiC filter"    & longwl=13.0 & shortwl=10.5 & end
    else: begin print,"Unknown filter: ", FrFilt & return & end
    endcase

    ; last scan is crap
    for f=0,nScan-2 do begin
        amplmax[f] = max(abs(fringe[*,f]))
        foufri[*,f]= abs( fft( fringe[*,f])) ^ 2
        totopd = abs( opd[scanLength-1,f] - opd[0,f]) * 1e6
        lowfreq = floor(totopd/longwl+0.5)
        highfreq= floor(totopd/shortwl+0.5)
        if (highfreq gt scanLength/2) then highfreq= scanLength/2
        ;print, totopd/13.,"...",totopd/7.7
        amplfou[f] = 2. * sqrt(total( foufri[ lowfreq:highfreq, f]))
                                ;print,ampl[f]
    endfor
    if Keyword_set(movie) then begin
        if Keyword_set(nofou) then begin
            fringe_movie, opd, fringe
        endif else begin
            fourier_movie, foufri
        endelse
        eniki = ""
        read,"Hit return to continue ",eniki
    endif
    wait, .3
    wdelete,0
    scrsz = get_screen_size()
    window, 0, Title="Fringe Amplitudes",$
      xpos=0,ypos=scrsz[1]/2+30, xsize=scrsz[0]-8,ysize=scrsz[1]/2-30

    multi = !P.multi
    !p.multi= [0, 0, 4, 0, 0]
    !p.charsize= 1.7
    plot, opd,     xtitle = "Frame no.", title= "Optical Path Difference"
    plot, fringe,  xtitle = "Frame no.", title= "Raw Amplitude"
    plot, flux1, yrange=[minflux,maxflux], xtitle = "Scan no.", title = "Fluxes"
    oplot,flux2, linestyle=1
    multi2 = !P.multi
    ;plot, amplfou,  xrange=[0,nScan-2],  xtitle = "Scan no.", title = "Fourier Fringe Amplitude (linear)"
    ; xrange added by FP, otherwise crash!
    !p.charsize= 1.0



    if not Keyword_set(nofou) then begin
        fringe_ampl = max(amplfou)
    endif else begin
        fringe_ampl = max(amplmax)
    endelse
    print,"Maximum amplitude: ", fringe_ampl

    correl_flux = fringe_ampl * 145/DIT
    print,"   in photons/DIT: ", correl_flux*DIT
    visibi= compute_visi(correl_flux, FluxA, FluxB, /verbose)

    ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

    window, 1, title="Histogram" ; this is in lower right corner
    window, 3, title="Fringe-Spectra" ; this is in lower left corner

    hist = histogram(amplfou,NBINS=40,OMIN=hmin,OMAX=hmax)
    xhist= findgen(40) * (hmax-hmin)/40. + hmin

    if NOT keyword_set(threshold) then threshold= hmax/2.

    ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
    meanramp = fltarr(40)
    noiseramp= fltarr(40)

    for i=0,N_elements(xhist)-1 do begin
        thres = xhist[i]
        goodfringes= where(amplfou gt thres, Ngood)
        meanramp[i]= total(amplfou[goodfringes])/Ngood
        badfringes = where(amplfou lt thres, Nbad)
        if Nbad gt 0 then noiseramp[i]= total(amplfou[badfringes])/Nbad
        ;print,"Threshold:",thres," => mean",meanramp[i]
    endfor
    ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

    goodthresh= threshold & oldgthres = 0
    badthresh = threshold & oldbthres = 0

    repeat begin
        print, ''
        print, "Noise fringes <",badthresh
        print, "Object fringes >",goodthresh

        wset,0
        !p.multi= multi2
        !p.charsize= 1.7
        plot, amplfou, xtitle = "Scan no.", title = "Fourier Fringe Amplitude (linear)"
        if (Ngood gt 1) then begin
            oplot,goodfringes,replicate(oldgthres,Ngood),psym=3,color=0
            oplot,goodfringes,replicate(oldbthres,Ngood),psym=3,color=0
        endif
        goodfringes= where(amplfou gt goodthresh, Ngood)
        badfringes = where(amplfou lt badthresh,  Nbad)
        ;print,Ngood," good fringes"
        if (Ngood gt 1) then begin
            oplot,goodfringes,replicate(goodthresh,Ngood),psym=3
            oplot,goodfringes,replicate(badthresh,Ngood),psym=3
            oldgthres= goodthresh
            oldbthres= badthresh
        endif
        !P.multi = multi
        !P.charsize= 1.0

        wset,3
        !P.multi = [ 0,0,2,0,0 ]
        if (Ngood gt 0) then begin
            goodfou = foufri[*,goodfringes]
            plot, goodfou[*,0], yrange=[0,max(foufri)], $
              title="Good Fringes", xtitle= "Frequency", ytitle="Spectral Power"
            for f=1,Ngood-1 do oplot, goodfou[*,f]

            totopd = abs(opd[scanLength-1,goodfringes[0]]-opd[0,goodfringes[0]])*1e6
            lowfreq = floor(totopd/longwl+0.5)
            highfreq= floor(totopd/shortwl+0.5)
            if (highfreq gt scanLength/2) then highfreq= scanLength/2
            oplot,[ lowfreq, lowfreq],[0,max(foufri)],linestyle=2
            oplot,[highfreq,highfreq],[0,max(foufri)],linestyle=2

            if (Ngood gt 1) then begin
                totfou = total(goodfou,2) / Ngood
                oplot, totfou, color=100
            endif
        endif
        if (Nbad gt 0) then begin
            badfou = foufri[*,badfringes]
            plot, badfou[*,0], yrange=[0,max(foufri)], $
              title="Bad Fringes", xtitle= "Frequency", ytitle="Spectral Power"
            for f=1,Nbad-1 do oplot, badfou[*,f]

            if (Nbad gt 1) then begin
                totfou = total(badfou,2) / Nbad
                oplot, totfou, color=100
            endif
        endif
                                ;help,foufri,totfou

        ampl = 2.*sqrt(total(totfou[ floor(totopd/13.+0.5):floor(totopd/7.7+0.5)]))
        print, "scan length", totopd," micrometer"
        print, "Max. Amplitude of fouriers: ",max(amplfou)
        print, "Amplitude of mean fouriers: ",ampl

        wset,1
        !P.multi = [0,0,2,0,0]
        plot, xhist,meanramp,title="Mean amplitude as function of threshold",ystyle=2
        oplot,xhist,noiseramp,linestyle=1
        plot, xhist,hist, title="Histogram of Fourier-Amplitudes",$
          ystyle=2, psym=10
        oplot, [goodthresh,goodthresh], [0,max(hist)*1.2]
        oplot, [ badthresh, badthresh], [0,max(hist)*1.2], linestyle=1

        cursor, xmouse, ymouse, /DATA, /WAIT
        button= !Mouse.Button
        ;print, "position ", xmouse, ymouse, ", button", button
        !P.multi = multi

        endloop = 0
        case button of
            4: endloop = 1
            2:    badthresh = xmouse
            else: goodthresh= xmouse
        endcase
    endrep until endloop eq 1
    wset,0

    print, ""
    print, "Signal threshold:",goodthresh
    print, "Noise  threshold:", badthresh
    goodfringes= where(amplfou gt goodthresh)
    badfringes = where(amplfou lt badthresh)
    Ngood= N_elements(goodfringes)
    Nbad = N_elements(badfringes)

    good_ampl = total(amplfou[goodfringes])/Ngood
    bad_ampl  = total(amplfou[ badfringes])/Nbad
    fringe_ampl= sqrt(good_ampl*good_ampl - bad_ampl*bad_ampl)

    print, "mean integrated linear amplitude of signal:",good_ampl
    print, "mean integrated lunear amplitude of  noise:", bad_ampl
    print, "   corrected amplitude (sqrt(signal^2 - noise^2)):",fringe_ampl

    correl_flux = fringe_ampl * 145/DIT
    print,"   in photons/DIT: ", correl_flux*DIT
    visibi= compute_visi(correl_flux, FluxA, FluxB, /verbose)

    if PhFilt ne FrFilt then begin
        print, "! WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING !"
        print, "! Filters of Photometry and Fringes are not the same!                     !"
        print, "! Photometry was done with ",PhFilt,", Fringes with ",FrFilt
        print, "! WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING !"
    endif
end
