;
; This file is part of MIA.
;
; MIA is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 2 of the License, or
; (at your option) any later version.
;
; MIA is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with MIA; if not, write to the Free Software
; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
;
; $Id: utilities.pro,v 1.3 2003/11/08 06:44:19 koehler Exp $
;
; utilities.pro
; Created:     Fri Aug  8 11:34:40 2003 by Koehler@sun47
; Last change: Sun Oct 31 19:29:29 2004
;
; PURPOSE:
;	some utility routines related to xmidispvisi et al.
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; the official way to compute the trace_fit.
; guaranteed free of Chesneau-code!
; NOTE: it does not do the actual fitting,
;	it only creates the fitted trace!
;
; CALLING SEQUENCE:
;	trace_fit = make_trace(x_size,trace_coeff,sigma_coeff)
;
; INPUT:
;	x_size		size in x-direction (length of trace)
;	trace_coeff	coefficients of trace-fit
;	sigma_coeff	coefficients of sigma-fit
;
; OUTPUT:
;	2D-array:
;	trace_fit[0,*]	trace
;	trace_fit[1,*]	trace-FWHM
;	trace_fit[2,*]	trace+FWHM
;
Function make_trace, x_size, trace_coeff, sigma_coeff

  tf = fltarr(3,x_size)
  tf[0,*]= poly(findgen(x_size),trace_coeff)
  FWHM   = poly(findgen(x_size),sigma_coeff) * sqrt(2.*alog(2))
  tf[1,*]= tf[0,*] - FWHM
  tf[2,*]= tf[0,*] + FWHM
  return, tf
end

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; CALLING SEQUENCE:
;	plot_visibility, vis [, vis2...]
;
; INPUT:
;	up to 6 visibility-arrays as returned from xmidispvisi
;
; OUTPUT:
;	plot of visibility vs. lambda in current plot window
;
PRO plot_visibility, v0, v1, v2, v3, v4, v5,$
                     XTITLE=xtitle, YTITLE=ytitle, YRANGE=yrange, _EXTRA=extra

  if not keyword_set(xtitle) then xtitle= "Wavelength [micron]"
  if not keyword_set(ytitle) then ytitle= "Visibility"
  if not keyword_set(yrange) then yrange= [0,1]

  plot, v0[0,*], v0[2,*], YRANGE=yrange, XTITLE=xtitle, YTITLE=ytitle, _EXTRA=extra

  if N_params() gt 1 then oplot, v1[0,*], v1[2,*], linestyle=1
  if N_params() gt 2 then oplot, v2[0,*], v2[2,*], linestyle=2
  if N_params() gt 3 then oplot, v3[0,*], v3[2,*], linestyle=3
  if N_params() gt 4 then oplot, v4[0,*], v4[2,*], linestyle=4
  if N_params() gt 5 then oplot, v5[0,*], v5[2,*], linestyle=5

end

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; CALLING SEQUENCE:
;	mtv, data [, atv-flags]
;
; INPUT:
;	data - one slice of MIDI data (a struct with elements data1
;              and data2)
;	atv-flags - all flags accepted by atv
;
; OUTPUT:
;	calls atv to display the frame data
;
PRO mtv, data, _EXTRA = extra
  data = data[0]	;just in case someone gave us an array

  sz1 = size(data.data1)
  sz2 = size(data.data2)
  sz_x= sz1[1] + 1 + sz2[1]

  tags = strtrim(tag_names(data),2)
  if where(tags eq "DATA3") ge 0 then begin
      ; there is a data3 field, so this is probably sci-phot data
      sz3 = size(data.data3)
      sz4 = size(data.data4)

      sz_x= max( [sz_x, sz3[1] + 1 + sz4[1]] )
      sz_y= max( [sz1[2]+1+sz3[2], sz2[2]+1+sz4[2]] )

      img = fltarr(sz_x,sz_y)
      img[     0  : sz1[1]-1, sz3[2]+1 : sz_y-1 ] = data.data1
      img[ sz1[1]+1 : sz_x-1, sz4[2]+1 : sz_y-1 ] = data.data2
      img[     0  : sz3[1]-1,     0  : sz3[2]-1 ] = data.data3
      img[ sz3[1]+1 : sz_x-1,     0  : sz4[2]-1 ] = data.data4

  endif else begin
      sz_y= max( [sz1[2],sz2[2]] )

      img = fltarr(sz_x,sz_y)
      img[   0 : sz1[1]-1 ,* ] = data.data1
      img[ sz1[1]+1:sz_x-1,* ] = data.data2
  endelse
  ;print,sz_x," x ",sz_y

  atv, img, _EXTRA=extra
end

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; MIDI-filename processing:
;
; MIDIname_date(filename)   returns date-part of filename
; MIDIname_time(filename)   returns time-part of filename
; MIDIname_datime(filename) returns date- and time-part
;
Function MIDIname_date, fname
  first_char= Strpos(fname,'/MIDI.20')+6
  fname	    = Strmid(fname,first_char)
  last_char = Strpos(fname,'T')
  return, Strmid(fname,0,last_char)
end

Function MIDIname_time, fname
  first_char= Strpos(fname,'/MIDI.20')+6
  first_char= first_char + Strpos(Strmid(fname,first_char),'T')+1
  last_char = Strpos(fname,'.fits')
  return, Strmid(fname,first_char,last_char-first_char)
end

Function MIDIname_datime, fname
  first_char= Strpos(fname,'/MIDI.20')+6
  last_char = Strpos(fname,'.fits')
  return, Strmid(fname,first_char,last_char-first_char)
end


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; Wrappers for C-programs
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

Function Cached_MeanRMS, filename
    first_char= STRPOS(filename,'/MIDI.20')+6
    last_char = STRPOS(filename,'.fits')
    rmsname = "PHOTOMETRY/Mean_" + $
      STRMID(filename,first_char,last_char-first_char) + ".fits"
    ; this gets rid of the other filenames in the list, too

    print,'Mean+RMS file is "', rmsname,'"'

    if not file_test(rmsname) then begin
        print,"Computing mean&rms..."
        cmd = getenv("MIAroot") + '/bin/oirMeanRMS'
        if file_test(cmd,/EXECUTABLE) then begin
            if not file_test("PHOTOMETRY") then spawn,'mkdir "PHOTOMETRY"'
            cmd = cmd + ' "' + filename + '" ' + rmsname
            print, cmd
            spawn, cmd
        endif else begin
            print, cmd + " not found, falling back on IDL code"
            return, oirgetmeanrms(filename)
        endelse
    endif
    return, oirgetdata(rmsname)
end

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

Function MIDIChopImageC, filename, BEFORE=before, AFTER=after

  ; default values used in MIDIChopImage (at least on 24-Jul-2003)

  if not keyword_set(before) then before= 0
  if not keyword_set(after)  then after = 1

  first_char= STRPOS(filename,'/MIDI.20')+6
  last_char = STRPOS(filename,'.fits')
  short_name= STRMID(filename,first_char,last_char-first_char)

  cachefile = string(Format='(%"PHOTOMETRY/Chop_%s_b%d_a%d.fits")',$
                     short_name, before, after)

  if not file_test(cachefile) then begin
      print, "Cached chop-frame does not exist, let's create it!"

      cmd = getenv("MIAroot") + '/bin/midiChopimg'
      if file_test(cmd,/EXECUTABLE) then begin
          if not file_test("PHOTOMETRY") then spawn,'mkdir "PHOTOMETRY"'
          cmd = cmd +	' -b ' + strtrim(before,2) + $
			' -a ' + strtrim(after,2) + $
			' "' + filename + '" ' + cachefile
          print,cmd
          spawn,cmd
      endif else begin
          print, cmd + " not found,"
          print, "falling back on IDL-code"
          chop_image= MIDIChopImage(filename, before=before, after=after)
          oirNewData, (oirFileList(filename))[0], cachefile, chop_image
          return, chop_image
      endelse
  endif
  chop_image= oirgetdata(cachefile)
  return, chop_image
end

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; maybe we can add nodding to the same function?

Function MIDISkySubImage, filename, BEFORE=before, AFTER=after

  chop_image= MIDIChopImageC(filename, BEFORE=before, AFTER=after)

  im = obj_new('imagedata',(oirFileList(filename))[0])
  he = im->prihead()
  beam= he->getpar('INS OPT1')
  obj_destroy,im
  if (beam EQ 'HIGH_SENS') then begin
      print,"MIDISkySubImage: High_sens"
      win1 = chop_image[0].data1 - chop_image[1].data1
      win2 = chop_image[0].data2 - chop_image[1].data2
      return, { int1: win1, int2: win2 }

  endif else if (beam EQ 'SCI_PHOT') then begin
      print,"MIDISkySubImage: sci_phot"
      win1 = chop_image[0].data2 - chop_image[1].data2
      win2 = chop_image[0].data3 - chop_image[1].data3
      phot1= chop_image[0].data1 - chop_image[1].data1
      phot2= chop_image[0].data4 - chop_image[1].data4
      return, { int1: win1, int2: win2, phot1: phot1, phot2: phot2 }

  endif else begin
      print,"INS OPT1 is",beam
      print,"Neither HIGH_SENS nor SCI_PHOT, don't know where the data is!"
      return, -1
  endelse
end

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

Function getCompressSpec, fringefile, maskfile, comprfile, REUSE=reuse, DWIM=dwim

  do_compress = 1
  if keyword_set(reuse) and file_test(comprfile) then begin
      do_compress = 0
      if not keyword_set(dwim) then begin
          print, ""
          print, "Compressed file exists already:"
          print, comprfile
          test="string" & read, "Shall I use it? [Y/n] ",test
          if (test eq 'n') then do_compress = 1
      endif
  endif
  if (do_compress eq 1) then begin
      if not file_test("PHOTOMETRY") then spawn,'mkdir "PHOTOMETRY"'
      print, ''
      cmd = getenv("MIAroot") + '/bin/oirCompressSpec'
      if file_test(cmd,/EXECUTABLE) then begin
          if file_test(comprfile) then file_delete, comprfile
          cmd = cmd + ' "' + fringefile +'" '+ maskfile +' '+ comprfile
          print, cmd
          spawn, cmd
      endif else begin
          print, cmd + " not found, falling back on IDL code"
          return, oirCompressData(fringefile, maskfile, nDim=1)
      endelse
  endif
  print, ''
  print, 'getting data...'
  return, oirgetdata(comprfile)
end
