;
; This file is part of MIA.
;
; MIA is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 2 of the License, or
; (at your option) any later version.
;
; MIA is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with MIA; if not, write to the Free Software
; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
;
;********************************************************
;#class#
;fringeUtilities
;#description#
;some utilities that are MIDI dependent for analyzing fringes
;#end_class#
;*******************************************************

FUNCTION midiRawFringe, inputFiles, maskfile, time, opd, tZero=tZero, $
   flipOpd=flipOpd, iErr=iErr, noC=noC
;********************************************************
;#function#
;midiRawFringe
;#description#
;calculate non dispersed fringes by compression and subtraction
;#call#
;Fringe=midiRawFringe(inputFiles, maskfile, time, opd, tZero=tZero, flipOpd=flipOpd)
;#inputs#
;\anArg{inputFiles}{string }{file(s) containing table format data}
;\anArg{maskfile}{string }{file containing weighting arrays}
;\anArg{time}{array of flt }{optional output array with time data}
;\anArg{time}{array of flt }{optional output array with opd data}
;\anArg{tZero}{int }{optional size of trend removal highpass filter; def=1sec}
;\anArg{flipOpd}{int }{optional interpret VLT opd with opposite sign}
;#return#
;\anArg{-}{data cube}{array of interferometric spectra in scans}
;#end_function#
;********************************************************
;   get compressed data
   if (KEYWORD_SET(noC)) then $
      data = oirCompressData(inputFiles, maskfile, opd=opd, time=time, $
         flipOpd = flipOpd, iErr=iErr) $
   else $
      data = oirCompressDataC(inputFiles, maskfile, opd=opd, time=time, $
         flipOpd = flipOpd ) 
;   figure out default tZero
   dtime = time[1,0] - time[0,0]
   if (N_ELEMENTS(tZero) LE 0) then tZero = 1.
   nZero = (tZero/dtime) > 10
;   figure out scan size from header
   filename = oirFileList(inputFiles)
   im = OBJ_NEW('imagedata', filename[0])
   imh = im->head()
   scanLength = imh->getpar('maxstep')
;    figure out channel location from OPT1
   prih = im->prihead()
   beam = prih->getpar('INS OPT1')
   OBJ_DESTROY,im
;    calculate scan size data
   nData = N_ELEMENTS(data)
   nScan = nData/scanLength
   if (beam EQ 'HIGH_SENS') then fringe = data.data1-data.data2 $
      else if (beam EQ 'SCI_PHOT') then fringe = data.data2 - data.data3
; 1st two points are garbage
   fringe[0:1] = MEDIAN(fringe[2:200])
; last two opd points are garbage
   opd[nData-2:nData-2] = median(opd[nData-200:nData-10])
; remove low freq trend
   fringe = fringe - SMOOTH(fringe, nZero,/EDGE_TRUNCATE)
; reform into scans, droping last fractional scan if any
   nFringe = LONG(scanLength)*LONG(nScan)
   fringe = REFORM(fringe[0:nFringe-1], scanLength, nScan,/overwrite)
   opd = 1.e6*REFORM(opd[0:nFringe-1], scanLength, nScan,/overwrite)
   time = REFORM(time[0:nFringe-1], scanLength, nScan,/overwrite)
RETURN, fringe
END

FUNCTION midiRawFringeSpec, inputFiles, maskfile, time, opd, tZero=tZero, $
   flipOpd=flipOpd, noC = noC, noScan=noScan, iErr=iErr
;********************************************************
;#function#
;midiRawFringeSpec
;#description#
;calculate spectrally dispersed fringes by compression and subtraction
;#call#
;SpectralFringe=midiRawFringeSpec(inputFiles, maskfile, timd, opd, tZero=tZero)
;#inputs#
;\anArg{inputFiles}{string }{file(s) containing table format data}
;\anArg{maskfile}{string }{file containing weighting arrays}
;\anArg{time}{array of flt }{optional output array with time data}
;\anArg{time}{array of flt }{optional output array with opd data}
;\anArg{tZero}{int }{optional size of trend removal highpass filter}
;\anArg{flipOpd}{int }{optional interpret VLT opd with opposite sign}
;#return#
;\anArg{-}{data cube}{array of interferometric spectra in scans}
;#end_function#
;********************************************************
;  get detector with window coordinates
   if (KEYWORD_SET(noC)) then $
      data = oirCompressData(inputFiles, maskfile, opd=opd, time=time, $
         nDim=1, flipOpd = flipOpd, iErr=iErr) $
   else $
      data = oirCompressDataC(inputFiles, maskfile, opd=opd, time=time, $
         nDim=1, flipOpd = flipOpd )
;   figure out default tZero
   dtime = (time[1,0] - time[0,0])
   if (N_ELEMENTS(tZero) LE 0) then tZero = 1.
   nZero = (tZero/dtime) > 10
   filename = oirFileList(inputFiles)
;  get image data header to read scan size
   im = OBJ_NEW('imagedata', filename[0])
   imh = im->head()
   scanLength = imh->getpar('maxstep')
;  figure out detector layout, for now either 4 windows or two
   prih = im->prihead()
   beam = prih->getpar('INS OPT1')
   OBJ_DESTROY,im
;   get sizes so we can break result into scans
   nData = N_ELEMENTS(data)
   nScan = nData/scanLength
;   subtract I1 - I2
   if (beam EQ 'HIGH_SENS') then begin
      size1 = SIZE(data[0].data1)
      if (0 NE TOTAL(size1 NE SIZE(data[0].data2))) then begin
         print,'I1 and I2 spectra are of different size'
         return,0
      endif
      fringe = data.data1-data.data2 
   endif else if (beam EQ 'SCI_PHOT') then begin
      size2 = SIZE(data[0].data2)
      if (0 NE TOTAL(size2 NE SIZE(data[0].data3))) then begin
         print,'I1 and I2 spectra are of different size'
         return,0
      endif
      fringe = data.data2 - data.data3
   endif
   fringeSize = SIZE(fringe)
   specSize   = fringeSize[1]
; 1st two points are garbage
   fringe[*,0] = fringe[*,2]
   fringe[*,1] = fringe[*,2]
; last two opd points are garbage
   opd[nData-2:nData-2] = median(opd[nData-200:nData-10])
; remove low freq trend
   for i=0, specSize-1 do fringe[i,*] = fringe[i,*]  - $
      SMOOTH(REFORM(fringe[i,*]), nZero,/EDGE_TRUNCATE)
; reform into scans, droping last fractional scan if any
   nFringe = LONG(scanLength)*LONG(nScan)
   if (KEYWORD_SET(noScan)) then begin
      opd = 1.e6*REFORM(opd); remove bogus dimensions
      time = REFORM(time); remove bogus dimensions
   endif else begin
      fringe = REFORM(fringe[*,0:nFringe-1], specSize, scanLength, nScan)
      opd = 1.e6*REFORM(opd[0:nFringe-1], scanLength, nScan)
      time = REFORM(time[0:nFringe-1], scanLength, nScan)
   endelse
RETURN, fringe
END

FUNCTION midiShiftSpecFringe, rawSpec, opd, waveNo
;********************************************************
;#function#
;midiShiftSpecFringe
;#description#
;Rotate fringes in spectral data to nominal 0 OPD.
;#call#
;Fringe=midiShiftSpecFringe(rawSpec, opd, waveNo) 
;#inputs#
;\anArg{Raw}{array of flt }{output of midiRawFringeSpec} 
;\anArg{opd}{array of flt }{array of reported opds (micron)}
;\anArg{waveNo}{array of flt }{array of waveNo; output of oirGetWaveNo;rad/micron}
;#return#
;\anArg{-}{array of complex}{Raw shifted by exp(i*opd*spfreq)}
;#end_function#
;********************************************************
   rawSize = SIZE(rawSpec,/dimensions)
   nRaw = N_ELEMENTS(rawSize)
   n2 = 1
   for i=1,nRaw-1 do n2=n2*rawSize[i]
   roll = TRANSPOSE(REFORM(COMPLEX(rawSpec), rawSize[0], n2),$
      [1,0])
   nOpd = REFORM(opd, n2)
   ii = -COMPLEX(0., 1.) * waveno
   for i=0, n2-1 do roll(i,*)=roll(i,*)*exp(ii*nOpd[i])
RETURN, roll
END

FUNCTION midiAutoSpec,spec, xsmooth
;********************************************************
;#function#
;midiAutoSpec
;#description#
;calculate by FFT autocorrelation (in time) per freq
;#call#
;auto = midiAutoSpec(spectrum, xsmooth)
;#inputs#
;\anArg{spec}{2-d array of complex }{output of midiShiftSpecFringe}
;\anArg{xsmooth}{float}{smooth spec with xsmooth 1st to supress image info}
;#return#
;\anArg{-}{2-dlarray of float}{autoCorrelation fn [delay,freq]}
;#end_function#
;********************************************************
   if (N_PARAMS() LT 2) then xsmooth = 1.5
   specSize = SIZE(spec,/dimensions)
   auto = fltarr(specSize)
;  weight = auto
;  if (N_ELEMENTS(dataWeight) LE 0) then weight(*) = 1. $
;  else begin
;     weightSize = SIZE(dataweight)
;  endif
   for i=0, specSize[1]-1 do begin
;     sr = fsmooth1(float(spec(*,i)),xsmooth)
;     si = fsmooth1(imaginary(spec(*,i)),xsmooth)
;     sc = COMPLEX(sr,si)
      sc = csmooth1(reform(spec[*,i]), xsmooth)
      auto(*,i)=float(FFT(ABS(FFT(sc,1))^2,-1))
    endfor
RETURN,auto
END

FUNCTION midiAutoPeak, auto, dTime, cTime
;********************************************************
;#function#
;midiAutoPeak
;#description#
;interpolate value of autocorrelation fn at zero offset for each freq
;#call#
;autoPeak = midiAutoPeak(auto, dTime, corrTime)
;#inputs#
;\anArg{auto}{2-d array of flt }{output of midiAutoSpec[time,freq]} 
;\anArg{dTime}{float }{time interval between points}
;\anArg{cTime}{float }{max. time to include in fit}
;#return#
;\anArg{-}{array of float }{per freq interpolated autoCorr at zero offset}
;#end_function#
;********************************************************
   autoSize = SIZE(auto,/dimensions)
   peak = fltarr(autoSize[1])
   nFit = fix(cTime/dTime)
   xx =FINDGEN(2*nFit+1)-nFit
   ww = ABS(xx) GT 1
   for i=0, autoSize[1]-1 do begin
      a = SHIFT(auto(*,i), nFit)
      b=a(0:2*nFit)
      c = POLYFITW(xx,b,ww,2,y)
      peak(i) =y[nFit]
   endfor
RETURN, peak
END

FUNCTION midiSpectralPower, inputFiles, maskFile, freqFile, $
   waveNo, cTime, xSmooth=xSmooth, flipOpd=flipOpd, iErr=iErr
;********************************************************
;#function#
;midiSpectralPower
;#description#
;Compute (incoherent) correlated spectral power from dispersed fringes
;#call#
;spPower = midiSpectralPower(input,File, maskFile, freqFile,$
;           waveNo, cTime, flipOpd=flipOpd)
;#inputs#
;\anArg{inputFiles}{string}{space-separated names of data files}
;\anArg{maskFile}{string}{name of file containing weighting mask}
;\anArg{freqFile}{string}{name of file containing frequency data}
;\anArg{waveNo}{arr of flt}{waveNos (output) of data arrays}
;\anArg{cTime}{float}{estimate of coherence time of OPD,sec;def=1.}
;\anArg{flipOpd}{?}{if set, flip DelayLine OPD before use}
;#return#
;\anArg{-}{array of float }{spectral power at each waveNo}
;#end_function#
;********************************************************
   if (N_PARAMS() LT 5) then cTime = 1.
   if (N_ELEMENTS(xSmooth) LT 1) then xSmooth=1.5
   print,'Getting data'
   rawSpec= midiRawFringeSpec(inputFiles[0], maskfile, time, opd, $
      flipOpd=flipOpd, iErr=iErr)
   dTime = time(2,0) - time(1,0)
   t = STRSPLIT(freqFile,' ',/extract)
   t = t[0]
   waveNo = oirGetWaveNo(t)
   print,'Rotating data'
   shiftFringe = midiShiftSpecFringe(rawSpec, opd, waveNo)
   auto = midiAutoSpec(shiftFringe, xSmooth)
   peak = midiAutoPeak(auto, dTime, cTime)
RETURN,peak
END

FUNCTION midiChopImage, inputFiles, before=before, after=after
;********************************************************
;#function#
;midiChopImage
;#description#
;average separately Target and Sky images in chopped data
;#call#
;ChoppedDataStructure = midiChopImage(inputFiles)
;ChoppedDataStructure = midiChopImage(inputFiles, before=3, after=2)
;#inputs#
;\anArg{inputFiles}{string or array of strings }{file(s) containing data}
;\anArg{before}{int}{optional:no. of points before a change of phase to drop (def=0)}
;\anArg{after}{int}{optional:no. of points after a change of phase to drop (def=1)}
;#return#
;\anArg{-}{array[2] of struct}{[0]=average of Target-Sky images;[1]=average of Sky images}
;#end_function#
;********************************************************
; set defaults
   if (N_ELEMENTS(before) LT 1) then before = 0
   if (N_ELEMENTS(after) LT 1)  then after  = 1
; check if single input or multiple
   filename = oirFileList(inputFiles)
   fileSize = SIZE(filename)
   if (fileSize[0] eq 0) then nFiles = 1 $
      else nFiles=N_ELEMENTS(filename)
   fileRows = intarr(nFiles)
; one pass through data to get chopping phase information
; maybe this is unncessary but before/after is difficult without
; a separate pass
   for iFile = 0, nFiles-1 do begin
      imageObj = OBJ_NEW('imagedata', filename[iFile], ierr=ierr)
      if (ierr ne 0) then begin
         ewsPrintErrMsg
         return,0
      endif
      if (iFile EQ 0) then target = imageObj->readrows(col='tartyp2') $
         else target = [target, imageObj->readrows(col='tartyp2') ]
      naxis = imageObj->naxis()
      fileRows[iFile] = naxis[1]
      OBJ_DESTROY, imageObj
   endfor
; edit target list
   target  = target.tartyp2
   nTarget = N_ELEMENTS(target)
   iTarget = intarr(nTarget)
   for i= 0, nTarget-1 do begin
      if ((i LT after) OR (i GE (nTarget - before))) then iTarget[i] = 0 $
      else if ((target[i] NE target[i+before]) OR (target[i] NE target[i-after])) then iTarget[i]=0$
      else if (target[i] EQ 'T ') then iTarget[i] = 1 $
      else if (target[i] EQ 'S ') then iTarget[i] = -1
   endfor
; create output array
   imageObj = obj_new('imagedata', filename[0], ierr=ierr)
   dataColumns = imageObj->dataColumns()
   targetColumns = imageObj->targetColumns()
   nData = N_ELEMENTS(dataColumns)
   data = imageObj->readRowsf(1, col=dataColumns, ierr=ierr)
   outData = REPLICATE(data,2)
   for iData = 0, nData-1 do outData.(iData) = 0.
   OBJ_DESTROY,imageObj
   readColumns = [dataColumns,targetColumns[0]]
;
; initialize counters
   row = 0
; read and select data
   for iFile = 0, nFiles-1 do begin
      imageObj = obj_new('imagedata', filename[iFile], ierr=ierr)
      for iRow = 1, fileRows[iFile] do begin
         data = imageObj->readRowsf(iRow, col=readColumns, ierr=ierr)
         if (iTarget[row] EQ -1) then $
            for iData = 0,nData-1 do outData[1].(iData)=$
               outData[1].(iData) + data.(iData) $
         else if (iTarget[row] EQ 1) then $
            for iData = 0,nData-1 do outData[0].(iData)=$
               outData[0].(iData) + data.(iData)
         row = row + 1
      endfor
      OBJ_DESTROY, imageObj
   endfor
; average data
   nS = TOTAL(iTarget EQ -1)
   if (nS GT 0) then $
      for iData = 0,nData-1 do outData[1].(iData)=$
               outData[1].(iData)/nS
   nT = TOTAL(iTarget EQ 1)
   if (nT GT 0) then $
      for iData = 0,nData-1 do outData[0].(iData)=$
               outData[0].(iData)/float(nT)
    for iData = 0,nData-1 do outData[0].(iData) = outData[0].(iData)-outData[1].(iData)
RETURN, outdata
END


FUNCTION midiChopImageArray, inputFiles, before=before, after=after, blockno=blockno
;********************************************************
;#function#
;midiChopImageArray
;#description#
;average Target and Sky images in block chopped data
;#call#
;ChoppedDataStructure = midiChopImageArray(inputFiles)
;ChoppedDataStructure = midiChopImageArray(inputFiles, before=3, after=2)
;#inputs#
;\anArg{inputFiles}{string or array of strings }{file(s) containing data}
;\anArg{before}{int}{optional:no. of points before a change of phase to drop (def=0)}
;\anArg{after}{int}{optional:no. of points after a change of phase to drop (def=1)}
;#return#
;\anArg{-}{array[2] of struct}{array of images containing average of each block of S/T images}
;#end_function#
;********************************************************
; set defaults
   if (N_ELEMENTS(before) LT 1) then before = 0
   if (N_ELEMENTS(after) LT 1)  then after  = 1
; check if single input or multiple
   filename = oirFileList(inputFiles)
   fileSize = SIZE(filename)
   if (fileSize[0] eq 0) then nFiles = 1 $
      else nFiles=N_ELEMENTS(filename)
   fileRows = intarr(nFiles)
; one pass through data to get chopping phase information
; maybe this is unncessary but before/after is difficult without
; a separate pass
   for iFile = 0, nFiles-1 do begin
      imageObj = OBJ_NEW('imagedata', filename[iFile], ierr=ierr)
      if (ierr ne 0) then begin
         ewsPrintErrMsg
         return,0
      endif
      if (iFile EQ 0) then target = imageObj->readrows(col='tartyp2') $
         else target = [target, imageObj->readrows(col='tartyp2') ]
      naxis = imageObj->naxis()
      fileRows[iFile] = naxis[1]
      OBJ_DESTROY, imageObj
   endfor
; edit target list
   target  = target.tartyp2
   nTarget = N_ELEMENTS(target)
   iTarget = intarr(nTarget)
   for i= 1, nTarget-1 do begin
      if ((i LT after) OR (i GE (nTarget - before))) then iTarget[i] = 0 $
      else if ((target[i] NE target[i+before]) OR (target[i] NE target[i-after])) then iTarget[i]=0$
      else if (target[i] EQ 'T ') then iTarget[i] = 1 $
      else if (target[i] EQ 'S ') then iTarget[i] = -1
   endfor
;flag first point
   iTarget[0] = 0
; group images in blocks
   tTarg = intarr(nTarget)
   tTarg[0] = -1
   tt = -1
   for iT = 1, nTarget -1 do begin
      if (iTarget[iT] EQ 0) then tTarg[iT] = -1 $
      else begin
         if (iTarget[iT] NE iTarget[iT-1]) then tt = tt + 1
         tTarg[iT] = tt
      endelse
   endfor
   tt = tt+1
   weights = INTARR(tt)
   for i=0,tt-1 do weights[i] = TOTAL(tTarg EQ i)
; create output array
   imageObj = obj_new('imagedata', filename[0], ierr=ierr)
   dataColumns = imageObj->dataColumns()
   targetColumns = imageObj->targetColumns()
   nData = N_ELEMENTS(dataColumns)
   data = imageObj->readRowsf(1, col=[dataColumns,targetColumns], ierr=ierr)
   for iData = 0, nData-1 do data.(iData) = 0.
   outData = REPLICATE(data,tt)
   OBJ_DESTROY,imageObj
   readColumns = [dataColumns,targetColumns[0:1]]
; initialize row counter
   row = 0
; read and select data
   for iFile = 0, nFiles-1 do begin
      imageObj = obj_new('imagedata', filename[iFile], ierr=ierr)
      for iRow = 1, fileRows[iFile] do begin
         data = imageObj->readRowsf(iRow, col=readColumns, ierr=ierr)
         if (tTarg[row] GE 0) then begin
            for iData = 0,nData-1 do outData[tTarg[row]].(iData)=$
               outData[tTarg[row]].(iData) + data.(iData) 
            outData[tTarg[row]].tartyp2 = data.tartyp2
         endif
         row = row + 1
      endfor
      OBJ_DESTROY, imageObj
   endfor
; average data
   for iRow = 0, tt-1 do $
      if (weights[iRow] GT 0) then $
         for iData = 0,nData-1 do outData[iRow].(iData)=$
            outData[iRow].(iData)/weights[iRow]
; if requested, output block target numbers
   blockno = tTarg
RETURN, outdata
END


FUNCTION midiChopFrames, inputFiles, before=before, after=after
;********************************************************
;#function#
;midiChopFrames
;#description#
;average Sky images and subtract from individual Object frames
;#call#
;ChoppedDataStructure = midiChopFrames(inputFiles)
;ChoppedDataStructure = midiChopFrames(inputFiles, before=3, after=2)
;#inputs#
;\anArg{inputFiles}{string or array of strings }{file(s) containing data}
;\anArg{before}{int}{optional:no. of points before a change of phase to drop (def=0)}
;\anArg{after}{int}{optional:no. of points after a change of phase to drop (def=1)}
;#return#
;\anArg{-}{array[2] of struct}{array of images containing average of each block of S/T images}
;#end_function#
;********************************************************
; get block averages
   PRINT,'Calculate block averages'
   blockData = midiChopImageArray(inputFiles, before=before, after=after,blockno=blockTarg)
   maxBlock = MAX(blockTarg)
; set defaults
   if (N_ELEMENTS(before) LT 1) then before = 0
   if (N_ELEMENTS(after) LT 1)  then after  = 1
; check if single input or multiple
   filename = oirFileList(inputFiles)
   fileSize = SIZE(filename)
   if (fileSize[0] eq 0) then nFiles = 1 $
      else nFiles=N_ELEMENTS(filename)
   fileRows = intarr(nFiles)
; one pass through data to get file sizes
   for iFile = 0, nFiles-1 do begin
      imageObj = OBJ_NEW('imagedata', filename[iFile], ierr=ierr)
      if (ierr ne 0) then begin
         ewsPrintErrMsg
         return,0
      endif
      naxis = imageObj->naxis()
      fileRows[iFile] = naxis[1]
      OBJ_DESTROY, imageObj
   endfor
; create output array
   imageObj = obj_new('imagedata', filename[0], ierr=ierr)
   dataColumns = imageObj->dataColumns()
   targetColumns = imageObj->targetColumns()
   nData = N_ELEMENTS(dataColumns)
   data = imageObj->readRowsf(1, col=[dataColumns,targetColumns], ierr=ierr)
   for iData = 0, nData-1 do data.(iData) = 0.
   outData = REPLICATE(data,TOTAL(blockTarg NE 0))
   OBJ_DESTROY,imageObj
   readColumns = [dataColumns,targetColumns[0:1]]
; initialize row counter
   row    = 0
   outRow = 0
; read and select data
   for iFile = 0, nFiles-1 do begin
      imageObj = obj_new('imagedata', filename[iFile], ierr=ierr)
      for iRow = 1, fileRows[iFile] do begin
         data = imageObj->readRowsf(iRow, col=readColumns, ierr=ierr)
; check if "U" "S" or "T"
         if (blockTarg[row] GE 0) then begin
; if sky subtract average for its own block
            if (data.tartyp2 EQ 'S ') then $
               for iData = 0,nData-1 do outData[outRow].(iData)=$
                  data.(iData) - blockData[blockTarg[row]].(iData) 
; if target subtract average of preceding and following block
            if (data.tartyp2 EQ 'T ') then begin
               precede = blockTarg[row]-1
               if (precede LT 0) then precede = blockTarg[row]+1
               follow = blockTarg[row]+1
               if (follow GT maxBlock) then follow = precede
               for iData = 0,nData-1 do outData[outRow].(iData)=$
                  data.(iData) - 0.5*(blockData[precede].(iData) $
                     + blockData[follow].(iData))
            endif
            outData[outRow].tartyp2 = data.tartyp2
            outRow = outRow + 1
         endif
         row = row + 1
      endfor
      OBJ_DESTROY, imageObj
   endfor
RETURN, outdata
END

FUNCTION midiFilterFringes, rawFringes, opd, filterFile, outputOpd, $
   outType=outType
;********************************************************
;#function#
;midiFilterFringes
;#description#
;filter raw fringes to accept given frequency range
;#call#
;Filtered=midiFilterFringes(rawFringes, opd, filterFile, outputOpd)
;#inputs#
;\anArg{rawFringes}{array of flt }{output of midiRawFringes}
;\anArg{opd}{array of flt }{opd output of midiRawFringes}
;\anArg{filterFile}{string}{name of file containing filter discription}
;\anArg{GridData}{flt[3] }{specification of opd grid containing output data}
;\anArg{}{}{[lowest opd(m), highest opd[m], number of pts(2^n)]}
;\anArg{outputOpd}{arr of flt}{1d array of opd values (m) for filtered data}
;\anArg{outType}{string}{'AMP' (default), 'COMPLEX', 'REAL'}
;#return#
;\anArg{-}{data cube}{array of filtered interferometric spectra in scans}
;#end_function#
;********************************************************
;  check defaults
   if (N_ELEMENTS(outType) LE 0) then outType = 'AMP'
   outType =STRUPCASE(outType)
;  set up fourier transform 
   gridData=[min(opd), (max(opd)-min(opd))/1024.,1024.]
   fourier = OBJ_NEW('fft1d', gridData)
   filter = OBJ_NEW('filterData', filterFile, iErr=iErr)
   if (iErr NE 0) then begin
      ewsPrintErrMsg
      RETURN,0
   endif
   fourier->set, filter=filter->interpolate(1.e6*fourier->kGrid())
;  fourier->set, filter=1.
   OBJ_DESTROY, filter
   rawSize = SIZE(rawFringes)
   nScans = rawSize[2]
   filterFringe = fltarr(GridData[2], nScans)
   if (outType EQ 'REAL') then $
      for iScan = 0, nScans-1 do filterFringe(*, iScan) = $
      FLOAT(fourier->Cfilter(opd(*,iScan), rawFringes(*,iScan))) $
   else if (outType EQ 'COMPLEX') then $
      for iScan = 0, nScans-1 do filterFringe(*, iScan) = $
      (fourier->Cfilter(opd(*,iScan), rawFringes(*,iScan))) $
   else for iScan = 0, nScans-1 do filterFringe(*, iScan) = $
      ABS(fourier->Cfilter(opd(*,iScan), rawFringes(*,iScan))) 
   outputOpd = fourier->coord()
   OBJ_DESTROY, fourier
RETURN, filterFringe
END

FUNCTION midiFilterDisp, rawFringes, opd, detectorFile, outputOpd, $
   outType=outType, compressScans=compressScans, noShift=noShift
;*************************************************************
;#function#
;midiFilterDisp
;#description#
;fft dispersed fringes into OPD domain
;#call#
;Filtered=midiFilterDisp(rawFringes, opd, detectorFile, outputOpd, $
;  outType=O,/compress)
;#inputs#
;\anArg{rawFringes}{array of flt }{output of midiRawFringeSpec}
;\anArg{opd}{array of flt }{opd output of midiRawFringeSpec}
;\anArg{detectorFile}{string}{file with IMAGING_DETECTOR data containing freq info}
;\anArg{outputOpd}{arr of flt}{1d array of opd values (m) for filtered data}
;\anArg{outType}{string}{'AMP' (default), 'COMPLEX', 'REAL'}
;#return#
;\anArg{-}{data cube}{array of flux vs OPD in scans}
;#end_function#
;********************************************************
;  check defaults
   if (N_ELEMENTS(outType) LE 0) then outType = 'AMP'
   if (N_ELEMENTS(CompressScans) LE 0) then CompressScans = 0
   outType =STRUPCASE(outType)
   rawSize = SIZE(rawFringes)
;  get wavelength information (actually spatial frequency)
   spFreq = oirGetWaveNo(detectorFile)
   gridData = [min(spFreq), (max(spFreq)-min(spFreq))/512.,512]
;  set up fourier transform 
   ii = -COMPLEX(0., 1)*spFreq ; -i 
   fourier = OBJ_NEW('fft1d', gridData)
   output = COMPLEXARR(gridData[2], rawSize[2], rawSize[3])
;    correct phase for instrumental OPD and FFT to delay domain
   for i=0, rawSize[2]-1 do for j=0, rawSize[3]-1 do begin
      deZero = rawFringes[*,i,j]
;     deZero = deZero - mean(deZero)
      if (NOT KEYWORD_SET(noShift)) then  deZero = $
         deZero * EXP(ii * opd[i,j] )
      output(*,i,j) = fourier->cfft(spFreq, deZero )
   endfor
   outputOpd = fourier->kGrid()
   OBJ_DESTROY, fourier
   if (CompressScans) then output=total(output,2)/rawSize[2]
   if (outType EQ 'COMPLEX') then RETURN, output
   if (outType EQ 'REAL') then RETURN, FLOAT(output)
   RETURN, ABS(output)
END

FUNCTION midiGroupOpd, dataFile, maskFile, freqFile, outOpd, xsmooth
   if (N_ELEMENTS(xsmooth) LT 0) then xsmooth = 2.
   print,'Getting Data'
   raw   = midiRawFringeSpec(dataFile[0], maskfile, time, opd)
   print,'FFT to Delay space'
   group = midiFilterDisp(raw, opd, freqFile, outOpd, outType='COMPLEX')
   gSize = SIZE(raw,/dimensions)
   nRaw = N_ELEMENTS(gSize)
   n2 = 1
   for i=1,nRaw-1 do n2=n2*gSize[i]
   nOut = N_ELEMENTS(outOpd)
   group = TRANSPOSE(REFORM(group, nOut, n2),[1,0])
   print,'Smooth result'
   for i=0, nOut-1 do begin
;     group(*,i) = COMPLEX(fsmooth1(FLOAT(group(*,i)), xsmooth), $
;        fsmooth1(IMAGINARY(group(*,i)), xsmooth))
      group(*,i)= csmooth1(group(*,i), xsmooth)
   endfor
RETURN,ABS(group)
END

FUNCTION midiTrackGroupOpd, groupFringe, outOpd, range
   if (N_ELEMENTS(range) LE 0) then range = 100.
   dopd = outOpd[1] - outOpd[0]
   meanPeak = pk(findgen(N_ELEMENTS(outOpd)), total(groupFringe,1))
   gSize = SIZE(groupFringe,/dimensions)
   pkOpd = fltarr(gSize[0])
   r2 = (range * .5 / dopd)
;  out = outopd(meanpeak-r2:meanpeak+r2)
;  for i=0, gSize[0]-1 do pkOpd[i] = pk(out, $
;     groupFringe(i,meanpeak-r2:meanpeak+r2))
  for i=0, gSize[0]-1 do pkOpd[i] = pk(outopd,  groupFringe(i,*))
   
   mp = median(pkOpd)
   mpp = median(pkOpd,range)
   wp = where(abs(pkOpd-mpp) GT range)
   pkOpd(wp) = mpp(wp)
RETURN, pkOpd
END

FUNCTION pseudoComplex, x
   s = SIZE(x,/dim)
   xx = 2*lindgen(s[0]/2)
   rr = complex(x(xx,*),x(xx+1,*))
   rr = TRANSPOSE(rr)
RETURN,rr
END

FUNCTION midiGetKeyword, keyword, files
   nFiles = N_ELEMENTS(files) -1
   values = 0
   for i=0, nFiles do begin
      fn = STRSPLIT(files[i], ' ',/extract)
      f = OBJ_NEW('fitsfile', fn[0], iErr=iErr)
      if (iErr NE 0) then RETURN, values
      fh = f->head()
      if (i EQ 0) then values = fh->getpar(keyword) $
         else values = [values, fh->getpar(keyword)]
      OBJ_DESTROY,fh
      OBJ_DESTROY,f
   endfor
RETURN, values
END
