/*
 * This file is part of MIA.
 *
 * MIA is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/* $Id: oi_vis.h,v 1.3 2003/07/29 14:14:24 jaffe Exp $ */
/*******************************************************************************
* E.S.O. - VLT project
*
* "@(#) $Id: oi_vis.h,v 1.3 2003/07/29 14:14:24 jaffe Exp $"
*
* who       when      what
* --------  --------  ----------------------------------------------
* jaffe 2003-07-26 imported
*/

/************************************************************************
 *
 *----------------------------------------------------------------------
 */


#ifndef OI_VIS_H
#define OI_VIS_H
/*----------------------------------------------------------------------- */
/* Visibility data FITS binary tables routines                            */
/* Partially/fully calibrated data table for optical/IR interferometer    */
/*----------------------------------------------------------------------- */
#include "VLTItable.h"

typedef struct      /* table definition for visibility table */
{
  table_def *def;      /* basic table information */
  long   row;          /* last row (0-rel) read or written, -1=> none. */
  int    nWave;        /* number of points (=#wavelengths) per visibility */
  char  *date_obs;     /* UTC start date of obs */
  char  *arrname;      /* array name */
  char  *insname;      /* instrument/disperser name */
  int    target_col;   /* target number column */
  int    time_col;   /* time column */
  int    mjd_col;      /* date column */
  int    intTime_col;  /* integration time column */
  int    visamp_col;   /* visibility amplitude column number */
  int    visamperr_col;/* visibility amplitude error column number */
  int    visphi_col;   /* visibility phase column number */
  int    visphierr_col;/* visibility phase error column number */
  int    u_col;         /* u column number */
  int    v_col;         /* v column number */
  int    station_col;   /* station index column number */
  int    flag_col;      /* flag column number */
} vis_table_def;

typedef struct      /* structure used for visibility row */
{
  short   targetId;  /* pointer to target table */
  double  time;      /* time of obs (seconds) */
  double  mjd;       /* date of obs (days) */
  float   intTime;   /* integration time (sec)*/
  float  *visAmp;    /* visibility amplitude vector*/
  float  *visAmpErr; /* visibility amplitude error vector*/
  float  *visPhi;    /* visibility phase vector*/
  float  *visPhiErr; /* visibility phase error vector*/
  double  ucoord;    /* U coordinate (meters)*/
  double  vcoord;    /* U coordinate (meters)*/
  short   station[2];/* station indices */
  short  *flag;      /* flag array */
  table_row *raw;    /* generic table row */
} vis_table_row;

/* public functions */

/* vis_table_row functions */
vis_table_row* make_vis_row(vis_table_def* table);
/*----------------------------------------------------------------------- */
/*   Make an vis_table_row structure                                   */
/*   Inputs:                                                              */
/*      table     pointer to vis table structure                 */
/*   Returns:                                                             */
/*      An vis_table_row structure pointer ; NULL indicates failure    */
/*      Data pointers set to generic structure values                     */
/*----------------------------------------------------------------------- */

void kill_vis_row (vis_table_row *row);
/*----------------------------------------------------------------------- */
/*   Delete an vis_table_row structure                                   */
/*   Inputs:                                                              */
/*      row    pointer to the vis_table_row structure.                   */
/*----------------------------------------------------------------------- */

/* vis_table functions */
vis_table_def* 
make_vis_table_def(fitsfile *fptr, int nWave, char *date_obs, char* arrname,
   char *insname, int *ierr);
/*----------------------------------------------------------------------- */
/*   Make a new vis_table in an open FITS file                     */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*      nWave     number of points (=wavelength channels) per visibility  */
/*      date_obs  date of observation                                     */
/*      arrname   array name                                              */
/*      insname   instrument/disperser name                               */
/*   Output:                                                              */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*   Returns:                                                             */
/*      An vis_table_def pointer; NULL indicates failure           */
/*----------------------------------------------------------------------- */

vis_table_def* 
open_vis_table(fitsfile *fptr, int *ierr);
/*----------------------------------------------------------------------- */
/*   Open an existing vis_table in the active FITS file            */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*   Output:                                                              */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*   Returns:                                                             */
/*      An vis_table_def pointer; NULL indicates failure           */
/*----------------------------------------------------------------------- */

int read_vis_table_row(fitsfile *fptr,  vis_table_def* table, 
			    vis_table_row *row, long row_number,
			    int *ierr);
/*----------------------------------------------------------------------- */
/*   Read the next row from an vis_table                           */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*      table     table definitions structure from open or create         */
/*      row_number which row (1-rel) in the table                         */   
/*                 -1 => next row                                         */  
/*   Input/output:                                                        */
/*      row       pointer to an extant row structure                      */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*   Returns:                                                             */
/*      0 if successful, else failed                                      */
/*----------------------------------------------------------------------- */

int write_vis_table_row(fitsfile *fptr, vis_table_def* table, 
			     vis_table_row *row,  long row_number,
			     int *ierr);
/*----------------------------------------------------------------------- */
/*   Write  the next row to an vis_table                         */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*      table     table definitions structure from open or create         */
/*      row       pointer to an extant row structure                      */
/*      row_number which row (1-rel) in the table                         */     
/*                 -1 => next row                                         */  
/*   Input/output:                                                        */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*   Returns:                                                             */
/*      0 if successful, else failed                                      */
/*----------------------------------------------------------------------- */

int close_vis_table(fitsfile *fptr, vis_table_def* table, 
			     int *ierr);
/*----------------------------------------------------------------------- */
/*   Close an open vis_table, associated structures are destroyed  */
/*   Note: this closes the full FITS file                                 */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*   Returns:                                                             */
/*      0 if successful, else failed                                      */
/*----------------------------------------------------------------------- */

void kill_vis_table_def (vis_table_def* table);
/*----------------------------------------------------------------------- */
/*   Delete an vis_table structure                                 */
/*   Inputs:                                                              */
/*      table     table definitions structure from open or create         */
/*----------------------------------------------------------------------- */

#endif /* end OI_VIS_H */
/* */
