/*
 * This file is part of MIA.
 *
 * MIA is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/* $Id: ABCD.c,v 1.4 2002/06/04 09:29:29 jdejong Exp $ */
#define _POSIX_SOURCE_1
#include "vltPort.h"

/*-----------------------------------------------------------------------*/
/* Generates simulated midi-like data and does ABCD fringe tracking      */
/* Simulates two spectra (two regions) with phase different by 180 deg.  */ 
/*  Simulator for Midi raw data/BABCD fringe fitter                      */
/*  1 mJy -> amplitude = 3 as per CL memo                                */ 
/* N=5, (390 mJy) = 1.0E4 electrons per msec to detector                 */
/*  of this 30% goes to photometric channel and FRACT% to each spectrum  */
/* for N=5 use amp = 250000 for 25 msec integration                      */
/*  Command line parameters:                                             */
/*   -out   Name of FITS file to contain the data ,def ABCD.fit          */
/*   -amp   Amplitude of correlated signal in mJy  def 390               */
/*   -rms   RMS OPD noise in microns, def 0                              */
/*   -delay Delay to add (microns)  def 0                                */
/*   -rate  Delay rate to add (microns/sec)  def. 0                      */
/*   -n     Number of frames, def 100                                    */
/*   -integ Integration time in msec, def 25                             */
/*   -nstep number of OPD delay steps, def 4                             */
/*   -step  OPD delay step in microns, def 2.5                           */
/*   -OPDfile  file containing OPD offsets                               */
/*   -tau0  Desired atmospheric coherence time (sec) def no change       */
/*Processing parameters are written in HISTORY records in the FITS header*/
/*-----------------------------------------------------------------------*/
#include "simulate.h"

/* global variables, possibly read from command line */
char     *outfile=NULL;     /* output FITS file name */
long     number=100;        /* number of samples */
int      integ=25;         /* integration time in msec */
long     amplitude = 3900;  /* Amplitude of correlated signalin mJy */
float    delay = 0.0;       /* Delay to add (microns) */
float    rate = 10.0;       /* Delay rate to add (microns/sec) */
float    rmsOPD = 0.0;      /* RMS OPD noise */
int      nstep = 4;         /* number of OPD delay steps */
float    step = 2.5;        /*  OPD delay step in microns*/
char     *OPDfile=NULL;     /*  OPD error file */
double   tau0 = -1.0;       /* Desired atm. coherence time (sec) */

/* internal prototypes */
void ABCDin (int argc, char **argv, int *ierr);
void Usage(void);
void main_loop (fitsfile *fptr, imaging_data_table_def* tabledef, 
		       int *ierr);
void history (fitsfile *fptr, int *ierr);
float ABCD_fit(imaging_data_row *row[100], int nf, int nstep, float *OPDfit,
	      float *OPDtotal);

int main ( int argc, char **argv )
/*----------------------------------------------------------------------- */
/*   Program to write a FITS binary table                                 */
/*----------------------------------------------------------------------- */
{
  int  ierr, nx, ny, nreg;
  imaging_data_table_def* tabledef;
  fitsfile *outfptr=NULL;

  /* Startup - parse command line */
  ierr = 0;
  ABCDin (argc, argv, &ierr);
  if (ierr!=0) return ierr;

  /* Open/create basic FITS output file*/
  fits_create_file(&outfptr, outfile, &ierr);
  if (ierr>0) { /* error check */
    fprintf(stderr,"ERROR %d opening output FITS file %s \n", ierr, outfile);
    table_print_error_stack();  /* Show any cfitsio error messages */
    return ierr;
  }/* end error test */

  /* Create basic table structures and write IMAGING_DETECTOR table */
  /* set basic paramaters */
  nx = NCHAN;
  ny = 1;
  nreg = 2;
  tabledef = make_data_table(outfptr, nx, ny, nreg, 4, &ierr);
  if ((tabledef==NULL) || (ierr>0)) { /* error test */
    fprintf (stderr,"error %d after table create\n",ierr);
    table_print_error_stack();  /* Show any cfitsio error messages */
    return ierr;
  } /* end error test */
  
  /* Simulate data/ fringe track */
  main_loop(outfptr, tabledef, &ierr);
  if (ierr>0) { /* error test */
    fprintf (stderr,"error %d after table write\n",ierr);
    table_print_error_stack();  /* Show any cfitsio error messages */
    return ierr;
  } /* end error test */
  
  /* delete table_def */
  kill_imaging_data_table_def(tabledef);
  
  /* Close FITS file - this will flush open table buffer */
  fits_close_file (outfptr, &ierr);

   /* Show any cfitsio error messages */
  table_print_error_stack(); 

  return ierr;
} /* end of main */

void ABCDin (int argc, char **argv, int *ierr)
/*----------------------------------------------------------------------- */
/*  Parse control info from command line                                  */
/*   Input:                                                               */
/*      argc   Number of arguments from command line                      */
/*      argv   Array of strings from command line                         */
/*   Output:                                                              */
/*      ierr       Error code: 0 => ok                                    */
/*   Global values                                                        */
/*      outfile;    output FITS file name                                 */
/*      number      number of samples                                     */
/*      integ       integration time in msec                              */
/*      amplitude   Amplitude of correlated signal in counts              */
/*      delay       Delay to add (microns)                                */
/*      rate        Delay rate to add (microns/sec)                       */
/*      nstep       Number of OPD modulation steps                        */
/*      step        Size in microns of OPD modulation steps               */
/*      OPDfile     OPD error file                                        */
/*      tau0        Desired atm. coherence time (sec)                     */
/*----------------------------------------------------------------------- */
{
  int ax;
  char *arg;

  /*debug    fprintf(stderr, "number of arguments = %d\n",argc);*/
  if (argc<=1) Usage(); /* must have arguments */
  /* parse command line */
  for (ax=1; ax<argc; ax++) {
    arg = argv[ax];
    /*debug    fprintf(stderr, "arg = %s\n",arg);*/
    if (strcmp(arg, "-out") == 0) {
      outfile = argv[++ax];
    } else if (strcmp(arg, "-integ") == 0){ /* integration time */
      sscanf (argv[++ax],"%d",&integ);
    } else if (strcmp(arg, "-n") == 0){    /* number of integrations */
      sscanf (argv[++ax],"%ld",&number);
    } else if (strcmp(arg, "-amp") == 0){  /* amplitude */
      sscanf (argv[++ax],"%ld",&amplitude);
    } else if (strcmp(arg, "-delay") == 0){/* delay */
      sscanf (argv[++ax],"%f",&delay);
    } else if (strcmp(arg, "-rate") == 0){ /* delay rate */
      sscanf (argv[++ax],"%f",&rate);
    } else if (strcmp(arg, "-rms") == 0){ /* OPD noise */
      sscanf (argv[++ax],"%f",&rmsOPD);
    } else if (strcmp(arg, "-nstep") == 0){ /* OPD noise */
      sscanf (argv[++ax],"%d",&nstep);
    } else if (strcmp(arg, "-step") == 0){ /* OPD noise */
      sscanf (argv[++ax],"%f",&step);
    } else if (strcmp(arg, "-OPDfile") == 0){ /* OPD model file */
      OPDfile = argv[++ax];
    } else if (strcmp(arg, "-tau0") == 0){ /* Tau_0 */
      sscanf (argv[++ax],"%lf",&tau0);
    } else { /* unknown argument */
      Usage();
    }
  }
  
  /* must specify file */
  /*  if (!outfile) Usage();*/
  if (!outfile) outfile = "!ABCD.fit";
  *ierr = 0;
} /* end ABCDin */

void Usage(void)
/*----------------------------------------------------------------------- */
/*   Tells about usage of program and bails out                           */
/*----------------------------------------------------------------------- */
{
    fprintf(stderr, "Usage: ABCD -out file -t integ -n number -amp amp -delay delay -rate rate -tau0 time\n");
    fprintf(stderr, "Simulated midi data/ ABCD fringe tracker\n");
    fprintf(stderr, "Arguments:\n");
    fprintf(stderr, "  -out   Name of FITS file to contain the data, def none\n");
    fprintf(stderr, "  -amp   Amplitude of correlated signal in mJy  def 390\n");
    fprintf(stderr, "  -delay Delay to add (microns)  def 0\n");
    fprintf(stderr, "  -rate  Delay rate to add (microns/sec)  def. 0\n");
    fprintf(stderr, "  -rms   RMS OPD noise in microns, def 0\n");
    fprintf(stderr, "  -n     Number of frames, def 100\n");
    fprintf(stderr, "  -integ Integration time in msec, def 25\n");
    fprintf(stderr, "  -nstep number of OPD delay steps, def 4\n");
    fprintf(stderr, "  -step  OPD delay step in microns, def 2.5\n");
    fprintf(stderr, "  -OPDfile  OPD model file\n");
    fprintf(stderr, "  -tau0  Desired atmospheric coher. time (sec) def no change \n");
    
    /*/exit(1);  bail out */
  }/* end Usage */

void main_loop (fitsfile *fptr, imaging_data_table_def* tabledef, 
		       int *ierr)
/*----------------------------------------------------------------------- */
/*   Loop simulating, fitting data                                        */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*      tabledef  imaging_data table definition structure                 */
/*   Output:                                                              */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*----------------------------------------------------------------------- */
{
  long i, j, irow, ireg, nreg, istep;
  imaging_data_row *row[100]; 
  float step_sec, *bg, *gain, OPD_off = 0.0, OPD_fit=-100000.0, OPDtotal;
  float amp;
  FILE *hFile = NULL;
  int nf;

  /* if using OPD_model */
  if (OPDfile!=NULL) hFile = OPDfile_open(OPDfile, tau0);

  if (*ierr>0) return; /* existing error condition? */
  if (tabledef==NULL) {*ierr=10000; return;} /* check pointer */

  /* initialize randum number generator */
  amp = GaussNoise (-1.0);

  /* frequency dependent gain factors 1% fluctuations */
  nf = tabledef->detector->naxis[0][0]; /* number of pixels in Freq */
  gain = (float*)malloc(nf*sizeof(float));
  for (i=0; i<nf; i++) gain[i] = 1.0 + GaussNoise(0.01);

  /* frequency dependent 30% background ramp*/
  bg =  (float*)malloc(nf*sizeof(float));
  for (i=0; i<nf; i++) 
    bg[i] = gain[i] * (1.0 + ((0.3 * (i-nf/2)) / (float)nf));

  /* number of regions */
  nreg = tabledef->nregion;       

  /* make row structures for ABCD */
  for (i=0; i<nstep; i++)
    row[i] = make_imaging_data_row (tabledef);

  /* modulation step size in seconds */
   step_sec = 1.0e-6 * step / CLIGHT;

   /* init row strructures */
   for (j=0; j<nstep; j++) {

  /* set correlations types */
    for (i=0; i<nreg; i++)
      row[j]->targtype[i] = 2; /* data type */
    
    /* set constant values */
    row[j]->exp = 0.001 * integ; /* exposure in seconds */
    row[j]->offset[0] = 0.0;
    row[j]->offset[1] = 0.0;
    row[j]->rotation = 0.0;
    row[j]->reference = 0;
    row[j]->opd = 0.0;
    for (ireg=0; ireg<row[j]->nregion; ireg++) {
      row[j]->targtype[ireg] = 1;
      row[j]->source[ireg] = 1;
    }
   } /* end loop over step */

  /* looping over rows*/
  istep = 0; /* OPD modulation step */
  for (irow = 1; irow<=number; irow++) {

    /* fill in data record */
    row[istep]->time = irow * (row[istep]->exp / 86400.0); /* time in days */
    row[istep]->step = istep + 1; 
    row[istep]->local_opd = istep * step_sec; /* add modulation delay */

    /* next OPD model if necessary */
    if (OPDfile!=NULL) 
      OPD_off = OPDfile_read (hFile, 86400.0*row[istep]->time);

    /* initialize OPD error to zero, correct for sequence */
    if (OPD_fit<-10000.0) {
      OPD_fit = OPD_off + nstep/2 * step;
      OPDtotal = OPD_fit;
    }
  
    /* rest of model */
    simulate_data (row[istep], OPD_off - OPDtotal,  
		   amplitude, integ, delay, rate, rmsOPD, 
		   gain, bg, tabledef->detector);
    
    /* write next row */
     write_imaging_data_table(fptr, tabledef, row[istep], -1, ierr);
     if (*ierr>0) { /* error test */
       fprintf(stderr,"ERROR %d writing IMAGING_DATA_TABLE\n", *ierr);
       return;
     }

      /* cycle done? */
     istep++;
     if (istep>=nstep) {
       amp = ABCD_fit (row, nf, nstep, &OPD_fit, &OPDtotal); /* fit, track fringe */
       OPD_fit += nstep/2 * step; /* correct for stepping sequence */
       istep = 0;
       /* print result */
       fprintf (stdout,"%d Amp2 = %f Delay = %f Total = %f Actual = %f error = %f\n",
	   irow-nstep/2,amp,OPD_fit,OPDtotal-5.0,OPD_off,OPD_off-OPDtotal+5.0);
     } /* end ABCD fit */ 

  } /* end loop over rows */
  
  /* delete row structures */
  for (i=0; i<nstep; i++)
    kill_imaging_data_row(row[i]);

  /* close OPD model file */
  if (OPDfile!=NULL) OPDfile_close(hFile);

} /* end main_loop */

void history (fitsfile *fptr, int *ierr)
/*----------------------------------------------------------------------- */
/*   Write model parameters in FITS history                               */
/*   Inputs                                                               */
/*   Also writes to stdout                                                */
/*      fptr      cfitsio pointer to open FITS file                       */
/*   Output:                                                              */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*----------------------------------------------------------------------- */
{
  char hiline[FLEN_CARD], ttime[FLEN_CARD];

  sprintf(hiline,"*** VLTI Midi data ABCDator ***");
  fits_write_history (fptr, hiline, ierr);
  fprintf(stdout, "%s in file %s\n", hiline,outfile);
  history_time(ttime); /* get current time/date */
  sprintf(hiline,"ABCD run %s", ttime);
  fits_write_history (fptr, hiline, ierr);
  fprintf(stdout, "%s in file %s\n", hiline, outfile);
  sprintf(hiline,"ABCD Amplitude = %ld",amplitude);
  fits_write_history (fptr, hiline, ierr);
  fprintf(stdout, "%s\n", hiline);
  sprintf(hiline,"ABCD Integration = %d msec",integ);
  fits_write_history (fptr, hiline, ierr);
  fprintf(stdout, "%s\n", hiline);
  sprintf(hiline,"ABCD Delay offset = %f micron",delay);
  fits_write_history (fptr, hiline, ierr);
  fprintf(stdout, "%s\n", hiline);
  sprintf(hiline,"ABCD Delay rate = %f micron/s",rate);
  fits_write_history (fptr, hiline, ierr);
  fprintf(stdout, "%s\n", hiline);
  sprintf(hiline,"ABCD RMS OPD noise = %f micron",rmsOPD);
  fits_write_history (fptr, hiline, ierr);
  fprintf(stdout, "%s\n", hiline);
  sprintf(hiline,"ABCD Readout noise = %ld",(long)RON);
  fits_write_history (fptr, hiline, ierr);
  fprintf(stdout, "%s\n", hiline);
  sprintf(hiline,"ABCD Sky brightness per ms = %ld",(long)SKY);
  fits_write_history (fptr, hiline, ierr);
  fprintf(stdout, "%s\n", hiline);
  fprintf(stdout, "ABCD Number of samples is %ld\n",number);
  sprintf(hiline,"ABCD number of OPD steps = %d",nstep);
  fits_write_history (fptr, hiline, ierr);
  fprintf(stdout, "%s\n", hiline);
  sprintf(hiline,"ABCD OPD step size (microns) = %f",step);
  fits_write_history (fptr, hiline, ierr);
  fprintf(stdout, "%s\n", hiline);
  if (OPDfile) {
    sprintf(hiline,"ABCD OPD model file = %s",OPDfile);
    fits_write_history (fptr, hiline, ierr);
    fprintf(stdout, "%s\n", hiline);
     sprintf(hiline,"ABCD Tau_0 = %lf",tau0);
    fits_write_history (fptr, hiline, ierr);
    fprintf(stdout, "%s\n", hiline);
  }
} /* end history */
float ABCD_fit(imaging_data_row *row[100], int nf, int nstep, float *OPDfit,
	      float *OPDtotal)
/*----------------------------------------------------------------------- */
/*   Do ABCD fringe fitting                                               */
/*   Inputs:                                                              */
/*      row     array of imaging_data row structures                      */
/*      nf      number of frequencies                                     */
/*      nstep   number of modulation steps                                */
/*   In/Output:                                                           */
/*      OPDtotal  total of the delay in microns                           */
/*   Output:                                                              */
/*      OPDfit  estimate of the delay offset in microns                   */
/*   Returns:                                                             */
/*      The amplitude squared.                                             */
/*----------------------------------------------------------------------- */
{
  float ABCD[100], Q, U, sum, amp2, phase, fract_corr = 1.0;
  int istep, ifreq;

  /* sum over frequencies for each OPD step*/
  for (istep=0; istep<nstep; istep++) {
    ABCD[istep] = 0.0;
    for (ifreq=0; ifreq<nf; ifreq++) {
      ABCD[istep] += 
      /* debug 
	row[istep]->data[0][ifreq];
      end debug */
	fract_corr * (row[istep]->data[0][ifreq] - row[istep]->data[1][ifreq]);
    } /* end loop over frequency */
  } /* end loop over step */

  if (nstep==4) { /* simple ABCD */
    sum = abs(ABCD[0]) + abs(ABCD[1]) + abs(ABCD[2]) + abs(ABCD[3]);
    Q = ABCD[0] - ABCD[2];
    U = ABCD[1] - ABCD[3];
    amp2 = 2.0 * (Q*Q + U*U) / (sum*sum);
    phase = atan2(U, Q) + PI; /* correct for offset */
    if (phase>PI)  phase = phase - TWOPI;
    if (phase<-PI) phase = phase + TWOPI;
    /* delay offset */
    *OPDfit = -WAVELENGTH * phase / TWOPI;
  } else { /* unknown */
    *OPDfit = 0.0;
  }

  *OPDtotal += *OPDfit; /* accumulate */

  return amp2;
} /* end ABCD_fit */
/* */
