/*
 * This file is part of MIA.
 *
 * MIA is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/* $Id: FLUORfit.c,v 1.4 2002/06/04 09:29:29 jdejong Exp $ */
#define _POSIX_SOURCE_1
#include "vltPort.h"

/*-----------------------------------------------------------------------*/
/* Reads FLUOR data from a FITS binary table                             */
/* Expects data in form written by FLUOR2FITS                            */
/*  Command line parameters:                                             */
/*   -in    Name of FITS file containing the data ,def FLUOR.fit         */
/*   -out   Name of FITS output file with solutions, none=>no output     */
/*   -pix   Name of FITS image cube of delay-rate, none=>DRpix.fit       */
/*   -SI    Solution interval of fit in samples  (def. 1)                */
/*   -photo 1=make photometric corrections, 0=>not (def. 1)              */
/*-----------------------------------------------------------------------*/
#include <math.h>
#include <malloc.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "imaging_data.h"
#include "cal_table.h"
#include "interpol.h"
#include "history.h"
#include "FOfit.h"
#include "FLUORfsub.h"

/* global variables, possibly read from command line */
char     *infile="FLUOR.fit";       /* input FITS file name */
char     *pixfile="!FLpix.fit";     /* delay-rate image file */
char     *outfile="!FLUORfit.fit";  /* output FITS file name */
int      SI = 1;                    /* Solution interval in samples. */
int      photo = 1;                 /* make photometric corrections */

/* cfitsio pointer for input file */
fitsfile *inputfptr;

#define CLIGHT 2.997924562e8 /* speed of light */
#define TWOPI 6.283185308    /* two pi */

/* internal prototypes */
void FLUORin (int argc, char **argv, int *ierr);
void Usage(void);
void read_data_table (fitsfile *fptr, imaging_data_table_def* tabledef, 
		       int *ierr);
fitsfile* init_caltable (imaging_data_table_def* tabledef,
			 cal_table_def **table, 
			 cal_table_row *cal_rows[2], int *ierr);
void write_caltable(fitsfile *outfptr, cal_table_def* table,
		    double time, float timei, 
		    cal_table_row *cal_rows[2], int *ierr);
void close_caltable(fitsfile *outfptr, cal_table_def* table,
		    cal_table_row *cal_rows[2], int *ierr);
void history (fitsfile *fptr, int *ierr);
int main ( int argc, char **argv )
/*----------------------------------------------------------------------- */
/*   Program to fringe fit FLUOR data                                     */
/*----------------------------------------------------------------------- */
{
  int  ierr;
  imaging_data_table_def* tabledef;
  fitsfile *infptr=NULL;

  /* Startup - parse command line */
  ierr = 0;
  FLUORin (argc, argv, &ierr);
  if (ierr!=0) return ierr;

  /* Open FITS input file */
  fits_open_file(&infptr, infile, READONLY, &ierr);
  if (ierr>0) { /* error check */
    fprintf(stderr,"ERROR %d opening input FITS file %s \n", &ierr, infile);
    table_print_error_stack();  /* Show any cfitsio error messages */
    return ierr;
  }/* end error test */

  /* Get basic table structures and read IMAGING_DETECTOR table */
  tabledef = open_imaging_data_table(infptr, &ierr);
  if ((tabledef==NULL) || (ierr>0)) { /* error test */
    fprintf (stderr,"error %d after table create\n",ierr);
    table_print_error_stack();  /* Show any cfitsio error messages */
    return ierr;
  } /* end error test */
  
  /* Read table  fitting data*/
  read_data_table(infptr, tabledef, &ierr);
  if (ierr>0) { /* error test */
    fprintf (stderr,"error %d after table read\n",ierr);
    table_print_error_stack();  /* Show any cfitsio error messages */
    return ierr;
  } /* end error test */
  
  /* Close FITS file */
  fits_close_file (infptr, &ierr);

   /* Show any cfitsio error messages */
  table_print_error_stack(); 

  return ierr;
} /* end of main */

void FLUORin (int argc, char **argv, int *ierr)
/*----------------------------------------------------------------------- */
/*  Parse control info from command line                                  */
/*   Input:                                                               */
/*      argc   Number of arguments from command line                      */
/*      argv   Array of strings from command line                         */
/*   Output:                                                              */
/*      ierr       Error code: 0 => ok                                    */
/*   Global values                                                        */
/*      infile;     input FITS file name                                  */
/*      SI          Solution interval in scans.                           */
/*----------------------------------------------------------------------- */
{
  int ax;
  char *arg;

  /*debug    fprintf(stderr, "number of arguments = %d\n",argc);*/
  if (argc<=1) Usage(); /* must have arguments */
  /* parse command line */
  for (ax=1; ax<argc; ax++) {
    arg = argv[ax];
    /*debug    fprintf(stderr, "arg = %s\n",arg);*/
    if (strcmp(arg, "-in") == 0) {
      infile = argv[++ax];
    } else if (strcmp(arg, "-out") == 0) {
      outfile = argv[++ax];
    } else if (strcmp(arg, "-pix") == 0) {
      pixfile = argv[++ax];
    } else if (strcmp(arg, "-SI") == 0){    /* solution interval */
      sscanf (argv[++ax],"%d",&SI);
    } else if (strcmp(arg, "-photo") == 0){    /* photometric corr? */
      sscanf (argv[++ax],"%d",&photo);
    } else { /* unknown argument */
      Usage();
    }
  }
  *ierr = 0;
} /* end FLUORin */

void Usage(void)
/*----------------------------------------------------------------------- */
/*   Tells about usage of program and bails out                           */
/*----------------------------------------------------------------------- */
{
    fprintf(stderr, "Usage: FLUORfit -in file -out file -photo n\n");
    fprintf(stderr, "Fringe fit FLUOR data\n");
    fprintf(stderr, "Arguments:\n");
    fprintf(stderr, "  -in   Name of FITS file to contain the data, def FLUOR.fit\n");
    fprintf(stderr, "  -out  Name of FITS output file with solutions, none=>no output\n");
    fprintf(stderr, "  -photo  1=>make photom. corr, 0=>not\n");
    /*    fprintf(stderr, "  -pix  Name of FITS image cube of delay-rate, none=>no output\n");*/
    /*    fprintf(stderr, "  -SI   Solution interval in scans (def. 2)\n");*/
     /*/exit(1);  bail out */
  }/* end Usage */

void read_data_table (fitsfile *fptr, imaging_data_table_def* tabledef, 
		       int *ierr)
/*----------------------------------------------------------------------- */
/*   Loop reading table, fitting fringes, writing solutions if requested  */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*      tabledef  imaging_data table definition structure                 */
/*   Output:                                                              */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*----------------------------------------------------------------------- */
{
  long i, j, irow, nreg, step, nx, integ, maxstep, iscan, tcount;
  double  time, taudif;
  double sumtau=0.0, sumd=0.0, sumd2=0.0, cnttau=0.0, cntd2=0.0;
  int done=0, FOinit=0;
  imaging_data_row *row; 
  float *OPD_off=NULL, *times=NULL, timei, dOPD, fract_corr=-1.0;
  cal_table_def  *caltable=NULL;
  cal_table_row  *cal_rows[2] = {NULL, NULL};
  fitsfile *outfptr = NULL;
  float *IntChan=NULL;
  /* which columns are the interferometric data */
  int i1col=0, i2col=1;
  long input_last, input_nrows;

  if (*ierr>0) return; /* existing error condition? */
  if (tabledef==NULL) {*ierr=10000; return;} /* check pointer */

  inputfptr = fptr; /* save cfitsio pointer in global variable */

  /* allocate storage array */
  maxstep = tabledef->maxstep;
  IntChan = (float*)malloc(2*maxstep*sizeof(float));
  OPD_off = (float*) malloc(maxstep*sizeof(float));
  times = (float*) malloc(maxstep*sizeof(float));
  for (i=0; i<2*maxstep; i++) IntChan[i] = 0.0;
  for (i=0; i<maxstep; i++) OPD_off[i] = 0.0;
  for (i=0; i<maxstep; i++) times[i] = 0.0;
  /*fprintf (stdout,"maxstep = %ld \n",maxstep);*/

  /* set up output if necessary */
  outfptr = init_caltable(tabledef, &caltable, cal_rows, ierr);
  if (*ierr>0) { /* error test */
    fprintf(stderr,"ERROR %d writing CALIBRATION_TABLE\n", *ierr);
    return;
  }

  nx  = tabledef->detector->naxis[0][0]; /* number of pixels in Freq */
  if (nx>1) { /* error test */
    fprintf(stderr,"ERROR: too many (%d) frequency channels\n", nx);
    return;
  }

  nreg = tabledef->nregion;        /* number of regions */

  row = make_imaging_data_row (tabledef);

  /* get size of table */
  fits_get_num_rows (fptr, &input_nrows, ierr);

  /* read first row */
  input_last = 1;
  read_imaging_data_table(fptr, tabledef, row, input_last, ierr);
  if (*ierr>0) { /* error test */
    fprintf(stderr,"ERROR %d reading IMAGING_DATA_TABLE\n", *ierr);
    return;
  }

  /* initialize fringe fitter. */
  if (!FOinit) {
    dOPD = 0.001;
    setup_FLUORfringe_fit(pixfile, fptr, maxstep, 300, photo, dOPD, 
			  tabledef->detector, input_nrows, input_last);
    FOinit = 1;
    }
	
  /* looping over scans */
  while (!done) {
  
    /* read next scan */
    get_FLUOR_scan(fptr, tabledef, row, maxstep, OPD_off, times, IntChan,
		   &integ, ierr);
    if (*ierr>0) { /* error test */
      fprintf(stderr,"ERROR %d reading IMAGING_DATA_TABLE\n", *ierr);
      return;
    }
    if (*ierr==-1) { /* finished? */
      done = 1;
      *ierr = 0;
    }

    if (done) break; /* no more data? */

    /* average time */
    time = 0.0; timei = 0.0; tcount = 0;
    for (i=0; i<integ; i++) {
      time += times[i];
      tcount++;
    }
    if (tcount>0) time = time /  tcount; /* average time */
    timei = row->exp * tcount;

    /* get times in seconds relative to reference */
    for (i=0; i<integ; i++) times[i] = (times[i] - time) * 86400.0;

    /* fit data */
    cal_rows[1]->time = time;
    FLUOR_fringe_fit_data (IntChan, integ, iscan, row->exp, OPD_off,
			   times, tabledef->detector, cal_rows[1]);
    
    /* write solution  */
    write_caltable(outfptr, caltable, time, timei, cal_rows, ierr);
    if (*ierr>0) { /* error test */
      fprintf(stderr,"ERROR %d writing CALIBRATION_TABLE\n", *ierr);
      return;
    }
  } /* end loop over scans */
  
  /* shutdown output caltable if necessary */
  close_caltable(outfptr, caltable, cal_rows, ierr);
  if (*ierr>0) { /* error test */
    fprintf(stderr,"ERROR %d closing CALIBRATION_TABLE\n", *ierr);
    return;
  }

  /* shutdown output delay-rate fitting process */
  shutdown_FLUORfringe_fit(ierr);
  if (*ierr>0) { /* error test */
    fprintf(stderr,"ERROR %d closing delay-rate process\n", *ierr);
    return;
  }

  /* delete row structure */
  kill_imaging_data_row(row);

  /* deallocate work arrays */
  if (IntChan) free (IntChan);
  if (OPD_off) free (OPD_off);
  if (times)   free (times);

} /* end read_data_table */

fitsfile* init_caltable (imaging_data_table_def* tabledef,
			 cal_table_def **table, cal_table_row *cal_rows[2], 
			  int *ierr)
/*----------------------------------------------------------------------- */
/*   If global variable outfile is set then it is used as the output FITS */
/*   file name, otherwise no file is written.                             */
/*      If output is desired, the file is created/opened and the          */
/*   cal_table row structures are initialized                             */
/*   Copies old, writes new history                                       */
/*   Inputs:                                                              */
/*      tabledef    imaging_data table definition                         */
/*   Outputs:                                                             */
/*      cal_rows    cal_table structures initialized for telescopes       */
/*                  1 and 2 with1 as the reference telescope              */
/*      table       table_def for I/O                                     */
/*      ierr        Error code: 0 => ok                                   */
/*   Returns                                                              */
/*      cfitsio file pointer, NULL indicated no file to be written.       */
/*----------------------------------------------------------------------- */
{
  int i;
  fitsfile *out=NULL;

  cal_rows[0] = NULL;
  cal_rows[1] = NULL;
  *table = NULL;

  /* don't write in no output file specified */
  if (outfile==NULL) return out;

  /* Open/create basic FITS output file*/
  fits_create_file(&out, outfile, ierr);
  if (*ierr>0) { /* error check */
    fprintf(stderr,"ERROR %d opening output FITS file %s \n", *ierr, outfile);
    table_print_error_stack();  /* Show any cfitsio error messages */
    return NULL;
  }/* end error test */

  /* create table_def for table */
  *table = make_cal_table_def(out, tabledef->date_obs, 0, ierr);
  if ((*ierr>0) || (*table==NULL)) return NULL;

  fits_write_key(out, TSTRING, "OBJECT", (void*)("MIDI simulation"),
		 "Fringe fitted simulated data", ierr);
  fits_write_key(out, TSTRING, "ORIGIN", (void*)("midi fringe"),
		 "Creator of FITS file", ierr);

  /* copy history */
  copy_history (inputfptr, out, ierr);

  /* add parameters in history */
  history(out, ierr);

  /* make table row objects */
  for (i=0; i<2; i++) {
    cal_rows[i] = make_cal_table_row(*table);
    cal_rows[i]->source = 0;
    cal_rows[i]->tele = i+1;
    cal_rows[i]->ref = 1;
    cal_rows[i]->weight = 10.0;
  }
  return out;
} /* end init_caltable */

void write_caltable(fitsfile *outfptr, cal_table_def* table, 
		    double time, float timei, 
		    cal_table_row *cal_rows[2], int *ierr)
/*----------------------------------------------------------------------- */
/*   Write calibration table entries to open FITS file                    */
/*   Inputs:                                                              */
/*      outfptr     cfitsio file pointer for output FITS file             */
/*      table       table_def for I/O                                     */
/*      time        reference time in days                                */
/*      timei       time span for solution                                */
/*      cal_rows    cal_table structures to be written                    */
/*   Outputs:                                                             */
/*      ierr        Error code: 0 => ok                                   */
/*----------------------------------------------------------------------- */
{
  /* are we writing output? */
  if (outfptr==NULL) return; /* nope */

  /* save time info */
  cal_rows[0]->time = time;
  cal_rows[1]->time = time;
  cal_rows[0]->timei = timei;
  cal_rows[1]->timei = timei;

  /* write 'em */
  write_cal_table(outfptr, table, cal_rows[0], (long)-1, ierr);
  write_cal_table(outfptr, table, cal_rows[1], (long)-1, ierr);
} /* end write_caltable */

void close_caltable(fitsfile *outfptr, cal_table_def* table,
		    cal_table_row *cal_rows[2], int *ierr)
/*----------------------------------------------------------------------- */
/*   Close calibration table in FITS file (closes file)                   */
/*   Also does other cleanup                                              */
/*   Inputs:                                                              */
/*      outfptr     cfitsio file pointer for output FITS file             */
/*      table       table_def for I/O                                     */
/*      cal_rows    cal_table structures to be deleted                    */
/*   Outputs:                                                             */
/*      ierr        Error code: 0 => ok                                   */
/*----------------------------------------------------------------------- */
{
  /* are we writing output? */
  if (outfptr==NULL) return; /* nope */

  /* Close FITS file - this will flush open table buffer */
   close_cal_table(outfptr, table, ierr);

  /* delete structures */
  if (cal_rows!=NULL) {
    kill_cal_table_row (cal_rows[0]);
    kill_cal_table_row (cal_rows[1]);
  }

} /* end close_caltable */

void history (fitsfile *fptr, int *ierr)
/*----------------------------------------------------------------------- */
/*   Write model parameters in FITS history                               */
/*   Inputs                                                               */
/*   Also writes to stdout                                                */
/*      fptr      cfitsio pointer to open FITS file                       */
/*   Output:                                                              */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*----------------------------------------------------------------------- */
{
  char hiline[FLEN_CARD], ttime[FLEN_CARD];
  sprintf(hiline,"*** VLTI Midi testbed fringe fitter ***");
  fits_write_history (fptr, hiline, ierr);
  fprintf(stdout, "%s in file %s\n", hiline, outfile);
  history_time(ttime); /* get current time/date */
  sprintf(hiline,"FLUORfit run %s", ttime);
  fits_write_history (fptr, hiline, ierr);
  fprintf(stdout, "%s in file %s\n", hiline, outfile);
  sprintf(hiline,"calib input data file = %s",infile);
  sprintf(hiline,"FLUORfit input file = %s",infile);
  fits_write_history (fptr, hiline, ierr);
  fprintf(stdout, "%s\n", hiline);
} /* end history */

/* */
