/*
 * This file is part of MIA.
 *
 * MIA is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/* $Id: calib.c,v 1.5 2002/06/04 09:29:29 jdejong Exp $ */
#define _POSIX_SOURCE_1
#include "vltPort.h"

/*-----------------------------------------------------------------------*/
/* Reads simulated midi-like data from a FITS binary table and applies   */
/* Fringe fitting calibration                                            */
/*  Command line parameters:                                             */
/*   -in    Name of FITS file containing the data, def simul.fit         */
/*   -in2   FITS file containing the CALIBRATION table, def fringe.fit   */
/*   -out   Name of FITS output file with solutions, def calib.fit       */
/*   -CI    Calibration interval of fit in samples  (def. 10)            */
/*   -chAv  Number of channels to average (def. 10)            */
/*   -minSNR minimum fringe SNR for an acceptable solution (def. 5)      */
/*-----------------------------------------------------------------------*/
#include <math.h>
#include <malloc.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "imaging_data.h"
#include "cal_table.h"
#include "visibility.h"
#include "fourg.h"
#include "dft.h"
#include "pfit.h"
#include "interpol.h"
#include "history.h"

/* global variables, possibly read from command line */
char     *infile="simul.fit";   /* input FITS file with data */
char     *in2file="fringe.fit"; /* input FITS file with calibration */
char     *outfile="!calib.fit"; /* output FITS file name */
int      CI = 10;               /* Calibration interval in samples. */
int      chAv = 10;             /* Number of frequency channels to average */
float    minSNR = 5;            /* min. fringe SNR for an acceptable solution */

/* global parameters */
#define T_FILTER_WIDTH 0 /* width in pixels of time notch filter, 0=only zero */
#define F_FILTER_WIDTH 3 /* width in pixels of freq notch filter */
float  fblank=-1.2345e5;  /* magic value blanking */

/* calibration structures */
cal_table_row *cal_row = NULL;  /* current calibration record */
cal_table_row *cal_tmp = NULL;  /* calibration record to read */
long  ical=0;                   /* current cal table row number */
long  num_cal=0;                /* number of cal records */

/* work arrays */
static float FTwork1[4096], FTwork2[4096];

#define CLIGHT 2.997924562e8 /* speed of light */
#define TWOPI 6.283185308    /* two pi */
#define FRACT 0.35      /* Fraction of the photons that go into each cross correlation */
#define COHER 0.60      /* Instrumental coherence factor (poln mismatch)*/
#define COUNTS 2.564e4  /* counts per second per mJy */

/* internal prototypes */
void calibin (int argc, char **argv, int *ierr);
void Usage(void);
void calibrate_data_table (fitsfile *fptr, imaging_data_table_def* tabledef, 
		       int *ierr);
int calibr_data (float **spectrum, int nx, int nt, 
		 float *OPD_off, float *phase_corr, float *weight,
		 imaging_detector_table_def* detector,
		 visibility_row *vis_row);
void notch_filter (float **array, int nf, int nt);
void correct_delay (float **array, int nf, int nt, float *OPD_off, 
		    float *phase_corr, 
		    imaging_detector_table_def* detector);
int avg_in_time (float **array, int nf, int nt, float *weight, float *wt);
void avg_in_freq (float **array, float *weight, int nf, float *vis, float *wt);
void print_array (int type, float **array, int first_delay, int last_delay, 
		 int first_rate, int last_rate, char *title);
fitsfile* init_caltable (imaging_data_table_def* tabledef,
			 cal_table_def **table, int *ierr);
void cal_delay(fitsfile *outfptr, cal_table_def* table,
	       double time, float *calDelay, float *calPhase, int *ierr);
void close_caltable(fitsfile *outfptr, cal_table_def* table, int *ierr);
fitsfile* init_outdata (imaging_data_table_def* tabledef, 
			fitsfile *infptr, fitsfile *calfptr,
			visibility_table_def **table, int *ierr);
void write_outdata(fitsfile *outfptr, visibility_table_def* table,
		   visibility_row *visrow, imaging_data_row *row,
		   int *ierr);
void close_outdata(fitsfile *outfptr, visibility_table_def* table, 
		   int *ierr);
void history (fitsfile *fptr, int *ierr);


int main ( int argc, char **argv )
/*----------------------------------------------------------------------- */
/*   Program to fringe fit simulated midi data                            */
/*----------------------------------------------------------------------- */
{
  int  ierr;
  imaging_data_table_def* tabledef;
  fitsfile *infptr=NULL;

  /* Startup - parse command line */
  ierr = 0;
  calibin (argc, argv, &ierr);
  if (ierr!=0) return ierr;

  /* Open FITS input file */
  fits_open_file(&infptr, infile, READONLY, &ierr);
  if (ierr>0) { /* error check */
    fprintf(stderr,"ERROR %d opening input FITS file %s \n", ierr, infile);
    table_print_error_stack();  /* Show any cfitsio error messages */
    return ierr;
  }/* end error test */

  /* Get basic table structures and read IMAGING_DETECTOR table */
  tabledef = open_imaging_data_table(infptr, &ierr);
  if ((tabledef==NULL) || (ierr>0)) { /* error test */
    fprintf (stderr,"error %d after table create\n",ierr);
    table_print_error_stack();  /* Show any cfitsio error messages */
    return ierr;
  } /* end error test */
  
  /* Read table  calibrating data */
  calibrate_data_table(infptr, tabledef, &ierr);
  if (ierr>0) { /* error test */
    fprintf (stderr,"error %d after table read\n",ierr);
    table_print_error_stack();  /* Show any cfitsio error messages */
    return ierr;
  } /* end error test */
  
  /* delete table_def */
  kill_imaging_data_table_def(tabledef,1);
  
  /* Close FITS file */
  fits_close_file (infptr, &ierr);

   /* Show any cfitsio error messages */
  table_print_error_stack(); 

  return ierr;
} /* end of main */

void calibin (int argc, char **argv, int *ierr)
/*----------------------------------------------------------------------- */
/*  Parse control info from command line                                  */
/*   Input:                                                               */
/*      argc   Number of arguments from command line                      */
/*      argv   Array of strings from command line                         */
/*   Output:                                                              */
/*      ierr       Error code: 0 => ok                                    */
/*   Global values                                                        */
/*      infile;     input FITS file name                                  */
/*      CI          Calibration interval in sec.                             */
/*----------------------------------------------------------------------- */
{
  int ax;
  char *arg;

  /*debug    fprintf(stderr, "number of arguments = %d\n",argc);*/
  if (argc<=1) Usage(); /* must have arguments */
  /* parse command line */
  for (ax=1; ax<argc; ax++) {
    arg = argv[ax];
    /*debug    fprintf(stderr, "arg = %s\n",arg);*/
    if (strcmp(arg, "-in") == 0) {
      infile = argv[++ax];
    } else if (strcmp(arg, "-out") == 0) {
      outfile = argv[++ax];
    } else if (strcmp(arg, "-in2") == 0) {
      in2file = argv[++ax];
    } else if (strcmp(arg, "-CI") == 0){    /* calibration interval */
      sscanf (argv[++ax],"%d",&CI);
    } else if (strcmp(arg, "-chAv") == 0){    /* number of channels to average */
      sscanf (argv[++ax],"%d",&chAv);
    } else if (strcmp(arg, "-minSNR") == 0){ /* min SNR */
      sscanf (argv[++ax],"%f",&minSNR);
    } else { /* unknown argument */
      Usage();
    }
  }
  *ierr = 0;
} /* end calibin */

void Usage(void)
/*----------------------------------------------------------------------- */
/*   Tells about usage of program and bails out                           */
/*----------------------------------------------------------------------- */
{
    fprintf(stderr, "Usage: calib -in file -t integ -n number -amp amp -delay delay -rate rate\n");
    fprintf(stderr, "Simulated midi data as VLTI binary tables\n");
    fprintf(stderr, "Arguments:\n");
    fprintf(stderr, "  -in   Name of FITS file to contain the data, def simul.fit\n");
    fprintf(stderr, "  -in2  FITS file containing the CALIBRATION table, def fringe.fit\n");
    fprintf(stderr, "  -out  Name of FITS output file with solutions, def calib.fit\n");
    fprintf(stderr, "  -CI   Calibration interval in samples (def. 10)\n");
    fprintf(stderr, "  -chAv Number of channels to average (def. 10) \n");
    fprintf(stderr, "  -minSNR Min acceptable fringe SNR (def. 5)\n");
     /*/exit(1);  bail out */
  }/* end Usage */

void calibrate_data_table (fitsfile *fptr, imaging_data_table_def* tabledef, 
		       int *ierr)
/*----------------------------------------------------------------------- */
/*   Loop reading table, calibrating data, writing results                */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS input data file            */
/*      tabledef  imaging_data table definition structure                 */
/*   Output:                                                              */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*----------------------------------------------------------------------- */
{
  long i, indx, j, irow, nreg, step, number, nx, integ;
  int OK;
  double  time, taudif=0., sumd=0.0, sumd2=0.0, cntd2=0.0;
  imaging_data_row *row; 
  visibility_row *visrow; 
  float **spectrum, *OPD_off, *phase_corr, *weight, fract_corr=-1.0;
  float timei, calDelay, calPhase;
  cal_table_def        *caltable=NULL;
  visibility_table_def *vistable=NULL;
  fitsfile             *calfprt=NULL, *outfprt=NULL;

  if (*ierr>0) return; /* existing error condition? */
  if (tabledef==NULL) {*ierr=10000; return;} /* check pointer */

  /* set up calibration table */
  calfprt = init_caltable(tabledef, &caltable, ierr);
  if (*ierr>0) { /* error test */
    fprintf(stderr,"ERROR %d opening CALIBRATION_TABLE\n", *ierr);
    return;
  }

  /* set up output */
  outfprt = init_outdata(tabledef, fptr, calfprt, &vistable, ierr);
  if (*ierr>0) { /* error test */
    fprintf(stderr,"ERROR %d opening VISIBILITY_TABLE\n", *ierr);
    return;
  }
  visrow = make_visibility_row(vistable);

  nx  = tabledef->detector->naxis[0][0]; /* number of pixels in Freq */

  /* allocate 2D data array */
  spectrum = (float**) malloc(CI*sizeof(float*));
  for (i=0; i<CI; i++)
    spectrum[i] = (float*) malloc(2*nx*sizeof(float));

  OPD_off = (float*) malloc(CI*sizeof(float));    /* delay correction */
  phase_corr = (float*) malloc(CI*sizeof(float)); /* phase correction */
  weight = (float*) malloc(CI*sizeof(float));     /* data weight */

  nreg = tabledef->nregion;        /* number of regions */

  row = make_imaging_data_row (tabledef);

  /* get size of input data table */
  number = 0;
  fits_get_num_rows (fptr, &number, ierr);

  /* looping over rows*/
  integ = 0; indx = 0; time = 0.0; timei = 0.0;
  for (irow = 1; irow<=number; irow++) {
  
    /* read next row */
     read_imaging_data_table(fptr, tabledef, row, -1, ierr);
     if (*ierr>0) { /* error test */
       fprintf(stderr,"ERROR %d reading IMAGING_DATA_TABLE\n", *ierr);
       return;
     }

     /* correction for how the light is divided amoung spectra, etc */
     /* try to calibrate in mJy */
     if (fract_corr<0.0) 
       fract_corr = ((float)nx / (2.0 * FRACT * COHER)) /
	 (COUNTS * row->exp);
     
     /* get modulation and calibration delay */
     cal_delay(calfprt, caltable, row->time, &calDelay, &calPhase, ierr);
     if (*ierr>0) { /* error test */
       fprintf(stderr,"ERROR %d writing CALIBRATION_TABLE\n", *ierr);
       return;
     }
     if (calDelay!=fblank) {
       /* debug - use original */
       /*calDelay = row->opd/(1.0E6 * CLIGHT); calPhase = 0.0;*/

       OPD_off[integ]    = row->local_opd[0] + calDelay; 
       phase_corr[integ] = calPhase;
       weight[integ]     = 1.0;
       /* model delay passed by simul as row->opd (microns)*/
       taudif = row->opd[0] - (calDelay * 1.0E6 * CLIGHT);
       /* ignore really bad fits */
       if (abs(taudif)<15.0) {
	 sumd += taudif;
	 sumd2 += taudif*taudif;
	 cntd2 += 1.0;
       }

     } else {
       OPD_off[integ]    = row->local_opd[0];
       phase_corr[integ] = 0.0;
       weight[integ]     = 0.0;
      }

     /* copy data to work array */
     indx = 0;
     for (i=0; i<nx; i++){
       spectrum[integ][indx++] = 
	 fract_corr * (row->data[0][i] - row->data[1][i]); 
       spectrum[integ][indx++] = 0.0;             /* imaginary part */
     }
     integ++;

     /* sum time for average */
     time  += row->time;
     timei += row->exp;

     if (integ>=CI) { /* time to calibrate it? */
       /* calibrate data and average in time */
       OK = calibr_data (spectrum, nx, integ, OPD_off, phase_corr, 
			 weight, tabledef->detector, visrow);

       /* write result */
       if (OK) {
	 if (integ>0) time = time / integ;
	 visrow->time = time;
	 visrow->timei = timei;
	 write_outdata(outfprt, vistable, visrow, row, ierr);
	 if (*ierr>0) { /* error test */
	   fprintf(stderr,"ERROR %d writing VISIBILITY_TABLE\n", *ierr);
	   return;
	 }
       }
       time = 0.0; timei = 0.0;
       integ = 0; indx = 0;
     } /* end of do fit */
    
  } /* end loop over rows */
  
  /* calibrate any remaining data  */
  if (integ>0) {
    OK = calibr_data (spectrum, nx, integ, OPD_off, phase_corr, 
		      weight, tabledef->detector, visrow);
  /* write result */
    if (OK) {
      if (integ>0) time = time / integ;
      visrow->time = time;
      visrow->timei = timei;
      write_outdata(outfprt, vistable, visrow, row, ierr);
      if (*ierr>0) { /* error test */
	fprintf(stderr,"ERROR %d writing VISIBILITY_TABLE\n", *ierr);
	return;
      }
    }
  } /* end dealing with final data */

  /* shutdown output */
  close_outdata(outfprt, vistable, ierr);
  if (*ierr>0) { /* error test */
    fprintf(stderr,"ERROR %d closinging VISIBILITY_TABLE\n", *ierr);
    return;
  }

  /* Give RMS delay error */
  if (cntd2>0.0) {
    sumd = sumd / cntd2;
    sumd2  = sqrt (sumd2 / cntd2);
    fprintf (stdout, "RMS delay error = %f microns, avg = %f microns\n",
	     sumd2, sumd);
  } /* end compare model */

  /* shutdown reading calibrator table */
  close_caltable(calfprt, caltable, ierr);
  if (*ierr>0) { /* error test */
    fprintf(stderr,"ERROR %d closinging CALIBRATION_TABLE\n", *ierr);
    return;
  }

  /* delete row structures */
  kill_imaging_data_row(row);
  kill_visibility_row(visrow);

  /* delete temporary arrays */
  if (spectrum!=NULL) {
    for (i=0; i<CI; i++)
      if (spectrum[i]!=NULL)free(spectrum[i]);
    free(spectrum);
  }
} /* end read_data_table */

int calibr_data (float **spectrum, int nf, int nt, 
		 float *OPD_off, float *phase_corr, float *weight,
		 imaging_detector_table_def* detector, 
		 visibility_row *visrow)
/*----------------------------------------------------------------------- */
/*   Apply calibration to simulated telescope data                        */
/*   Steps:                                                               */
/*     1) apply 2D notch filter to suppress non source signals            */
/*        Also zero pad array to big for better resolution                */
/*     2) remove OPD from calibration                                     */
/*     3) average in time direction                                       */
/*   Inputs:                                                              */
/*      spectrum  Measured spectrum  (nf x nt) array of detector values   */
/*      nf        Number of frequency channels in spectrum                */
/*      nt        Number of times                                         */
/*      OPD_off   Calibration delay (seconds) per sample                  */
/*      phase_corr Calibration phase (rad) per sample                     */
/*      weight    data weight  per sample for average                     */
/*      detector  imaging_detector table definition structure             */
/*      visrow    Visibility record, filled in on output                  */
/*   Returns                                                              */
/*      1 if there is valid data in the average, else 0                   */
/*----------------------------------------------------------------------- */
{
  int OK;

  /* notch filter to kill most non source signals */
  notch_filter (spectrum, nf, nt);

  /* correct delays */
  correct_delay(spectrum, nf, nt, OPD_off, phase_corr, detector);
  
  /* average in time */
  OK =  avg_in_time (spectrum, nf, nt, weight, FTwork1);

  /* average in freq */
  if (OK)
    avg_in_freq (spectrum, FTwork1, nf, visrow->data, visrow->weight);

  return OK;

} /* end calibr_data */

void notch_filter (float **array, int nf, int nt)
/*----------------------------------------------------------------------- */
/*   Apply notch filter to a frequency x time array                       */
/*   Uses global parameters:                                              */
/*      T_FILTER_WIDTH  width in pixels of time notch filter, 0=only zero */
/*      F_FILTER_WIDTH  width in pixels of freq notch filter, 0=only zero */
/*   Inputs:                                                              */
/*      array     Measured spectrum  (nf x nt) array of detector values   */
/*      nf        Dimension of spectrum                                   */
/*      nt        Number of times                                         */
/*   Output:                                                              */
/*      array     Input array filtered                                    */
/*----------------------------------------------------------------------- */
{
  int   ifreq, itime, j, jfreq, ii;
  float scale;

  /* first filtering in time/rate */
  /* loop over frequency */
  for (ifreq = 0; ifreq<nf; ifreq++) {
    jfreq = 2 * ifreq; /* frequency index */
    /* collect data into FTwork1 */
    ii = 0;
    for (itime=0; itime<nt; itime++) { 
      FTwork1[ii++] = array[itime][jfreq];
      FTwork1[ii++] = array[itime][jfreq+1];
    }

    /* do the forward FFT */
    fourg (FTwork1, nt, 1, FTwork2);

    /* do the filtering */
    /* Note: fourg leaves the zero cell in the first element */
    FTwork1[0] = 0.0; FTwork1[1] = 0.0; 
    for (j=1; j<=T_FILTER_WIDTH/2; j++) {
      ii = j * 2;
      FTwork1[ii] = 0.0; FTwork1[ii+1] = 0.0; 
      ii = (nt - j) * 2;
      FTwork1[ii] = 0.0; FTwork1[ii+1] = 0.0; 
    }

    /* do the reverse FFT */
    fourg (FTwork1, nt, -1, FTwork2);

    /* copy back to array, renormalize for effects of fourg */
    ii = 0;
    scale = 1.0 / (float)nt;
    for (itime=0; itime<nt; itime++) { 
      array[itime][jfreq]   =  scale * FTwork1[ii++];
      array[itime][jfreq+1] =  scale * FTwork1[ii++];
    }
  } /* end loop over frequency */

  /* Then filtering in freq/delay */
  /* loop over time */
  for (itime=0; itime<nt; itime++) {

    /* do the forward FFT */
    fourg (array[itime], nf, 1, FTwork2);

    /* do the filtering */
    /* Note: fourg leaves the zero cell in the first element */
    array[itime][0] = 0.0; array[itime][1] = 0.0; 
    for (j=1; j<=F_FILTER_WIDTH/2; j++) {
      ii = j * 2;
      array[itime][ii] = 0.0; array[itime][ii+1] = 0.0; 
      ii = (nf - j) * 2;
      array[itime][ii] = 0.0; array[itime][ii+1] = 0.0; 
    }

    /* do the reverse FFT */
    fourg (array[itime], nf, -1, FTwork2);

    /* renormalize for effects of fourg */
    ii = 0;
    scale = 1.0 / (float)nf;
    for (j=0; j<nf; j++) {
      ii = j * 2;
      array[itime][ii]   =  scale * array[itime][ii];
      array[itime][ii+1] =  scale * array[itime][ii+1];
    }
  } /* end loop over time */
} /* end notch_filter */

void correct_delay (float **array, int nf, int nt, float *OPD_off, 
		 float *phase_corr, imaging_detector_table_def* detector)
/*----------------------------------------------------------------------- */
/*   Remove delay from data array                                         */
/*   Interpolation done using shift theorem and applying a phase ramp     */
/*   to each spectrum.                                                    */
/*   Inputs:                                                              */
/*      array     Measured spectrum  (nf x nt) array of detector values   */
/*      nf        Dimension of spectrum                                   */
/*      nt        Number of times                                         */
/*      OPD_off   OPD correction applied (seconds)                        */
/*      phase_corr Calibration phase (rad) per sample                     */
/*      detector  imaging_detector table definition structure             */
/*   Output:                                                              */
/*      array     Input array with initial OPD modulation removed         */
/*----------------------------------------------------------------------- */
{ 
  int   itime, ifreq, j, ii, jj;
  float temp;
  double freq, freq0, freq_incr, delay, phase_fact;
  float  ref_pix, phase, phase0, sp, cp, re, im, sum;

  /* get info from structures */
  freq0     = detector->crval[0][0]; /* frequency at reference pixel */
  ref_pix   = detector->crpix[0][0]; /* reference pixel 1-rel */
  freq_incr = detector->cd[0][0];    /* frequency spacing */

  /* refer phases to the reference time, freq */
  phase0 = TWOPI * OPD_off[nt/4] * (freq0-ref_pix*freq_incr);

  /* debug - print array */
  /*
  print_array(1, array, 0, 30, 0, nt, 
	      "Array before correct_delay");
  print_array(2, array, 0, 30, 0, nt, 
	      "Array before correct_delay");
  */

  /* loop over actual time slice */
  ii = 0;
  for (itime = 0; itime<nt; itime++) {
    phase_fact = TWOPI * OPD_off[itime];

    /* loop over frequency applying phase ramp */
    for (ifreq=0; ifreq<nf; ifreq++) {
      freq = freq0 + (ifreq-ref_pix)*freq_incr;
      phase = phase_fact * freq + phase_corr[itime];
      sp = sin(phase);
      cp = cos(phase);
      re = array[itime][ifreq*2];
      im = array[itime][ifreq*2+1];
      array[itime][ifreq*2]   = re*cp - im*sp;
      array[itime][ifreq*2+1] = re*sp + im*cp;
    } /* end of loop in freq */
  } /* end loop in time slice */
  /* debug - print array */
  /*
  print_array(1, array, 0, 60, 0, nt, 
	      "Array after correct_delay");
  print_array(2, array, 0, 60, 0, nt, 
	      "Array after correct_delay");
  */

} /* end  correct_delay */

int avg_in_time (float **array, int nf, int nt, float *weight, float *wt)
/*----------------------------------------------------------------------- */
/*   Do weighted average of data in time                                  */
/*   Result multiplied by 2.0 - (is this) to correct for having only reals*/
/*   Inputs:                                                              */
/*      array     Measured spectrum  (nf x nt) array of detector values   */
/*      nf        Dimension of spectrum                                   */
/*      nt        Number of times                                         */
/*      weight    weighting factors per time                              */
/*   Output:                                                              */
/*      array     Data averaged in first time slot                        */
/*      wt        Data weights of time averaged data                      */
/*   Returns                                                              */
/*      1 if there is valid data in the average, else 0                   */
/*----------------------------------------------------------------------- */
{
  int   ifreq, itime, j, ii;
  float sumwt, sumre, sumim, isumwt;

  /* find sum of weights */
  sumwt = 0.0;
  for (itime=0; itime<nt; itime++) sumwt += weight[itime];
  if (sumwt<1.0) return 0;  /* no good data */
  isumwt = 2.0 / sumwt;

  /* loop over frequency */
  for (ifreq = 0; ifreq<nf; ifreq++) {
    j = 2 * ifreq;

    /* sum data */
    sumre = 0.0; sumim = 0.0;
   for (itime=0; itime<nt; itime++) { 
      sumre += weight[itime] * array[itime][j];
      sumim += weight[itime] * array[itime][j+1];
    }
    /* save in first time slot */
    array[0][j]   = sumre * isumwt;
    array[0][j+1] = sumim * isumwt;
    wt[ifreq] = sumwt;

  } /* end loop in freq  */
  /* debug - print array */
  /*
  print_array(1, array, 0, 30, 0, 10, 
      "Array after avg_in_time");
  print_array(2, array, 0, 30, 0, 10, 
  "Array after avg_in_time");
  */

  return 1; /* OK */
} /* end  avg_in_time */

void avg_in_freq (float **array, float *weight, int nf, 
		 float *vis, float *wt)
/*----------------------------------------------------------------------- */
/*   Do average of data in freq by chAv channels                          */
/*   Inputs:                                                              */
/*      array     Measured spectrum  (nf x nt) array of detector values   */
/*                data in first time element                              */
/*      weight    Data weights of time averaged data                      */
/*      nf        Dimension of spectrum                                   */
/*   Output:                                                              */
/*      vis       Data averaged                                           */
/*      wt        Data weights = sum of weights                           */
/*----------------------------------------------------------------------- */
{
  int   ifreq, i, j, ii, indx, jndx, n;
  float sumwt, sumre, sumim;

  /* loop over frequency */
  indx = 0; jndx = 0;
  for (ifreq = 0; ifreq<nf; ifreq+=chAv) {

    /* sum data */
    sumwt = 0.0; sumre = 0.0; sumim = 0.0;
   for (i=ifreq; i<ifreq+chAv; i++) { 
     j = 2 * i;
     sumwt += weight[ifreq];
     sumre += weight[ifreq] * array[0][j];
     sumim += weight[ifreq] * array[0][j+1];
   }

   /* average */
   if (sumwt>0.0) {
     vis[indx++]  = sumre / sumwt;
     vis[indx++]  = sumim / sumwt;
     wt[jndx++] = sumwt;
   } else { 
     vis[indx++]  = 0.0;
     vis[indx++]  = 0.0;
     wt[jndx++] = 0.0;
   }
  } /* end loop in freq  */

  /* debug - print array */
    /* debug */
   n = (int) (((float)nf / (float)chAv) + 0.999);
   for (i=0; i<n; i++) {
     ii = i*2;
    array[0][ii]   = vis[ii];
    array[0][ii+1] = vis[ii+1];
   }
   
  print_array(1, array, 0, n, 0, 1, 
      "Array after avg_in_freq");
  print_array(2, array, 0, n, 0, 1, 
      "Array after avg_in_freq");
   

} /* end  avg_in_freq */

void print_array (int type, float **array, int first_delay, int last_delay, 
		 int first_rate, int last_rate, char *title)
/*----------------------------------------------------------------------- */
/*   Print selected window of  in 2D (d x r)  complex array               */
/*   Inputs:                                                              */
/*      type        1 = amplitude, 2 = phase                              */
/*      array       Complex array  as (real, imag)                        */
/*      first_delay First delay complex element to print                  */
/*      last_delay  Last  delay complex element to print                  */
/*      first_rate  First rate complex element to print                   */
/*      last_rate   Last rate complex element to print                    */
/*      title       title for display                                     */
/*----------------------------------------------------------------------- */
{
  float val;
  int   id, ir, iid;


  if (type==1) { /* amplitude */
    fprintf (stdout,"Amplitudes x 10**-3:  %s \n",title);
    for (id=first_delay; id<last_delay; id++) {
      fprintf (stdout,"  if=%3d ",id);
      iid = id*2;
      for (ir=first_rate; ir<last_rate; ir++) {
	/* index into real array */
	val = array[ir][iid]*array[ir][iid] + 
	  array[ir][iid+1]*array[ir][iid+1];
	val = 0.001*sqrt(val);
	fprintf (stdout," %8.2f", val);
      }
      /*fprintf (stdout,"\n");*/
    }
  } else { /* phase */
    fprintf (stdout,"Phases:  %s \n",title);
    for (id=first_delay; id<last_delay; id++) {
      fprintf (stdout,"  if=%3d ",id);
      iid = id*2;
      for (ir=first_rate; ir<last_rate; ir++) {
	/* index into real array */
	val = 57.296 * atan2(array[ir][iid+1], array[ir][iid]);
	fprintf (stdout," %8.1f", val);
      }
      fprintf (stdout,"\n");
    }
  }
    fprintf (stdout,"\n");
} /* end print array */

fitsfile* init_caltable (imaging_data_table_def* tabledef,
			 cal_table_def **table, int *ierr)
/*----------------------------------------------------------------------- */
/*   Initialize reading calibration table                                 */
/*   Calibration information stores in a global structure                 */
/*   Inputs:                                                              */
/*      tabledef    input data set descriptor                             */
/*   Outputs:                                                             */
/*      table       table_def for I/O                                     */
/*      ierr        Error code: 0 => ok                                   */
/*   Returns                                                              */
/*      cfitsio file pointer, NULL indicated no file to be written.       */
/*----------------------------------------------------------------------- */
{
  int i;
  fitsfile *out=NULL;

  /* Open/create basic FITS output file*/
  fits_open_file(&out, in2file, READONLY, ierr);
  if (*ierr>0) { /* error check */
    fprintf(stderr,"ERROR %d opening calibration FITS file %s \n", *ierr, in2file);
    table_print_error_stack();  /* Show any cfitsio error messages */
    return NULL;
  }/* end error test */

  /* create table_def for table */
  *table = open_cal_table(out, ierr);
  if ((*ierr>0) || (*table==NULL)) return NULL;

  /* make table row objects to store calibration info */
  cal_row = make_cal_table_row(*table);
  cal_tmp = make_cal_table_row(*table);

  /* get size of table */
  fits_get_num_rows (out, &num_cal, ierr);

  /* force read on first call to cal_delay */
  ical = 0; /* 1-rel row number */
  cal_row->time  = -1.0e5;
  cal_row->timei = 0.0;

  return out;
} /* end init_caltable */

void cal_delay(fitsfile *calfptr, cal_table_def* table,
	       double time, float *calDelay, float *calPhase, int *ierr)
/*----------------------------------------------------------------------- */
/*  Read calibration table entries from open FITS file                    */
/*  Note: this only works for a single baseline                           */
/*   Inputs:                                                              */
/*      calfprt     cfitsio file pointer for output FITS file             */
/*      table       table_def for I/O                                     */
/*      time        time in days for calibration values                   */
/*   Outputs:                                                             */
/*      calDelay    calibration delay (sec) flbank = invalid calibration  */
/*      calPhase    calibration phase (rad) flbank = invalid calibration  */
/*      ierr        Error code: 0 => ok                                   */
/*----------------------------------------------------------------------- */
{
  /* locate calibration record for time */
  if ((ical<num_cal) &&
      (cal_row->time+5.787037e-6*cal_row->timei) < time) {
    cal_tmp->time  = cal_row->time;
    cal_tmp->timei = cal_row->timei;
    /* look for right time */
    while ((ical<num_cal) && 
	   ((cal_tmp->time+5.787037e-6*cal_tmp->timei) < time)) {
      ical++;
      read_cal_table(calfptr, table, cal_tmp, ical, ierr);
      /* ignore reference telescope values */
      while ((ical<num_cal) && 
	      (cal_tmp->tele == cal_tmp->ref)) {
	ical++;
	read_cal_table(calfptr, table, cal_tmp, ical, ierr);
      }
    }
    /* if found a better one use it */
    if (cal_tmp->tele != cal_tmp->ref)
      read_cal_table(calfptr, table, cal_row, ical, ierr);
  } /* end find new record */

  /* is this one acceptable? */
  if (cal_row->weight < minSNR) {
    *calDelay = fblank;
    *calPhase = fblank;
    return;
  }

  /* get delay at time time */
  *calDelay = (cal_row->delay +
    86400.0 * (time - cal_row->time) * cal_row->rate);

  /* get phase at reference time */
  *calPhase = atan2(cal_row->gain[1], cal_row->gain[0]);
  /* *calPhase = 0.0;   debug */

  /* debug 
   *calDelay = 0.0;
   *calDelay = cal_row->delay + 
   86400.0 * (time - cal_row->time) * cal_row->rate;
   *calPhase = 0.0;
   */
  return;
} /* end cal_delay */

void close_caltable(fitsfile *calfprt, cal_table_def* table, int *ierr)
/*----------------------------------------------------------------------- */
/*   Close calibration table in FITS file (closes file)                   */
/*   Also does other cleanup                                              */
/*   Inputs:                                                              */
/*      calfprt     cfitsio file pointer for output FITS file             */
/*      table       table_def for I/O                                     */
/*      cal_rows    cal_table structures to be deleted                    */
/*   Outputs:                                                             */
/*      ierr        Error code: 0 => ok                                   */
/*----------------------------------------------------------------------- */
{
  /* are we writing output? */
  if (calfprt==NULL) return; /* nope */

  /* Close FITS file - this will flush open table buffer */
   close_cal_table(calfprt, table, ierr);

  /* delete structures */
   kill_cal_table_row (cal_row);
   cal_row = NULL;
} /* end close_caltable */

fitsfile* init_outdata (imaging_data_table_def* tabledef, 
			fitsfile *infptr, fitsfile *calfptr,
			visibility_table_def **table, int *ierr)
/*----------------------------------------------------------------------- */
/*   Open output and set up to write                                      */
/*   Copies old, writes new history                                       */
/*   Inputs:                                                              */
/*      tabledef    input data set descriptor                             */
/*      infptr      input FITS file cfitsio                               */
/*      calfptr     input cal table FITS file cfitsio                     */
/*   Outputs:                                                             */
/*      table       visibility table_def for I/O                          */
/*      ierr        Error code: 0 => ok                                   */
/*   Returns                                                              */
/*      cfitsio file pointer, NULL indicated no file to be written.       */
/*----------------------------------------------------------------------- */
{
  int i;
  fitsfile *out=NULL;
  visibility_geom *geom;
  double crval[2];  /* reference value for each axis */
  float  crpix[2];  /* reference pixels */
  float  cd[2][2];  /* cd transformation matrix */
  int  naxis[2];    /* dimension of two axes of data, weight arrays */
  char *date_obs;   /* observing date */
  char *ctype[2];   /* labels for the two axes */
  imaging_detector_table_def *detec;
  int n, nf;

  *table = NULL;

  /* Open/create basic FITS output file*/
  fits_create_file(&out, outfile, ierr);
  if (*ierr>0) { /* error check */
    fprintf(stderr,"ERROR %d opening output FITS file %s \n", *ierr, outfile);
    table_print_error_stack();  /* Show any cfitsio error messages */
    return NULL;
  }/* end error test */

  /* make geometry descriptor - assume region 1 */
  detec = tabledef->detector;
  date_obs = detec->date_obs;
  /* how many channels after averaging */
  nf = detec->naxis[0][0] ;
  n = (int) (((float)nf / (float)chAv) + 0.999);
  naxis[0] = n;
  naxis[1] = detec->naxis[0][1];
  ctype[0] = detec->ctype[0][0];
  ctype[1] = detec->ctype[0][1];
  crval[0] = detec->crval[0][0];
  crval[1] = detec->crval[0][1];
  crpix[0] = ((detec->crpix[0][0]-1.0) / chAv) + 1.0; /* including averaging */
  crpix[1] = detec->crpix[0][1];
  cd[0][0] = detec->cd[0][0] * chAv; /* including averaging */
  cd[0][1] = detec->cd[0][1];
  cd[1][0] = detec->cd[0][2];
  cd[1][1] = detec->cd[0][3];
  geom = make_visibility_geom(naxis, ctype, crval, crpix, cd, date_obs);

  /* create/init table */
  *table = make_visibility_table_def(out, geom, ierr);
  if ((*ierr>0) || (*table==NULL)) return NULL;

  /* copy data history */
  /* copy_history (infptr, out, ierr); */

  /* copy cal table history - this will include the data history*/
  copy_history (calfptr, out, ierr);

  /* add parameters in history */
  history(out, ierr);

  return out;
} /* end init_outdata */

void write_outdata(fitsfile *outfptr, visibility_table_def* table, 
		    visibility_row *visrow, imaging_data_row *row,
		   int *ierr)
/*----------------------------------------------------------------------- */
/*   Write calibration table entries to open FITS file                    */
/*   Inputs:                                                              */
/*      outfptr     cfitsio file pointer for output FITS file             */
/*      table       table_def for I/O                                     */
/*      visrow      visibility structures to be written                   */
/*      row         imaging_data row representative of visrow             */
/*   Outputs:                                                             */
/*      ierr        Error code: 0 => ok                                   */
/*----------------------------------------------------------------------- */
{
  /* are we writing output? */
  if (outfptr==NULL) return; /* nope */

  /* copy descriptive values */
  visrow->source = row->source[0];
  visrow->offset[0] = row->offset[0];
  visrow->offset[1] = row->offset[1];
  visrow->rotation = row->rotation;

  /* write it */
  write_visibility_table(outfptr, table, visrow, (long)-1, ierr);
} /* end write_outdata */

void close_outdata(fitsfile *outfptr, visibility_table_def* table, int *ierr)
/*----------------------------------------------------------------------- */
/*   Close visibility table in FITS file (closes file)                    */
/*   Inputs:                                                              */
/*      outfptr     cfitsio file pointer for output FITS file             */
/*      table       table_def for I/O                                     */
/*   Outputs:                                                             */
/*      ierr        Error code: 0 => ok                                   */
/*----------------------------------------------------------------------- */
{
  /* are we writing output? */
  if (outfptr==NULL) return; /* nope */

  /* Close FITS file - this will flush open table buffer */
   close_visibility_table(outfptr, table, ierr);
} /* end close_outdata */

void history (fitsfile *fptr, int *ierr)
/*----------------------------------------------------------------------- */
/*   Write model parameters in FITS history                               */
/*   Inputs                                                               */
/*   Also writes to stdout                                                */
/*      fptr      cfitsio pointer to open FITS file                       */
/*   Output:                                                              */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*----------------------------------------------------------------------- */
{
  char hiline[FLEN_CARD], ttime[FLEN_CARD];

  sprintf(hiline,"*** VLTI Midi testbed data calibrator ***");
  fits_write_history (fptr, hiline, ierr);
  fprintf(stdout, "%s in file %s\n", hiline, outfile);
  history_time(ttime); /* get current time/date */
  sprintf(hiline,"calib run %s", ttime);
  fits_write_history (fptr, hiline, ierr);
  fprintf(stdout, "%s in file %s\n", hiline, outfile);
  sprintf(hiline,"calib input data file = %s",infile);
  fits_write_history (fptr, hiline, ierr);
  fprintf(stdout, "%s\n", hiline);
  sprintf(hiline,"calib input cal file = %s",in2file);
  fits_write_history (fptr, hiline, ierr);
  fprintf(stdout, "%s\n", hiline);
  sprintf(hiline,"calib Calibration interval = %d",CI);
  fits_write_history (fptr, hiline, ierr);
  fprintf(stdout, "%s\n", hiline);
  sprintf(hiline,"calib No. channels to average = %d",chAv);
  fits_write_history (fptr, hiline, ierr);
  fprintf(stdout, "%s\n", hiline);
  sprintf(hiline,"calib min fringe SNR = %f ",minSNR);
  fits_write_history (fptr, hiline, ierr);
  fprintf(stdout, "%s\n", hiline);
} /* end history */
/* */
