/*
 * This file is part of MIA.
 *
 * MIA is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/* $Id: cube2table.c,v 1.4 2002/06/04 09:29:29 jdejong Exp $ */
#define _POSIX_SOURCE_1
#include "vltPort.h"

/*
   Fix scaling problem with image data.
 */

/* $Id: cube2table.c,v 1.4 2002/06/04 09:29:29 jdejong Exp $ */
/*-----------------------------------------------------------------------*/
/*   This program used an Imaging Detector FITS binary table to convert  */
/* MIDI image cube into a FITS binary table.  No state information is    */
/* a written into the output and the times are replaced with a sequence  */
/* number.                                                               */
/* The output file contains an imaging detector table and an imaging     */
/* data table with the detector windows.                                 */
/*                                                                       */
/* Command line arguments:                                               */
/* -input  Input parameter file name                                     */
/*        def. cube2table.in                                             */
/* -incube Input FITS image cube with detector frames.                   */
/*        def. MIDIcube.fit                                              */
/* -inID   Input FITS binary file with Imaging Detector table            */
/*         def. makeID.fit                                               */
/* -output Output FITS binary file with Imaging Detector table           */
/*         def. MIDItable.fit                                            */
/*-----------------------------------------------------------------------*/
#include "imaging_detector.h"
#include "image.h"
#include "imaging_data.h"
#include "parser.h"

/* global variables, possibly read from command line */
char     *input_file="cube2table.in";       /* input parameter file name */
char     *incube  = "MIDIcube.fit";  /* input cube */
char     *intable = "makeID.fit";    /* input imaging detector table */
char     *output_file="!cube2table.fit";    /* output FITS file name */

/* global variables possibly read from the input_file */
long start;   /* initial sequence number */

/* internal prototypes */
void cube2tablein (int argc, char **argv, int *ierr);
void Usage(void);
/* read input Imaging Detector Table */
imaging_detector_table_def* read_ID_table (char *file, int *ierr);

/* convert image cube to Imaging data table */
void convert_to_table (fitsfile *outfptr,
		       image_info* imageinfo,
		       imaging_detector_table_def* tabledef, 
		       char *out,
		       long start,
		       int *ierr);
/* convert detector frame to table row */
void frame_to_row (long nx, long ny, DataType* cube_data, 
		   int nreg, int corn[20][4],
		   imaging_data_row* row);
/* create parser list */
parse_list* make_parse_list();
/* copy parse_list to global variables */
void parse_list_to_global (parse_list* list);

int main ( int argc, char **argv )
/*----------------------------------------------------------------------- */
/*   Program to convert a detector image cube to binary table format.     */
/*----------------------------------------------------------------------- */
{
  int  ierr;
  imaging_detector_table_def* tabledef=NULL;
  image_info* imageinfo=NULL;
  fitsfile *outfptr=NULL;

  /* Startup - parse command line */
  ierr = 0;
  cube2tablein (argc, argv, &ierr);
  if (ierr!=0) return ierr;

  /* Open/create basic FITS output file*/
  fits_create_file(&outfptr, output_file, &ierr);
  if (ierr>0) { /* error check */
    fprintf(stderr,"ERROR %d opening output FITS file %s \n", 
	    ierr, output_file);
    table_print_error_stack();  /* Show any cfitsio error messages */
    return ierr;
  }/* end error test */

  /* Open image file */
  imageinfo = open_image(incube, &ierr);
  if (ierr>0) { /* error check */
    fprintf(stderr,"ERROR %d opening inputput FITS cube %s \n", 
	    ierr, incube);
    table_print_error_stack();  /* Show any cfitsio error messages */
    return ierr;
  }/* end error test */

  /* read imaging detector table */
  tabledef = read_ID_table(intable, &ierr);
  if ((tabledef==NULL) || (ierr>0)) { /* error test */
    fprintf (stderr,"error %d reading Imaging Detector table\n",ierr);
    table_print_error_stack();  /* Show any cfitsio error messages */
    return ierr;
  } /* end error test */

  /* convert cube to table */
  convert_to_table (outfptr, imageinfo, tabledef, output_file, start, 
		    &ierr);
  if (ierr>0) { /* error test */
    fprintf (stderr,"error %d converting cube to table\n",ierr);
    table_print_error_stack();  /* Show any cfitsio error messages */
    return ierr;
  } /* end error test */
  
  /* close/delete image_info */
  image_close(imageinfo, &ierr);
  kill_image_info (imageinfo);

   /* Show any cfitsio error messages */
  table_print_error_stack(); 

  return ierr;
} /* end of main */

void cube2tablein (int argc, char **argv, int *ierr)
/*----------------------------------------------------------------------- */
/*  Parse control info from command line                                  */
/*   Input:                                                               */
/*      argc   Number of arguments from command line                      */
/*      argv   Array of strings from command line                         */
/*   Output:                                                              */
/*      ierr       Error code: 0 => ok                                    */
/*----------------------------------------------------------------------- */
{
  int ax;
  char *arg;
  parse_list* list;

  /* command line arguments */
  /*debug    fprintf(stderr, "number of arguments = %d\n",argc);*/
  if (argc<=1) Usage(); /* must have arguments */
  /* parse command line */
  for (ax=1; ax<argc; ax++) {
    arg = argv[ax];
    /*debug    fprintf(stderr, "arg = %s\n",arg);*/
    if (strcmp(arg, "-input") == 0){ /* input parameters */
      input_file = argv[++ax];
    } else if (strcmp(arg, "-output") == 0){ /* output parameters */
      output_file = argv[++ax];
    } else if (strcmp(arg, "-incube") == 0) { /* input image cube */
      incube = argv[++ax];
    } else if (strcmp(arg, "-intable") == 0){ /* input ID table */
      intable = argv[++ax];
    } else { /* unknown argument */
      Usage();
    }
    }
  
  /* must specify file */
  if (!input_file) {Usage(); exit(1);}

  /* parse input file */
  /* create/init parse_list */
  list = make_parse_list();

  /* parse input file, bail out on error */
  if (parse_input_file(input_file, list)) exit(1);

  /* copy to input_info */
  parse_list_to_global (list);

  /* debug 
     return;*/

  /* delete parser list */
  kill_parse_list(list);

  *ierr = 0;
} /* end cube2tablein */

void Usage(void)
/*----------------------------------------------------------------------- */
/*   Tells about usage of program and bails out                           */
/*----------------------------------------------------------------------- */
{
    fprintf(stderr, "Usage: cube2table -input file\n");
    fprintf(stderr, "Convert detector image cube and Imaging Detector table\n");
    fprintf(stderr, "to FITS binary Imaging Data file.\n");
    fprintf(stderr, "Arguments:\n");
    fprintf(stderr, "  -input input parameter file, def cube2table.in\n");
    fprintf(stderr, "  -output output FITS file, def cube2table.fit\n");
    fprintf(stderr, "  -incube input detector cube file, def MIDIcube.fit\n");
    fprintf(stderr, "  -intable input ID table file, def makeID.fit\n");
    
    /*/exit(1);  bail out */
  }/* end Usage */

  parse_list* make_parse_list() {
/*----------------------------------------------------------------------- */
/*  create parser list for cube2table                                         */
/*----------------------------------------------------------------------- */
  parse_list *out = create_parse_list();

  /* add parser items */
  /* initial sequence number */
  out->list[(out->NumItem)++] = 
    make_parse_item ("start", PINTEGER, 0.0, 1, NULL);

  return out;
} /* end make_parse_list */

/* copy parse_list to global variables */
void parse_list_to_global (parse_list* list) {
/*----------------------------------------------------------------------- */
/* copy parse_list to global variables                                    */
/*----------------------------------------------------------------------- */
  long indx;

  /* number of windows */
  indx = find_parse_entry("start", list);
  if (indx>=0) start = list->list[indx]->ivalue;
} /* end  parse_list_to_global */

imaging_detector_table_def* read_ID_table (char *file, int *ierr)
/*----------------------------------------------------------------------- */
/*   Read an imaging detector table from a file                           */
/*   Inputs:                                                              */
/*      file      file name to read from                                  */
/*   Output:                                                              */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*   Returns:                                                             */
/*       imaging_data_table_def structure                                 */
/*       NULL on failure                                                  */
/*----------------------------------------------------------------------- */
{
  imaging_detector_table_def *out=NULL;
  fitsfile *fptr;
  int ireg;

  if (*ierr>0) return NULL; /* existing error condition? */

  /* attempt to open file */
  fits_open_file(&fptr, file, READONLY, ierr);
  if (*ierr>0) { /* error check */
    fprintf(stderr,"ERROR %d opening input FITS ID table %s \n", *ierr, 
	    file);
    table_print_error_stack();  /* Show any cfitsio error messages */
    return out;
  }/* end error test */

  /* read imaging_detector table  */
  out = read_imaging_detector_table(fptr, ierr);
  if (*ierr>0) { /* error check */
    fprintf(stderr,"ERROR %d reading input FITS ID table %s \n", *ierr, 
	    file);
    table_print_error_stack();  /* Show any cfitsio error messages */
    kill_imaging_detector_table_def(out); /* cleanup */
    return NULL;
  }/* end error test */

  /* close input file */
  fits_close_file(fptr, ierr);

  return out;
} /* end read_ID_table */

/* convert image cube to Imaging data table */
void convert_to_table (fitsfile *outfptr,
		       image_info* imageinfo,
		       imaging_detector_table_def* tabledef, 
		       char *out,
		       long start,
		       int *ierr) 
/*----------------------------------------------------------------------- */
/* Convert image cube to Imaging data table                               */
/*   Inputs:                                                              */
/*      outfptr   FITS output file, open on call, closed on return.       */
/*      imageinfo Detector cube image_info to read                        */
/*      tabledef  Imaging detector table definition                       */
/*      out       name of output file                                     */
/*      start     first sequence number to use                            */
/*   Output:                                                              */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*----------------------------------------------------------------------- */
{
  imaging_data_table_def *outtable;
  long i, size, nx, ny;
  DataType* cube_data; /* pointer to data */
  imaging_data_row* row;
  int nreg, maxstep, corn[20][4];

  if (*ierr>0) return; /* existing error? */
  if (!imageinfo) {*ierr=10000; return;}
  if (!tabledef)  {*ierr=10000; return;}
  if (!out)       {*ierr=10000; return;}

  /* save windowing info into local array */
  nreg = tabledef->nregion;
  for (i=0; i<nreg; i++) {
    corn[i][0] = tabledef->corner[i][0];
    corn[i][2] = tabledef->corner[i][1];
    corn[i][1] = tabledef->corner[i][0] + tabledef->naxis[i][0] - 1;
    corn[i][3] = tabledef->corner[i][1] + tabledef->naxis[i][1] - 1;
    /* convert to zero relative */
    corn[i][0]--;
    corn[i][2]--;
    corn[i][1]--;
    corn[i][3]--;
  }

  /* write imaging_detector table and open/init imaging data table */
  maxstep = 128; /* maximum value of OPD sequence step number*/
  outtable = make_imaging_data_table_def(outfptr, tabledef, maxstep, ierr);
  if ((out==NULL) || (*ierr>0)) { /* failed */
    fprintf(stderr,"ERROR %d create/open output Imaging Data table \n", 
	    *ierr);
    if (*ierr==0) *ierr = 10000;
    return;
  } /* end error test */

  /* create row structure */
  row = make_imaging_data_row(outtable);
  if (row==NULL) {
    fprintf(stderr,"ERROR: creating  Imaging Data table row \n");
    *ierr = 10000;
    return;
  }

  /* allocate image cube memory */
  size = imageinfo->dim[0] * imageinfo->dim[1];
  cube_data = (DataType*)malloc(size * FITS_type_to_size(DATACODE));
  nx = imageinfo->dim[0];  /* size of detector frame */
  ny = imageinfo->dim[1];

  /* loop over image converting to table rows */
  for (i=0; i<imageinfo->dim[2]; i++) {

    /* read detector frame */
    image_read_raw (imageinfo, (void*)cube_data, ierr);
    if (*ierr>0) {
      fprintf(stderr,"ERROR %d reading detector frame %ld \n", 
	      *ierr, i);
      return;
    } /* end error check */

    /* convert detector frame to table row */
    frame_to_row (nx, ny, cube_data, nreg, corn, row);

    /* set time to sequence number */
    row->time = i + start;

    write_imaging_data_table(outfptr, outtable, row, i+1, ierr);
    if (*ierr>0) {
      fprintf(stderr,"ERROR %d writing Imaging Data table row %ld \n", 
	      *ierr, i);
      return;
    } /* end error check */
  } /* end loop converting */

  /* delete row structure */
   kill_imaging_data_row (row);

  /* Close FITS file - this will flush open table buffer */
  close_imaging_data_table (outfptr, outtable, ierr);
} /* end convert_to_table */

void frame_to_row (long nx, long ny, DataType* cube_data, 
		   int nreg, int corn[20][4],
		   imaging_data_row* row)
/*----------------------------------------------------------------------- */
/* Convert detector frame to table row                                    */
/*   Inputs:                                                              */
/*      nx         X dimension of cube_data                               */
/*      ny         Y dimension of cube_data                               */
/*      cube_data  Detector image                                         */
/*      nreg       Number of regions                                      */
/*      corn       Corners, lo x, hi x, lo y, hi y                        */
/*   Output:                                                              */
/*      row        table row structure                                    */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*----------------------------------------------------------------------- */
{
  long ix, iy, ireg;
  DataType *in, *out;

  for (ireg=0; ireg<nreg; ireg++) {
    out = row->data[ireg];
    for (iy=corn[ireg][2]; iy<=corn[ireg][3]; iy++) {
      in = cube_data + iy * nx + corn[ireg][0];
      for (ix=corn[ireg][0]; ix<=corn[ireg][1]; ix++) {
	*out++ = *in++;
      } /* end loop over x */
    } /* end loop over y */
  } /* end loop over region */
} /* end frame_to_row */
/* */
