/*
 * This file is part of MIA.
 *
 * MIA is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/* $Id: demoread.c,v 1.4 2002/05/02 15:20:09 jdejong Exp $ */
#define _POSIX_SOURCE_1
#include "vltPort.h"

/*-----------------------------------------------------------------------*/
/*  Demo for reading VLTI raw data tables                                */
/*  Reads IMAGING_DETECTOR and IMAGING_DATA table and                    */
/*  prints contents on stdout                                            */
/*-----------------------------------------------------------------------*/
#include <math.h>
#include <malloc.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "imaging_data.h"
#include "source_table.h"


/* global variables */
char     *infile=NULL;      /* input FITS file name */

/* internal prototypes */
void testfin (int argc, char **argv, int *ierr);
void Usage(void);
void read_data_table (fitsfile *fptr, imaging_data_table_def* tabledef, 
		      int *ierr);
void print_source_table(fitsfile *fptr, source_table_def *tabledef, 
			int *ierr);

int main ( int argc, char **argv )
/*----------------------------------------------------------------------- */
/*   Program to read a FITS binary table                                  */
/*----------------------------------------------------------------------- */
{
  int  ierr;
  imaging_data_table_def* tabledef;
  source_table_def *sourcedef;
  fitsfile *infptr=NULL;

  /*  Startup - parse command line arguments*/
  ierr = 0;
  testfin (argc, argv, &ierr);
  if (ierr!=0) return ierr;

  /* Open input file*/
  fits_open_file(&infptr, infile, READONLY, &ierr);
  if (ierr>0) { /* error check */
    fprintf(stderr,"ERROR %d opening input FITS file %s \n", ierr, infile);
    table_print_error_stack();  /* Show any cfitsio error messages */
    ierr = 1;
    return ierr;
  }
  
  /* find and read source table */
  sourcedef = open_source_table(infptr, &ierr);
  ierr = 0; /* ignore if not there */
  print_source_table(infptr, sourcedef, &ierr); /* print */
  kill_source_table_def (sourcedef);


  /*  Now get Data table - this reads the IMAGING_DETECTOR table 
      and opens the IMAGING_DATA table */
  tabledef = open_imaging_data_table(infptr, &ierr);
  if (ierr>0) { /* error check */
    fprintf (stderr,"error %d after table open\n",ierr);
    table_print_error_stack();  /* Show any cfitsio error messages */
    return ierr;
  } /* end of error check */

  /* Print IMAGING_DETECTOR descriptor (attached to tabledef) */
  print_imaging_detector_table(tabledef->detector);

  /* Read and print IMAGING_DATA table */
  read_data_table(infptr, tabledef, &ierr);
  if (ierr>0) { /* error check */
    fprintf (stderr,"error %d after table read\n",ierr);
    table_print_error_stack();  /* Show any cfitsio error messages */
    return ierr;
  } /* end of error check */
 
  /* Close FITS file*/
  fits_close_file (infptr, &ierr);

  /* Show any cfitsio error messages */
  table_print_error_stack(); 
    
  return ierr;
} /* end of main */

void testfin (int argc, char **argv, int *ierr)
/*----------------------------------------------------------------------- */
/*  Parse control info from command line */
/*   Input: */
/*      argc   Number of arguments from command line */
/*      argv   Array of strings from command line */
/*   Output: */
/*      ierr       Error code: 0 => ok */
/*----------------------------------------------------------------------- */
{
  int ax;
  char *arg;

  if (argc<=1) Usage(); /* must have arguments */
/* parse command line */
  for (ax=1; ax<argc; ax++) {
    arg = argv[ax];
    if (strcmp(arg, "-in") == 0) {
      infile = argv[++ax];
    }     else { /* unknown argument */
      Usage();
    }
  }

  /* must specify files */
  /*  if (!infile) Usage();*/
  if (!infile) infile = "simul.fit";
  *ierr = 0;
} /* end testfin */

void Usage(void)
/*----------------------------------------------------------------------- */
/*   Tells about usage of program and bails out */
/*----------------------------------------------------------------------- */
{
    fprintf(stderr, "Usage: demoread -in input_file\n");
    fprintf(stderr, "Test for reading binary VLTI data tables\n");
    fprintf(stderr, "Argument:\n");
    fprintf(stderr, "  input_file is FITS file (def simul.fit)\n");
    
    /* exit(1);  bail out */
  }/* end Usage */


void read_data_table (fitsfile *fptr, imaging_data_table_def* tabledef, 
		      int *ierr)
/*----------------------------------------------------------------------- */
/*   Read and print input IMAGING_DATA table                              */
/*   Inputs                                                             : */
/*      fptr      cfitsio pointer to open FITS file                       */
/*      tabledef  table definition structure describing table             */
/*   Output:                                                              */
/*      ierr      Error code: 0 => ok                                     */
/*----------------------------------------------------------------------- */
{
  long i, j, n, irow, nrow, lrow, ireg, max;
  char dtype;
  imaging_data_row *row; 
  table_def *def;

  if (*ierr>0) return;        /* existing error condition? */
  if (tabledef==NULL) return; /* sanity check */

  def = tabledef->def;  /* this is referred to alot */

  /* print some info about the file, keywords */
  fprintf (stdout, "Dump of IMAGING_DATA table from file = '%s' \n",infile);

  /* This prints all of the application keywords from the IMAGING_DATA table */
  for (i=0; i<def->nkey; i++) {
    /* show keyword */
    fprintf (stdout, "%s", def->keywords[i]);
  /* print values by type */
    switch (def->keytype[i]) {
    case TINT:
      fprintf (stdout," = %d",*((int*)def->keydata[i]));
      break;
    case TSHORT:
      fprintf (stdout," = %d",*((int*)def->keydata[i]));
      break;
    case TSTRING:
      fprintf (stdout," = '%s'",((char*)def->keydata[i]));
      break;
    case TFLOAT:
      fprintf (stdout," = %f",*((float*)def->keydata[i]));
      break;
    case TDOUBLE:
      fprintf (stdout," = %f",*((double*)def->keydata[i]));
      break;
    case TLONG:
      fprintf (stdout," = %ld",*((long*)def->keydata[i]));
      break;
    default:
      break;
    } /* end switch on keytype */
    /* add comment and newline */
    fprintf (stdout, " / %s\n", def->keycom[i]);
  } /* end loop over keywords */

  /* Describe columns */
  fprintf (stdout, "\n");
  for (i=0; i<def->ncol; i++) {
    dtype = FITS_type_to_code(def->coldatatype[i]); /* get datatype */
    fprintf (stdout, "col %ld %s count=%ld type=%c units='%s' dim='%s'\n",
	     i+1,def->coltype[i],def->colcount[i], dtype,
	     def->colunits[i],def->coldim[i]);
  } /* end loop over columns */
  fprintf (stdout, "\n");
  
  /* make table_row structure */
  row = make_imaging_data_row (tabledef);

  /* get size of table */
  fits_get_num_rows (fptr, &nrow, ierr);
  
  /* loop over rows, NOTE row numbers are 1 relative */
  for (irow = 1; irow<=nrow; irow++) {
  
    /* read next row */
    lrow = -1;
    read_imaging_data_table(fptr, tabledef, row, lrow, ierr);
    if (*ierr>0) { /* error test */
      fprintf(stderr,"ERROR %d reading IMAGING_DATA_TABLE\n", *ierr);
      return;
    }

    /* print values */
    fprintf (stdout, "\nData for row %ld\n", row->raw->trow);
    fprintf (stdout, "Time = %f, exposure = %f, frame = %ld\n", 
	     row->time, row->exp, row->frame);
    fprintf (stdout, "OPD = %g, localopd = %g\n", 
	     row->opd[0], row->local_opd[0]);
    fprintf (stdout, "Offset = %f, %f\n",
	     row->offset[0], row->offset[1]);
    fprintf (stdout, "Rotation = %f\n",
	     row->rotation);
    fprintf (stdout, "Reference source = %d\n",
	     row->reference);
    fprintf (stdout, "Modulation step = %d\n",
	     row->step);
    /* Loop over regions */
    for (ireg=0; ireg<row->nregion; ireg++) {
      fprintf (stdout, "\nRegion = %ld\n", ireg+1);
      fprintf (stdout, "Source = %d target type = %d\n",
	       row->source[ireg], row->targtype[ireg]);
      fprintf (stdout, "Pixel data is\n     ");
      max = def->colcount[tabledef->data_col[ireg]];
      for (i=0; i<max; i+=5) {
	for (j=0; j<5; j++) { /* put several on row */
	  if (i+j<max) 	fprintf (stdout, "%d ", 
				 row->data[ireg][i+j]);
	} /* end group */
	fprintf (stdout, "\n     "); /* new line */
      }
    } /* end loop over regions */
    fprintf (stdout, "\n"); /* FINAL new line */
  } /* end loop over rows */
  
  /* delete row structure */
  kill_imaging_data_row(row);

} /* end read_data_table */

void print_source_table(fitsfile *fptr, source_table_def *tabledef, 
			int *ierr)
/*----------------------------------------------------------------------- */
/*   Read and print input SOURCE table                                    */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*      tabledef  table definition structure describing table             */
/*   Output:                                                              */
/*      ierr      Error code: 0 => ok                                     */
/*----------------------------------------------------------------------- */
{
  long i, j, n, irow, nrow, lrow, ireg, max;
  char dtype;
  source_table_row *row; 
  table_def *def;

  if (*ierr>0) return;        /* existing error condition? */
  if (tabledef==NULL) return; /* sanity check */

  def = tabledef->def;  /* this is referred to alot */

  /* print some info about the file, keywords */
  fprintf (stdout, "Dump of SOURCE table from file = '%s' \n",infile);

  /* This prints all of the application keywords from the IMAGING_DATA table */
  for (i=0; i<def->nkey; i++) {
    /* show keyword */
    fprintf (stdout, "%s", def->keywords[i]);
  /* print values by type */
    switch (def->keytype[i]) {
    case TINT:
      fprintf (stdout," = %d",*((int*)def->keydata[i]));
      break;
    case TSHORT:
      fprintf (stdout," = %d",*((int*)def->keydata[i]));
      break;
    case TSTRING:
      fprintf (stdout," = '%s'",((char*)def->keydata[i]));
      break;
    case TFLOAT:
      fprintf (stdout," = %f",*((float*)def->keydata[i]));
      break;
    case TDOUBLE:
      fprintf (stdout," = %f",*((double*)def->keydata[i]));
      break;
    case TLONG:
      fprintf (stdout," = %ld",*((long*)def->keydata[i]));
      break;
    default:
      break;
    } /* end switch on keytype */
    /* add comment and newline */
    fprintf (stdout, " / %s\n", def->keycom[i]);
  } /* end loop over keywords */

  /* Describe columns */
  fprintf (stdout, "\n");
  for (i=0; i<def->ncol; i++) {
    dtype = FITS_type_to_code(def->coldatatype[i]); /* get datatype */
    fprintf (stdout, "col %ld %s count=%ld type=%c units='%s' dim='%s'\n",
	     i+1,def->coltype[i],def->colcount[i], dtype,
	     def->colunits[i],def->coldim[i]);
  } /* end loop over columns */
  fprintf (stdout, "\n");
  
  /* make table_row structure */
  row = make_source_table_row (tabledef);

  /* get size of table */
  fits_get_num_rows (fptr, &nrow, ierr);
  
  /* loop over rows, NOTE row numbers are 1 relative */
  for (irow = 1; irow<=nrow; irow++) {
    
    /* read next row */
    lrow = -1;
    read_source_table(fptr, tabledef, row, lrow, ierr);
    if (*ierr>0) { /* error test */
      fprintf(stderr,"ERROR %d reading SOURCE\n", *ierr);
      return;
    }
    
    /* print values */
    fprintf (stdout, "\nSource table row %ld\n", row->raw->trow);
    fprintf (stdout, "Name = %s, number = %d calcode = %s\n", 
	     row->source, row->source_id, row->calcode);
    fprintf (stdout, "RAapp = %g, decapp = %f\n", 
	     row->rapp, row->decapp);
    fprintf (stdout, "Diameter = %f\n",
	     row->diameter);
    for (i=0; i<tabledef->nbands; i++) {
      fprintf (stdout, "Freq = %g Hz, flux = %g Jy\n",
	       row->freq[i], row->iflux[i]);
    }
  } /* end loop over rows */
  
  /* delete row structure */
   kill_source_table_row ((source_table_row*)row);

} /* end print_source_table */
/* */
