/*
 * This file is part of MIA.
 *
 * MIA is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/* $Id: demowrite.c,v 1.4 2002/05/02 15:20:09 jdejong Exp $ */
#define _POSIX_SOURCE_1
#include "vltPort.h"

/*-----------------------------------------------------------------------*/
/*  Demo for writing VLTI raw data tables                                */
/*  Writes IMAGING_DETECTOR and IMAGING_DATA table.                      */
/*-----------------------------------------------------------------------*/
#include <math.h>
#include <malloc.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "imaging_data.h"


/* global variables */
char     *outfile=NULL;     /* output FITS file name */

/* internal prototypes */
void testfin (int argc, char **argv, int *ierr);
void Usage(void);
imaging_data_table_def* 
make_data_table (fitsfile *fptr, int nx, int ny, int nreg, int *ierr);
void write_data_table (fitsfile *fptr, int nx, int ny, int reg,
		       imaging_data_table_def* tabledef, int *ierr);


int main ( int argc, char **argv )
/*----------------------------------------------------------------------- */
/*   Program to write a FITS binary table                                 */
/*----------------------------------------------------------------------- */
{
  int  ierr, nx, ny, nreg;
  imaging_data_table_def* tabledef;
  fitsfile *outfptr=NULL;

  /* Startup - parse command line */
  ierr = 0;
  testfin (argc, argv, &ierr);
  if (ierr!=0) return ierr;

  /* Open/create basic FITS output file*/
  fits_create_file(&outfptr, outfile, &ierr);
  if (ierr>0) { /* error check */
    fprintf(stderr,"ERROR %d opening output FITS file %s \n", ierr, outfile);
    table_print_error_stack();  /* Show any cfitsio error messages */
    return ierr;
  }/* end error test */

  /* Create basic table structures and write IMAGING_DETECTOR table */
  /* set basic paramaters */
  nx = 256;
  ny = 1;
  nreg = 2;
  tabledef = make_data_table(outfptr, nx, ny, nreg, &ierr);
  if (ierr>0) { /* error test */
    fprintf (stderr,"error %d after table create\n",ierr);
    table_print_error_stack();  /* Show any cfitsio error messages */
    return ierr;
  } /* end error test */
  
  /* Write table */
  write_data_table(outfptr, nx, ny, nreg, tabledef, &ierr);
  if (ierr>0) { /* error test */
    fprintf (stderr,"error %d after table write\n",ierr);
    table_print_error_stack();  /* Show any cfitsio error messages */
    return ierr;
  } /* end error test */
  
  /* delete table_def */
  kill_imaging_data_table_def(tabledef,1);
  
  /* Close FITS file - this will flush open table buffer */
  fits_close_file (outfptr, &ierr);

   /* Show any cfitsio error messages */
  table_print_error_stack(); 

  return ierr;
} /* end of main */

void testfin (int argc, char **argv, int *ierr)
/*----------------------------------------------------------------------- */
/*  Parse control info from command line                                  */
/*   Input:                                                               */
/*      argc   Number of arguments from command line                      */
/*      argv   Array of strings from command line                         */
/*   Output:                                                              */
/*      ierr       Error code: 0 => ok                                    */
/*----------------------------------------------------------------------- */
{
  int ax;
  char *arg;

  if (argc<=1) Usage(); /* must have arguments */
  /* parse command line */
  for (ax=1; ax<argc; ax++) {
    arg = argv[ax];
    if (strcmp(arg, "-out") == 0) {
      outfile = argv[++ax];
    }
    else { /* unknown argument */
      Usage();
    }
  }
  
  /* must specify file */
  /*  if (!outfile) Usage();*/
  if (!outfile) outfile = "!demo.fit";
  *ierr = 0;
} /* end testfin */

void Usage(void)
/*----------------------------------------------------------------------- */
/*   Tells about usage of program and bails out                           */
/*----------------------------------------------------------------------- */
{
    fprintf(stderr, "Usage: demowrite -out output_file \n");
    fprintf(stderr, "Demo for writing VLTI binary tables\n");
    fprintf(stderr, "Arguments:\n");
    fprintf(stderr, "  output_file is new FITS file  (def demo.fit)\n");
    
    /*/exit(1);  bail out */
  }/* end Usage */

imaging_data_table_def* make_data_table (fitsfile *fptr, int nx, int ny, 
					 int nreg, int *ierr)
/*----------------------------------------------------------------------- */
/*   Create FITS file and initialize table                                */
/*   Inputs                                                             : */
/*      fptr      cfitsio pointer to open FITS file                       */
/*      nx        Size of data array in x direction                       */
/*      ny        Size of data array in y direction                       */
/*      nreg      Number of regions                                       */
/*   Output:                                                              */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*   Returns:                                                             */
/*       imaging_data_table_def structure to use to write table           */
/*       NULL on failure                                                  */
/*----------------------------------------------------------------------- */
{
  imaging_data_table_def *out;
  imaging_detector_table_def *detector;

  if (*ierr>0) return NULL; /* existing error condition? */

  /* make simple imaging_detector_table_def */
  detector = make_testing_detector_table_def(nx, ny, nreg);
  if (detector==NULL) { /* error check */
    *ierr = 10000;
    return NULL;
  } /* end error test */

  /* write imaging_detector table and open/init imaging data table */
  out = make_imaging_data_table_def(fptr, detector, 1, ierr);
  if (out==NULL) { /* failed */
    *ierr = 10000;
    kill_imaging_detector_table_def(detector); /* clean up */
    return NULL;
  } /* end error test */

  return out; /* done */
} /* end make_data_table */

void write_data_table (fitsfile *fptr, int nx, int ny, int nreg,
		       imaging_data_table_def* tabledef, 
		       int *ierr)
/*----------------------------------------------------------------------- */
/*   Demo of writing an entry in the data table                           */
/*   Inputs:                                                            : */
/*      fptr      cfitsio pointer to open FITS file                       */
/*      nx        Size of data array in x direction                       */
/*      ny        Size of data array in y direction                       */
/*      nreg      Number of regions                                       */
/*      tabledef  imaging_data table definition structure                 */
/*   Output:                                                              */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*----------------------------------------------------------------------- */
{
  long i, j, irow, ireg, indx;
  imaging_data_row *row; 

  if (*ierr>0) return; /* existing error condition? */

  /* make table_row structure */
  row = make_imaging_data_row (tabledef);

  /* set correlations types */
  for (i=0; i<nreg; i++)
    row->targtype[i] = 2; /* data type */

  /* any looping over rows here */
  for (irow = 1; irow<2; irow++) {
  
    /* fill in data, random values to be sure they get in the right place */
    row->time = 1.234567;
    row->opd[0] = 2.3456;
    row->local_opd[0] = 3.456;
    row->exp = 10.0;
    row->offset[0] = 1.0;
    row->offset[1] = 2.0;
    row->rotation = 3.0;
    row->step = 5;
    row->reference = -5;

    for (ireg=0; ireg<row->nregion; ireg++) {
      row->targtype[ireg] = 20+ireg;
      row->source[ireg] = 10+ireg;
      indx = 0; /* index into data array */
      for (j = 0; j<ny; j++) { /* loop over rows */
	for (i = 0; i<nx; i++) { /* loop over columns */
	  row->data[ireg][indx++] = (DataType)(i * 3.1415926); /* some number */
	} /* end loop over columns */
      } /* end loop over rows */
    } /* end loop over region */
    
    /* write next row */
     write_imaging_data_table(fptr, tabledef, row, -1, ierr);
     if (*ierr>0) { /* error test */
       fprintf(stderr,"ERROR %d writing IMAGING_DATA_TABLE\n", *ierr);
       return;
     }

  } /* end loop over rows */
  
  /* delete row structure */
  kill_imaging_data_row(row);

} /* end write_data_table */

/* */
