/*
 * This file is part of MIA.
 *
 * MIA is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/* $Id: fringe.c,v 1.4 2002/06/04 09:29:29 jdejong Exp $ */
#define _POSIX_SOURCE_1
#include "vltPort.h"

/*-----------------------------------------------------------------------*/
/* Reads simulated midi-like data from a FITS binary table               */
/* Test Fringe finder for simulated Midi raw data                        */
/* Currently assumes a positive bias put into OPD delay                  */
/* This version can use a steady variation of the OPD.                   */
/*  Command line parameters:                                             */
/*   -in    Name of FITS file containing the data ,def simul.fit         */
/*   -out   Name of FITS output file with solutions, none=>no output     */
/*   -pix   Name of FITS image cube of delay-rate, none=>DRpix.fit       */
/*   -SI    Solution interval of fit in samples  (def. 10)               */
/*   -narrow use narrow search window(false)                             */
/*-----------------------------------------------------------------------*/
#include <math.h>
#include <malloc.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "imaging_data.h"
#include "cal_table.h"
#include "fourg.h"
#include "dft.h"
#include "pfit.h"
#include "interpol.h"
#include "image.h"
#include "history.h"

/* global variables, possibly read from command line */
char     *infile="simul.fit"; /* input FITS file name */
char     *pixfile="!DRpix.fit";     /* delay-rate image file */
char     *outfile="!xxx.fit";       /* output FITS file name */
int      SI = 10;             /* Solution interval in samples. */
int      narrow=0;            /* if true use narrow search window */

/* cfitsio pointer for input file */
fitsfile *inputfptr;

/* global parameters */
#define T_FILTER_WIDTH 0 /* width in pixels of time notch filter, 0=only zero*/
#define F_FILTER_WIDTH 3 /* width in pixels of freq notch filter*/

/* work arrays */
static float FTwork[4096], FTwork1[4096], FTwork2[4096];

/* sine/cosine etc arrays for time DFT. */
DFT_info **DFTinfo_array = NULL;

/* image_info for delay-rate image output */
image_info *DRimage = NULL;
float      *DRplane = NULL;  /* work array for delay rate plane */

#define CLIGHT 2.997924562e8 /* speed of light */
#define TWOPI 6.283185308    /* two pi */
#define FRACT 0.35      /* Fraction of the photons that go into each cross correlation */
#define COHER 0.60      /* Instrumental coherence factor (poln mismatch)*/
#define COUNTS 2.564e4  /* counts per second per mJy */

/* internal prototypes */
void fringein (int argc, char **argv, int *ierr);
void Usage(void);
void read_data_table (fitsfile *fptr, imaging_data_table_def* tabledef, 
		       int *ierr);
void fringe_fit_data (float **spectrum, float **big, 
		      int nx, int nt, float time_incr, 
		      float *OPD_off, float *times, 
		      imaging_detector_table_def* detector,
		      cal_table_row *cal_row);
void find_peak (float **array, int first_delay, int last_delay, int nd, 
		 int first_rate, int last_rate, int nr, float peak_loc[2], 
		float *rms, float *peak);
void notch_filter (float **array, int nf, int nt, float **out);
void FT_old_time (float **array, int nf, int nt); /* debug routine */
void FT_in_time (float **array, int nf, int nt);
void init_FT_in_time (int nf, int nt, float delta_t,
		      imaging_detector_table_def* detector);
void kill_FT_in_time (int nf);
void FT_in_freq (float **array, int nf, int nt);
void correct_delay (float **array, int nf, int nt, float *OPD_off, 
		    imaging_detector_table_def* detector);
void avg_in_time (float **array, int nf, int nt);
void avg_in_freq (float **array, int nf, float vis[2]);
void demodulate (float **array, int nf, int nt, float *OPD_off, 
		 float target, imaging_detector_table_def* detector);
void print_array (int type, float **array, int first_delay, int last_delay, 
		 int first_rate, int last_rate, char *title);
fitsfile* init_caltable (imaging_data_table_def* tabledef,
			 cal_table_def **table, 
			 cal_table_row *cal_rows[2], int *ierr);
void write_caltable(fitsfile *outfptr, cal_table_def* table,
		    double time, float timei, 
		    cal_table_row *cal_rows[2], int *ierr);
void close_caltable(fitsfile *outfptr, cal_table_def* table,
		    cal_table_row *cal_rows[2], int *ierr);
image_info* init_DRpix(int nd, int nr, float del_d, float del_r, 
		       float del_t, char *obs_date, int *ierr);
void write_DRpix(image_info *image, int nd, int nr, float **array, 
		 float rms, int *ierr);
void close_DRpix(image_info *image, int *ierr);
void history (fitsfile *fptr, int *ierr);


int main ( int argc, char **argv )
/*----------------------------------------------------------------------- */
/*   Program to fringe fit simulated midi data                            */
/*----------------------------------------------------------------------- */
{
  int  ierr;
  imaging_data_table_def* tabledef;
  fitsfile *infptr=NULL;

  /* Startup - parse command line */
  ierr = 0;
  fringein (argc, argv, &ierr);
  if (ierr!=0) return ierr;

  /* Open FITS input file */
  fits_open_file(&infptr, infile, READONLY, &ierr);
  if (ierr>0) { /* error check */
    fprintf(stderr,"ERROR %d opening input FITS file %s \n", &ierr, infile);
    table_print_error_stack();  /* Show any cfitsio error messages */
    return ierr;
  }/* end error test */

  /* Get basic table structures and read IMAGING_DETECTOR table */
  tabledef = open_imaging_data_table(infptr, &ierr);
  if ((tabledef==NULL) || (ierr>0)) { /* error test */
    fprintf (stderr,"error %d after table create\n",ierr);
    table_print_error_stack();  /* Show any cfitsio error messages */
    return ierr;
  } /* end error test */
  
  /* Read table  fitting data*/
  read_data_table(infptr, tabledef, &ierr);
  if (ierr>0) { /* error test */
    fprintf (stderr,"error %d after table read\n",ierr);
    table_print_error_stack();  /* Show any cfitsio error messages */
    return ierr;
  } /* end error test */
  
  /* Close FITS file */
  fits_close_file (infptr, &ierr);

   /* Show any cfitsio error messages */
  table_print_error_stack(); 

  return ierr;
} /* end of main */

void fringein (int argc, char **argv, int *ierr)
/*----------------------------------------------------------------------- */
/*  Parse control info from command line                                  */
/*   Input:                                                               */
/*      argc   Number of arguments from command line                      */
/*      argv   Array of strings from command line                         */
/*   Output:                                                              */
/*      ierr       Error code: 0 => ok                                    */
/*   Global values                                                        */
/*      infile;     input FITS file name                                  */
/*      SI          Solution interval in sec.                             */
/*----------------------------------------------------------------------- */
{
  int ax;
  char *arg;

  /*debug    fprintf(stderr, "number of arguments = %d\n",argc);*/
  if (argc<=1) Usage(); /* must have arguments */
  /* parse command line */
  for (ax=1; ax<argc; ax++) {
    arg = argv[ax];
    /*debug    fprintf(stderr, "arg = %s\n",arg);*/
    if (strcmp(arg, "-in") == 0) {
      infile = argv[++ax];
    } else if (strcmp(arg, "-out") == 0) {
      outfile = argv[++ax];
    } else if (strcmp(arg, "-pix") == 0) {
      pixfile = argv[++ax];
    } else if (strcmp(arg, "-SI") == 0){    /* solution interval */
      sscanf (argv[++ax],"%d",&SI);
    } else if (strcmp(arg, "-narrow") == 0){    /* narrow search */
      narrow = 1;
    } else { /* unknown argument */
      Usage();
    }
  }
  *ierr = 0;
} /* end fringein */

void Usage(void)
/*----------------------------------------------------------------------- */
/*   Tells about usage of program and bails out                           */
/*----------------------------------------------------------------------- */
{
    fprintf(stderr, "Usage: fringe -in file -out file -pix file -SI n -narrow\n");
    fprintf(stderr, "Simulated midi data as VLTI binary tables\n");
    fprintf(stderr, "Arguments:\n");
    fprintf(stderr, "  -in   Name of FITS file to contain the data, def simul.fit\n");
    fprintf(stderr, "  -out  Name of FITS output file with solutions, none=>no output\n");
    fprintf(stderr, "  -pix  Name of FITS image cube of delay-rate, none=>no output\n");
    fprintf(stderr, "  -SI   Solution interval in samples (def. 10)\n");
    fprintf(stderr, "  -narrow use narrow search window(def false)\n");
     /*/exit(1);  bail out */
  }/* end Usage */

void read_data_table (fitsfile *fptr, imaging_data_table_def* tabledef, 
		       int *ierr)
/*----------------------------------------------------------------------- */
/*   Loop reading table, fitting fringes, writing solutions if requested  */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*      tabledef  imaging_data table definition structure                 */
/*   Output:                                                              */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*----------------------------------------------------------------------- */
{
  long i, indx, j, irow, nreg, step, number, nx, integ;
  double  time, taudif, sumtau=0.0, sumd=0.0, sumd2=0.0, cnttau=0.0, cntd2=0.0;
  imaging_data_row *row; 
  float **spectrum, **big, *OPD_off, *times, timei, fract_corr=-1.0;
  cal_table_def  *caltable=NULL;
  cal_table_row  *cal_rows[2] = {NULL, NULL};
  fitsfile *outfptr = NULL;

  if (*ierr>0) return; /* existing error condition? */
  if (tabledef==NULL) {*ierr=10000; return;} /* check pointer */

  inputfptr = fptr; /* save cfitsio pointer in global variable */

  /* set up output if necessary */
  outfptr = init_caltable(tabledef, &caltable, cal_rows, ierr);
  if (*ierr>0) { /* error test */
    fprintf(stderr,"ERROR %d writing CALIBRATION_TABLE\n", *ierr);
    return;
  }

  nx  = tabledef->detector->naxis[0][0]; /* number of pixels in Freq */
  if (nx<5) { /* error test */
    fprintf(stderr,"ERROR: too few (%d) frequency channels\n", nx);
    return;
  }

  /* require at least 5 channels */

  /* allocate 2D data arrays */
  spectrum = (float**) malloc(SI*sizeof(float*));
  for (i=0; i<SI; i++)
    spectrum[i] = (float*) malloc(2*nx*sizeof(float));
  /* now one twice as large*/
  big = (float**) malloc(2*SI*sizeof(float*));
  for (i=0; i<2*SI; i++) {
    big[i] = (float*) malloc(4*nx*sizeof(float));
  }

  OPD_off = (float*) malloc(SI*sizeof(float));
  times = (float*) malloc(SI*sizeof(float));

  nreg = tabledef->nregion;        /* number of regions */

  row = make_imaging_data_row (tabledef);

  /* get size of table */
  number = 0;
  fits_get_num_rows (fptr, &number, ierr);

  /* looping over rows*/
  integ = 0; indx = 0; time = 0.0; timei = 0.0;
  for (irow = 1; irow<=number; irow++) {
  
    /* read next row */
     read_imaging_data_table(fptr, tabledef, row, -1, ierr);
     if (*ierr>0) { /* error test */
       fprintf(stderr,"ERROR %d reading IMAGING_DATA_TABLE\n", *ierr);
       return;
     }

     /* average delay passed by simul as row->opd */
     if (row->opd!=0.0) {
       sumtau += row->opd;
       cnttau += 1.0;
     }

     /* correction for how the light is divided amoung spectra, etc */
     /* try to calibrate in mJy */
     if (fract_corr<0.0) 
       fract_corr = ((float)nx / (2.0 * FRACT * COHER)) /
	 (COUNTS * row->exp);
     
     /* copy data to work array */
     OPD_off[integ] = row->local_opd; /* modulation delay */
     times[integ] = row->time;        /* integration time */
     indx = 0;
     for (i=0; i<nx; i++){
       /* real part - difference two spectra */
       spectrum[integ][indx++] = 
	 fract_corr * (row->data[0][i] - row->data[1][i]); 
       spectrum[integ][indx++] = 0.0;   /* imaginary part - not measured */
     }
     integ++;

     /* sum time for average */
     time  += row->time;
     timei += row->exp;

     if (integ>=SI) { /* time for fit? */
       /* debug */
       /*fprintf (stdout, " record %d ",irow);*/
       
       /* fit data */
       /* get times in seconds relative to reference */
       if (integ>0) time = time / integ; /* average time */
       for (i=0; i<integ; i++) times[i] = (times[i] - time) * 86400.0;

       cal_rows[1]->time = time;
       fringe_fit_data (spectrum, big, nx, integ, row->exp, OPD_off,
			times, tabledef->detector, cal_rows[1]);

       /* write solution if necessary */
       write_caltable(outfptr, caltable, time, timei, cal_rows, ierr);
       if (*ierr>0) { /* error test */
	 fprintf(stderr,"ERROR %d writing CALIBRATION_TABLE\n", *ierr);
	 return;
       }
       time = 0.0; timei = 0.0;
       integ = 0; indx = 0;

       /* Compare model passed by simul with fitted */
       if (cnttau>0.0) {
	 sumtau  = sumtau / cnttau;
	 taudif = sumtau - (cal_rows[1]->delay * 1.0E6 * CLIGHT);
	 /* ignore really bad fits */
	 if (abs(taudif)<15.0) {
	   sumd  += taudif;
	   sumd2 += taudif*taudif;
	   cntd2 += 1.0;
	 }
	 sumtau = 0.0; cnttau = 0.0;
       } /* end compare model */
     } /* end of do fit */
    
  } /* end loop over rows */
  
  /* fit any remaining data  */
  /* need to redo time FT structures for this orphan data */
  /*if (integ>0) 
    fringe_fit_data (spectrum, big, nx, integ, row->exp,  OPD_off,
    times, tabledef->detector);*/

  /* Give RMS delay error */
  if (cntd2>0.0) {
    sumd = sumd / cntd2;
    sumd2  = sqrt (sumd2 / cntd2);
    fprintf (stdout, "RMS delay error = %lf microns, avg = %lf microns\n",
	     sumd2, sumd);
  } /* end compare model */

  /* shutdown output caltable if necessary */
  close_caltable(outfptr, caltable, cal_rows, ierr);
  if (*ierr>0) { /* error test */
    fprintf(stderr,"ERROR %d closinging CALIBRATION_TABLE\n", *ierr);
    return;
  }

  /* shutdown output delay-rate image if necessary */
  close_DRpix(DRimage, ierr);
  if (*ierr>0) { /* error test */
    fprintf(stderr,"ERROR %d closinging delay-rate image\n", *ierr);
    return;
  }
  kill_image_info (DRimage); /* delete global image_info structure */

  /* delete row structure */
  kill_imaging_data_row(row);

  /* delete time FFT structures */
  kill_FT_in_time (nx);

  /* delete temporary arrays */
  if (spectrum!=NULL) {
    for (i=0; i<SI; i++)
      if (spectrum[i]!=NULL)free(spectrum[i]);
    free(spectrum);
  }
  if (big!=NULL) {
    for (i=0; i<2*SI; i++)
      if (big[i]!=NULL) free(big[i]);
    free(big);
  }
} /* end read_data_table */

void fringe_fit_data (float **spectrum, float **big, 
		      int nf, int nt, float time_incr,
		      float *OPD_off, float *times, 
		      imaging_detector_table_def* detector,
		      cal_table_row *cal_row)
/*----------------------------------------------------------------------- */
/*   Fit simulated telescope data                                         */
/*   Also writes delay-rate images to FITS cube if requested              */
/*   Steps:                                                               */
/*     1) apply 2D notch filter to suppress non source signals            */
/*        Also zero pad array to big for better resolution                */
/*     2) remove instrumental OPD                                         */
/*     3) FFT in time direction                                           */
/*     4) FT in frequency direction                                       */
/*     5) Estimate location of the peak                                   */
/*   Inputs:                                                              */
/*      spectrum  Measured spectrum  (nf x nt) array of detector values   */
/*      big       Double size work array for zero padded fft              */
/*      nf        Number of frequency channels in spectrum                */
/*      nt        Number of times                                         */
/*      time_incr Time step between samples in sec                        */
/*      OPD_off   OPD offset applied (seconds) per sample                 */
/*      times     Times relative to reference time (sec)                  */
/*      detector  imaging_detector table definition structure             */
/*      cal_row   calibration table structure for result, ignore if NULL  */
/*                filled in on output                                     */
/*----------------------------------------------------------------------- */
{
  long i, j, ii, jj, first_delay, first_rate, last_delay, last_rate;
  double freq, freq0, freq_incr, freq1, freq2, lo_freq, delay, rate;
  float  ref_pix, noise, background, value, sum, target, amp;
  float  peak_val[3], temp, peak[2], rms, faz, vis[2], del_d, del_r;
  int    mf, mt, jerr;

  /* setup for Time DFTs if not done */
  if (DFTinfo_array == NULL)  
    init_FT_in_time (nf*2, nt*2, time_incr, detector);

  /* get info from structures */
  freq0     = detector->crval[0][0]; /* frequency at reference pixel */
  ref_pix   = detector->crpix[0][0]; /* reference pixel 1-rel */
  freq_incr = detector->cd[0][0];    /* frequency spacing */

  /* find low frequency */
  freq1 = freq0 + (0-ref_pix)*freq_incr;
  freq2 = freq0 + (nf-1-ref_pix)*freq_incr;
  if (freq1>freq2) lo_freq = freq2;
  else  lo_freq = freq1;

  /* notch filter to kill most non source signals in image */
  /* also zero pads to big for better dalay, rate resolution */
  notch_filter (spectrum, nf, nt, big);
  mf = nf * 2;
  mt = nt * 2;

  /* demodulate OPD */
  /* target = 100 microns in sec */
  target = 100.0 / (CLIGHT*1.0e6) ;
  demodulate(big, mf, mt, OPD_off, target, detector);
  
  /* FFT in time */
  if (nt>1) FT_in_time (big, mf, mt);
  /*debug if (nt>1) FT_old_time (big, nf, nt);*/

  /* FFT in freq */
  FT_in_freq (big, mf, mt);

  /* find peak */
  /* always ignore junk around and below zero delay */
  first_delay = (mf / 2) + 3;

  if (narrow) { /* narrow search */
    last_delay = mf/2 + mf/8; 
    first_rate = mt/2 - mt/8;
    last_rate = mt/2 + mt/8;
  } else { /* wide open search */
    last_delay = mf; 
    first_rate  = 0;
    last_rate = mt;
  } 
  if (first_delay>=mf) first_delay = mf-1;
  if (first_rate>=mt) first_rate = mt-1;
  if (first_rate<0) first_rate = 0;
  if (last_rate>mt) last_rate = mt;

  find_peak (big, first_delay, last_delay, mf, 
	     first_rate, last_rate, mt,
	     peak, &rms, peak_val);

  /* convert to delay in microns */
  delay =  CLIGHT*1.0e6 * (peak[0] - mf/2) / (mf*freq_incr);

  /* rate channels all set to lowest frequency - convert to micron/s */
  rate = (CLIGHT /lo_freq)  * 1.0e6 * (peak[1] - mt/2) / (mt * time_incr);

  /* save result if necessary */
  if (cal_row!=NULL) {
    /* convert results to seconds from microns */
    delay *= 1.0e-6 / CLIGHT;
    rate  *= 1.0e-6 / CLIGHT;
    amp = sqrt (peak_val[1]*peak_val[1] + peak_val[2]*peak_val[2]);
    if (amp<1.0e-5) amp = 1.0;
    cal_row->gain[0] = peak_val[1] / amp;
    cal_row->gain[1] = peak_val[2] / amp;
    cal_row->delay   = delay;
    cal_row->rate    = rate;
    cal_row->weight  = peak_val[0]/rms;
  } /* end save value */

  /* write delay - rate image if requested */
  /* initialize delay-rate image file if necessary */
  jerr = 0;
  if ((pixfile!=NULL) && (DRimage==NULL)) {
    del_d = CLIGHT*1.0e6  / (mf*freq_incr);
    del_r = (CLIGHT /lo_freq)  * 1.0e6 / (mt * time_incr);
    DRimage = init_DRpix (mf, mt, del_d, del_r, 
			  time_incr, detector->date_obs, &jerr);
    /* if there's a problem - turn off writing */
    if ((jerr!=0) || (DRimage==NULL)) {
      pixfile = NULL;
      DRimage = NULL;
    }
  }

  if (pixfile) {
    write_DRpix(DRimage, mf, mt, big, rms, &jerr);
    /* if there's a problem - turn off writing */
    if (jerr!=0) {
      pixfile = NULL;
      DRimage = NULL;
    }
  }

  /* now use this solution to correct and average the data to get phase */
  /* correct delays */
  for (i=0; i<nt; i++) OPD_off[i] += delay + (times[i] * rate);

  correct_delay(spectrum, nf, nt, OPD_off, detector);
  
  /* average in time */
  avg_in_time (spectrum, nf, nt);

  /* average in freq */
  avg_in_freq (spectrum, nf, vis);

  /* save phase (flip sign to make a correction) */
  amp = sqrt (vis[0]*vis[0] + vis[1]*vis[1]);
  if (amp<1.0e-5) amp = 1.0;
  cal_row->gain[0] =  vis[0] / amp;
  cal_row->gain[1] = -vis[1] / amp;

  /* phase */
  faz = 57.296 * atan2(vis[1], vis[0]);

  /* print result */
  delay /= 1.0e-6 / CLIGHT;
  rate  /= 1.0e-6 / CLIGHT;
  fprintf (stdout, "time = %lf  peak =  %f %f, amp = %f snr= %f  delay = %lg rate = %lg ph = %f\n",
	   cal_row->time*86400.0, peak[0], peak[1], peak_val[0], peak_val[0]/rms, delay, rate, faz);


} /* end fringe_fit_data */

void find_peak (float **array, int first_delay, int last_delay, int nd, 
		int first_rate, int last_rate, int nr, float peak_loc[2], 
		float *rms, float *peak)
/*----------------------------------------------------------------------- */
/*   Find the maximum magnitude element in a nd x nr complex array        */
/*   The returned position is the local centroid                          */
/*   Inputs:                                                              */
/*      array       Complex array  as (real, imag)                        */
/*      first_delay First delay complex element to consider               */
/*      last_delay  Last delay complex element to consider                */
/*      nd          dimension in delay                                    */
/*      first_rate  First rate complex element to consider                */
/*      last_rate   last rate complex element to consider                 */
/*      nr          dimension in rate                                     */
/*   Output:                                                              */
/*      peak_loc    Location of peak (pixels in delay, rate)              */
/*      rms         RMS of region far from zero delay, rate               */
/*      peak        magnitude, real, imag interpolated at peak            */
/*----------------------------------------------------------------------- */
{
  float val, peak_pos, sumv, sumd, sumr, norm, count=0.0, max_val=-1.0;
  float darray[9][9], sarray[9][9],  carray[9][9], fblank=-1.2345e5;
  float x, y, fit_max;
  double sum = 0;
  int   id, ir, iid, i, i1, i2, j1, j2,  max_delay=-1, max_rate=-1;
  int   jr, jd, rmin, rmax;
  int   realOnly=0; /* do search with real only or amplitude */

  /* debug - print search window*/
  rmin = nr/2 - 5;
  if (rmin<0) rmin = 0;
  rmax = nr/2 + 5;
  if (rmax>nr) rmax = nr;
  /*
  print_array(1, array, nd/2, nd/2+20, rmin, rmax, 
    "Array in Find_peak");
  print_array(2, array, nd/2, nd/2+20, rmin, rmax, 
    "Array in Find_peak");
  */

  /* double loop over allowed portion of the array */
  /* getting sum for RMS and finding peak in subregion */
  for (ir=first_rate; ir<nr; ir++) {
    for (id=first_delay; id<nd; id++) {
      iid = id*2;
      /* index into real array */
      val = array[ir][iid]*array[ir][iid] + 
	    array[ir][iid+1]*array[ir][iid+1];
      sum += val; count++;
      /* search real part? */
      if (realOnly) val =  array[ir][iid];
      /* don't search everywhere */
      if ((id<=last_delay) && (ir<=last_rate)) {
	if (val>max_val) { /* new maximum */
	  max_val = val;
	  max_delay = id;
	  max_rate = ir;
	}
      }
    } /* end loop over delay */
  } /* end loop over rate */

  /* blank arrays */
  for (ir=0; ir<9; ir++) {
    for (id=0; id<9; id++) {
      darray[ir][id] = fblank;
      sarray[ir][id] = fblank;
      carray[ir][id] = fblank;
    }
  }
  /* find local centroid for better estimate */
  i1 = max_delay-4; if (i1<first_delay) i1 = first_delay;
  i2 = max_delay+4; if (i2>=nd) i2 = nd-1;
  j1 = max_rate-4; if (j1<first_rate) j1 = first_rate;
  j2 = max_rate+4; if (j2>=nr) j2 = nr-1;

  /* double loop over allowed portion of the array */
  /* copy amplitudes to work array */
  for (ir=j1; ir<=j2; ir++) {
    for (id=i1; id<=i2; id++) {
      iid = id*2;
      /* index into real array */
      if (!realOnly) { /* amplitude search */
	val = array[ir][iid]*array[ir][iid] + 
	  array[ir][iid+1]*array[ir][iid+1];
	val = sqrt(val);
      } else { /* only real part */
	val = array[ir][iid];
      }
      /* save value */
      jr = ir - max_rate + 4;
      jd = id - max_delay + 4;
      carray[jr][jd] = array[ir][iid];
      sarray[jr][jd] = array[ir][iid+1];
      darray[jr][jd] = val;
      /* delete from rms sum */
      sum -= val*val; count--;
    } /* end loop over delay */
  } /* end loop over rate */

  /* get better positions, peak */
  if (pfit (darray, &fit_max, peak_loc, fblank)) {
    /* fit failed, use peak*/
    peak_loc[0] = (float)max_delay;
    peak_loc[1] = (float)max_rate;
  } else { /* add offsets */
     peak_loc[0] += (float)max_delay;
     peak_loc[1] += (float)max_rate;
  }

  /* normalization factor for FT */
  /* not sure why 8.0 is necessary */
  norm = 8.0 / ((float)nd * (float)nr);
  /* compute RMS */
  if (count>2.0)
    *rms = sqrt(sum/count) * norm;
  else
    *rms = sqrt(max_val) * norm;

  /* finished */
  /*peak[0] = sqrt(max_val) * norm;*/
  /* amplitude */
  peak[0] = fit_max * norm;

  /* real and imaginary */
  x = peak_loc[0] - max_delay + 4;
  y = peak_loc[1] - max_rate + 4;
  peak[1] =  imintp (&carray[0][0], 9, 9, 3, x, y, fblank) * norm;
  peak[2] =  imintp (&sarray[0][0], 9, 9, 3, x, y, fblank) * norm;

  return;
} /* end find_peak */

void notch_filter (float **array, int nf, int nt, float **out)
/*----------------------------------------------------------------------- */
/*   Apply notch filter to a frequency x time array                       */
/*   Uses global parameters:                                              */
/*      T_FILTER_WIDTH  width in pixels of time notch filter, 0=only zero */
/*      F_FILTER_WIDTH  width in pixels of freq notch filter, 0=only zero */
/*   Inputs:                                                              */
/*      array     Measured spectrum  (nf x nt) array of detector values   */
/*      nf        Dimension of spectrum                                   */
/*      nt        Number of times                                         */
/*   Output:                                                              */
/*      array     Input array filtered                                    */
/*      out       Input array filtered and zero padded                    */
/*----------------------------------------------------------------------- */
{
  int   ifreq, itime, j, jfreq, ii;
  float scale;

  /* zero fill out */
  for (itime=0; itime<2*nt; itime++) {
    for (ifreq=0; ifreq<4*nf; ifreq++) {
      out[itime][ifreq] = 0.0;
    }
  }

  /* first filtering in time/rate */
  /* loop over frequency */
  for (ifreq = 0; ifreq<nf; ifreq++) {
    jfreq = 2 * ifreq; /* frequency index */
    /* collect data into FTwork1 */
    ii = 0;
    for (itime=0; itime<nt; itime++) { 
      FTwork1[ii++] = array[itime][jfreq];
      FTwork1[ii++] = array[itime][jfreq+1];
    }

    /* do the forward FFT */
    fourg (FTwork1, nt, 1, FTwork2);

    /* do the filtering */
    /* Note: fourg leaves the zero cell in the first element */
    FTwork1[0] = 0.0; FTwork1[1] = 0.0; 
    for (j=1; j<=T_FILTER_WIDTH/2; j++) {
      ii = j * 2;
      FTwork1[ii] = 0.0; FTwork1[ii+1] = 0.0; 
      ii = (nt - j) * 2;
      FTwork1[ii] = 0.0; FTwork1[ii+1] = 0.0; 
      }

    /* do the reverse FFT */
    fourg (FTwork1, nt, -1, FTwork2);

    /* copy back to array, renormalize for effects of fourg */
    ii = 0;
    scale = 1.0 / (float)nt;
    for (itime=0; itime<nt; itime++) { 
      array[itime][jfreq]   =  scale * FTwork1[ii++];
      array[itime][jfreq+1] =  scale * FTwork1[ii++];
    }
  } /* end loop over frequency */

  /* Then filtering in freq/delay */
  /* loop over time */
  for (itime=0; itime<nt; itime++) {

    /* do the forward FFT */
    fourg (array[itime], nf, 1, FTwork2);

    /* do the filtering */
    /* Note: fourg leaves the zero cell in the first element */
    array[itime][0] = 0.0; array[itime][1] = 0.0; 
    for (j=1; j<=F_FILTER_WIDTH/2; j++) {
      ii = j * 2;
      array[itime][ii] = 0.0; array[itime][ii+1] = 0.0; 
      ii = (nf - j) * 2;
      array[itime][ii] = 0.0; array[itime][ii+1] = 0.0; 
      }

    /* do the reverse FFT */
    fourg (array[itime], nf, -1, FTwork2);

    /* renormalize for effects of fourg */
    ii = 0;
    scale = 1.0 / (float)nf;
    for (j=0; j<nf; j++) {
      ii = j * 2;
      array[itime][ii]   =  scale * array[itime][ii];
      array[itime][ii+1] =  scale * array[itime][ii+1];
      /* make copy in out */
      out[itime][ii]   =  array[itime][ii];
      out[itime][ii+1] =  array[itime][ii+1];
    }
  } /* end loop over time */
  /* debug - print array */
  /*print_array(2, array, 0, nf, 0, nt, 
    "Array after notch_filter");*/
} /* end notch_filter */

void FT_old_time (float **array, int nf, int nt)
/*----------------------------------------------------------------------- */
/*   Fourier transform a freq x time array in the time direction          */
/*   Inputs:                                                              */
/*      array     Measured spectrum  (nf x nt) array of detector values   */
/*      nf        Dimension of spectrum                                   */
/*      nt        Number of times                                         */
/*   Output:                                                              */
/*      array     Input array fourier transformed in time (2nd axis)      */
/*----------------------------------------------------------------------- */
{
  int   ifreq, itime, j, ii;

  /* loop over frequency */
  for (ifreq = 0; ifreq<nf; ifreq++) {
    j = 2 * ifreq;
    /* collect data into FTwork1 */
    ii = 0;
    for (itime=0; itime<nt; itime++) { 
      FTwork1[ii++] = array[itime][j];
      FTwork1[ii++] = array[itime][j+1];
    }

    /* do the FFT */
    fourg (FTwork1, nt, 1, FTwork2);

    /* rearrange to sensible order (center at center) */
    /* lower half first */
    ii = nt;
    for (itime=0; itime<nt/2; itime++) { 
      array[itime][j]   = FTwork1[ii++];
      array[itime][j+1] = FTwork1[ii++];
    }
    /* then upper */
    ii = 0;
    for (itime=nt/2; itime<nt; itime++) { 
      array[itime][j]   = FTwork1[ii++];
      array[itime][j+1] = FTwork1[ii++];
    }
  } /* end loop in freq slice */
  /* debug - print array */
  /*print_array(array, 0, nf, 0, nt, 
    "Array after FT_old_time");*/
} /* end  FT_old_time */

void FT_in_time (float **array, int nf, int nt)
/*----------------------------------------------------------------------- */
/*   Fourier transform a freq x time array in the time direction          */
/*   Inputs:                                                              */
/*      array     Measured spectrum  (nf x nt) array of detector values   */
/*      nf        Dimension of spectrum (second half expected zero padded)*/
/*      nt        Number of times (second half expected zero padded)      */
/*   Output:                                                              */
/*      array     Input array fourier transformed in time (2nd axis)      */
/*                values are sampled at the fringe rate in Hz of the      */
/*                lowest frequency channel                                */
/*----------------------------------------------------------------------- */
{
  int   ifreq, itime, j, i, ii;
  float ref_pix, du;

  /* loop over frequency */
  for (ifreq = 0; ifreq<nf/2; ifreq++) {
    j = 2 * ifreq;
    /* collect data into FTwork1 */
    ii = 0;
    for (itime=0; itime<nt/2; itime++) { 
      FTwork1[ii++] = array[itime][j];
      FTwork1[ii++] = array[itime][j+1];
    }

    /* do the FFT */
    DFTc2c (DFTinfo_array[ifreq], FTwork1, FTwork2);

    /* copy back (will be twice the size due to zero padding) */
    ii = 0;
    for (itime=0; itime<nt; itime++) { 
      array[itime][j]   = FTwork2[ii++];
      array[itime][j+1] = FTwork2[ii++];
    }
  } /* end loop in freq slice */

  /* debug - print array */
  /*print_array(1, array, 0, nf, 0, nt, 
    "Array after FT_in_time");*/
} /* end  FT_in_time */

void init_FT_in_time (int nf, int nt, float delta_t,
		      imaging_detector_table_def* detector)
/*----------------------------------------------------------------------- */
/*   Set up for time DFTs                                                 */
/*   Creates time DFT structures in global array DFTinfo_array            */
/*   The time FT use a DFT so that the ooutput rate channels in each      */
/*   spectral channel will be the same in s/s rather than Hz.             */
/*   The array to be transformed is expected zero padded in the second    */
/*   half in both time and frequency.                                     */
/*   Inputs:                                                              */
/*      nf        Number of frequency channels                            */
/*      nt        Number of times                                         */
/*      delta_t   Time increment in sec.                                  */
/*      detector  imaging_detector table definition structure             */
/*   Output:                                                              */
/*      values in global array array DFTinfo_array.                       */
/*----------------------------------------------------------------------- */
{
  int   ifreq, itime, j, i, ii;
  DFT_info *info;
  float ref_pix, du;
  double freq, freq0, freq1, freq2, lo_freq, freq_incr;

  /* create array */
  DFTinfo_array = (DFT_info**)malloc((nf/2)*sizeof(DFT_info*));

  /* get info from structures */
  freq0     = detector->crval[0][0]; /* frequency at reference pixel */
  ref_pix   = detector->crpix[0][0]; /* reference pixel 1-rel */
  freq_incr = detector->cd[0][0];    /* frequency spacing */

  /* find low frequency */
  freq1 = freq0 + (0-ref_pix)*freq_incr;
  freq2 = freq0 + (nf-1-ref_pix)*freq_incr;
  if (freq1>freq2) lo_freq = freq2;
  else  lo_freq = freq1;

  /* loop over frequency (second half zero padded so don't need FT */
  for (ifreq = 0; ifreq<nf/2; ifreq++) {

    /* prepare for FT, use fringe frequencies at low frequency end  */
    /* scale so that a given delay rate in s/s will be in the same channel
       inregardless of the frequency */
    freq = freq0 + (ifreq-ref_pix)*freq_incr;
    du = (freq/lo_freq) * (1.0 / (nt*delta_t));
    for (i=0; i<nt; i++) {
      FTwork[i]  = (i-nt/2) * du;
      FTwork2[i] = (i-nt/4) * delta_t;
    }

    /* get sines/cosines (last time half zero padded) */
    DFTinfo_array[ifreq] =  
      makeDFT(BACKWARD, (long)nt/2, FTwork2, (long)nt, FTwork); 

  } /* end loop in freq  */
} /* end  init_FT_in_time */

void kill_FT_in_time (int nf)
/*----------------------------------------------------------------------- */
/*   Delete globay structures  for time DFTs                              */
/*   Inputs:                                                              */
/*      nf        Number of frequency channels, structures assumed for    */
/*                first half only due to zero padding                     */
/*----------------------------------------------------------------------- */
{
  int   ifreq;

  if (DFTinfo_array==NULL) return;

  /* loop over frequency */
  for (ifreq = 0; ifreq<nf/2; ifreq++) killDFT (DFTinfo_array[ifreq]);
  free(DFTinfo_array);  /* get rid of main array */
  DFTinfo_array = NULL; /* just in case */
} /* end  kill_FT_in_time */

void FT_in_freq (float **array, int nf, int nt)
/*----------------------------------------------------------------------- */
/*   Fourier transform a freq x time array in the freq direction          */
/*   Inputs:                                                              */
/*      array     Measured spectrum  (nf x nt) array of detector values   */
/*      nf        Dimension of spectrum                                   */
/*      nt        Number of times                                         */
/*   Output:                                                              */
/*      array     Input array Fourier transfoemed in frequency            */
/*----------------------------------------------------------------------- */
{
  int   ifreq, itime, jfreq, jj, ii;
  float temp;

  /* debug - print search window*/
  /*print_array(array, nf/2, nf, 0, nt, 
    "Array in FT_in_freq before");*/

  /* loop over time */
  for (itime = 0; itime<nt; itime++) {

    /* do the FFT */
    fourg (array[itime], nf, 1, FTwork2);

    /* rearrange to sensible order (center at center) */
    for (ifreq=0; ifreq<nf/2; ifreq++) { 
      jfreq = nf/2 + ifreq;
      ii = ifreq*2;
      jj = jfreq*2;
      /* swap reals */
      temp = array[itime][jj];
      array[itime][jj] = array[itime][ii];
      array[itime][ii] = temp;
      /* swap imaginaries */
      ii++; jj++;
      temp = array[itime][jj];
      array[itime][jj] = array[itime][ii];
      array[itime][ii] = temp;
    } /* end freq loop */
  } /* end loop in time slice */
  /* debug - print search window*/
  /*print_array(array, nf/2-5, nf, 0, nt, 
    "Array in FT_in_freq after");*/

} /* end  FT_in_freq */

void demodulate (float **array, int nf, int nt, float *OPD_off, 
		 float target, imaging_detector_table_def* detector)
/*----------------------------------------------------------------------- */
/*   Remove instrumental OPD modulation by changing it to target          */
/*   Interpolation done using shift theorem and applying a phase ramp     */
/*   to each spectrum.                                                    */
/*   Inputs:                                                              */
/*      array     Measured spectrum  (nf x nt) array of detector values   */
/*      nf        Dimension of spectrum (last half zero padded)           */
/*      nt        Number of times (last half zero padded)                 */
/*      OPD_off   OPD offset applied (seconds)                            */
/*      target    target offset (seconds)                                 */
/*      detector  imaging_detector table definition structure             */
/*   Output:                                                              */
/*      array     Input array with initial OPD modulation removed         */
/*----------------------------------------------------------------------- */
{ 
  int   itime, ifreq, j, ii, jj;
  float temp;
  double freq, freq0, freq_incr, delay, phase_fact;
  float  ref_pix, phase, sp, cp, re, im, sum;

  /* debug */
  /* subtract mean from each spectrum to kill most of background */
  /* only bother with real part 
  for (itime = 0; itime<nt; itime++) {
    sum = 0.0;
    for (j=0; j<nf; j++) sum += array[itime][j*2];
    sum = sum / nf;
    for (j=0; j<nf; j++) array[itime][j*2] -= sum;
    }*/

  /* get info from structures */
  freq0     = detector->crval[0][0]; /* frequency at reference pixel */
  ref_pix   = detector->crpix[0][0]; /* reference pixel 1-rel */
  freq_incr = detector->cd[0][0];    /* frequency spacing */

  /* loop over actual time slice */
  ii = 0;
  for (itime = 0; itime<nt/2; itime++) {
    phase_fact = TWOPI * OPD_off[itime];

    /* loop over frequency applying phase ramp */
    for (ifreq=0; ifreq<nf/2; ifreq++) {
      freq = freq0 + (ifreq-ref_pix)*freq_incr;
      phase = phase_fact * freq;
      sp = sin(phase);
      cp = cos(phase);
      re = array[itime][ifreq*2];
      im = array[itime][ifreq*2+1];
      array[itime][ifreq*2]   = re*cp - im*sp;
      array[itime][ifreq*2+1] = re*sp + im*cp;
    } /* end of loop in freq */
  } /* end loop in time slice */
  /* debug - print search window*/
  /*
  print_array(2, array, 0, 30, 0, nt, 
    "Array after demodulate");
  */
} /* end  demodulate */

void correct_delay (float **array, int nf, int nt, float *OPD_off, 
		 imaging_detector_table_def* detector)
/*----------------------------------------------------------------------- */
/*   Remove delay from data array                                         */
/*   Interpolation done using shift theorem and applying a phase ramp     */
/*   to each spectrum.                                                    */
/*   Inputs:                                                              */
/*      array     Measured spectrum  (nf x nt) array of detector values   */
/*      nf        Dimension of spectrum                                   */
/*      nt        Number of times                                         */
/*      OPD_off   OPD correction applied (seconds)                        */
/*      detector  imaging_detector table definition structure             */
/*   Output:                                                              */
/*      array     Input array with initial OPD modulation removed         */
/*----------------------------------------------------------------------- */
{ 
  int   itime, ifreq, j, ii, jj;
  float temp;
  double freq, freq0, freq_incr, delay, phase_fact;
  float  ref_pix, phase, sp, cp, re, im, sum;

  /* get info from structures */
  freq0     = detector->crval[0][0]; /* frequency at reference pixel */
  ref_pix   = detector->crpix[0][0]; /* reference pixel 1-rel */
  freq_incr = detector->cd[0][0];    /* frequency spacing */

  /* debug - print array */
  /*
  print_array(1, array, 0, 30, 0, nt, 
	      "Array before correct_delay");
  print_array(2, array, 0, 30, 0, nt, 
	      "Array before correct_delay");
  */

  /* loop over actual time slice */
  ii = 0;
  for (itime = 0; itime<nt; itime++) {
    phase_fact = TWOPI * OPD_off[itime];

    /* loop over frequency applying phase ramp */
    for (ifreq=0; ifreq<nf; ifreq++) {
      freq = freq0 + (ifreq-ref_pix)*freq_incr;
      phase = phase_fact * freq;
      sp = sin(phase);
      cp = cos(phase);
      re = array[itime][ifreq*2];
      im = array[itime][ifreq*2+1];
      array[itime][ifreq*2]   = re*cp - im*sp;
      array[itime][ifreq*2+1] = re*sp + im*cp;
    } /* end of loop in freq */
  } /* end loop in time slice */
  /* debug - print array */
  /*
  print_array(1, array, 0, 60, 0, nt, 
	      "Array after correct_delay");
  print_array(2, array, 0, 60, 0, nt, 
	      "Array after correct_delay");
  */

} /* end  correct_delay */

void avg_in_time (float **array, int nf, int nt)
/*----------------------------------------------------------------------- */
/*   Doaverage of data in time                                            */
/*   Result multiplied by 2.0 -(is this) to correct for having only reals */
/*   Inputs:                                                              */
/*      array     Measured spectrum  (nf x nt) array of detector values   */
/*      nf        Dimension of spectrum                                   */
/*      nt        Number of times                                         */
/*   Output:                                                              */
/*      array     Data averaged in first time slot                        */
/*----------------------------------------------------------------------- */
{
  int   ifreq, itime, j, ii;
  float sumwt, sumre, sumim, isumwt;

  sumwt = (float)nt;    /* all considered good */
  isumwt = 2.0 / sumwt;

  /* loop over frequency */
  for (ifreq = 0; ifreq<nf; ifreq++) {
    j = 2 * ifreq;

    /* sum data */
    sumre = 0.0; sumim = 0.0;
   for (itime=0; itime<nt; itime++) { 
      sumre += array[itime][j];
      sumim += array[itime][j+1];
    }
    /* save in first time slot */
    array[0][j]   = sumre * isumwt;
    array[0][j+1] = sumim * isumwt;

  } /* end loop in freq  */
  /* debug - print array */
  /*
  print_array(1, array, 0, 30, 0, 10, 
      "Array after avg_in_time");
  print_array(2, array, 0, 30, 0, 10, 
  "Array after avg_in_time");
  */
} /* end  avg_in_time */

void avg_in_freq (float **array, int nf,  float vis[2])
/*----------------------------------------------------------------------- */
/*   Do average of all data in freq                                       */
/*   Inputs:                                                              */
/*      array     Measured spectrum  (nf x nt) array of detector values   */
/*                data in first time element                              */
/*      nf        Dimension of spectrum                                   */
/*   Output:                                                              */
/*      vis       Data averaged, as (re, im)                              */
/*----------------------------------------------------------------------- */
{
  int   i, j, ii, indx, jndx, n;
  float sumwt, sumre, sumim;

  indx = 0; jndx = 0; 

  /* sum data */
  sumwt = (float)nf; 
  sumre = 0.0; sumim = 0.0;
  for (i=0; i<nf; i++) { 
    j = 2 * i;
    sumre += array[0][j];
    sumim += array[0][j+1];
  }
  
  /* average */
  if (sumwt>0.0) {
    vis[0]  = sumre / sumwt;
    vis[1]  = sumim / sumwt;
  } else { 
    vis[0]  = 1.0;
    vis[1]  = 0.0;
  }
} /* end  avg_in_freq */

void print_array (int type, float **array, int first_delay, int last_delay, 
		 int first_rate, int last_rate, char *title)
/*----------------------------------------------------------------------- */
/*   Print selected window of  in 2D (d x r)  complex array               */
/*   Inputs:                                                              */
/*      type        1 = amplitude, 2 = phase                              */
/*      array       Complex array  as (real, imag)                        */
/*      first_delay First delay complex element to print                  */
/*      last_delay  Last  delay complex element to print                  */
/*      first_rate  First rate complex element to print                   */
/*      last_rate   Last rate complex element to print                    */
/*      title       title for display                                     */
/*----------------------------------------------------------------------- */
{
  float val;
  int   id, ir, iid;


  if (type==1) { /* amplitude */
    fprintf (stdout,"Amplitudes x 10**-3:  %s \n",title);
    for (id=first_delay; id<last_delay; id++) {
      fprintf (stdout,"  if=%3d ",id);
      iid = id*2;
      for (ir=first_rate; ir<last_rate; ir++) {
	/* index into real array */
	val = array[ir][iid]*array[ir][iid] + 
	  array[ir][iid+1]*array[ir][iid+1];
	val = 0.001*sqrt(val);
	fprintf (stdout," %8.1f", val);
      }
      fprintf (stdout,"\n");
    }
  } else { /* phase */
    fprintf (stdout,"Phases:  %s \n",title);
    for (id=first_delay; id<last_delay; id++) {
      fprintf (stdout,"  if=%3d ",id);
      iid = id*2;
      for (ir=first_rate; ir<last_rate; ir++) {
	/* index into real array */
	val = 57.296 * atan2(array[ir][iid+1], array[ir][iid]);
	fprintf (stdout," %8.1f", val);
      }
      fprintf (stdout,"\n");
    }
  }
    fprintf (stdout,"\n\n");
} /* end print array */


fitsfile* init_caltable (imaging_data_table_def* tabledef,
			 cal_table_def **table, cal_table_row *cal_rows[2], 
			  int *ierr)
/*----------------------------------------------------------------------- */
/*   If global variable outfile is set then it is used as the output FITS */
/*   file name, otherwise no file is written.                             */
/*      If output is desired, the file is created/opened and the          */
/*   cal_table row structures are initialized                             */
/*   Copies old, writes new history                                       */
/*   Inputs:                                                              */
/*      tabledef    imaging_data table definition                         */
/*   Outputs:                                                             */
/*      cal_rows    cal_table structures initialized for telescopes       */
/*                  1 and 2 with1 as the reference telescope              */
/*      table       table_def for I/O                                     */
/*      ierr        Error code: 0 => ok                                   */
/*   Returns                                                              */
/*      cfitsio file pointer, NULL indicated no file to be written.       */
/*----------------------------------------------------------------------- */
{
  int i;
  fitsfile *out=NULL;

  cal_rows[0] = NULL;
  cal_rows[1] = NULL;
  *table = NULL;

  /* don't write in no output file specified */
  if (outfile==NULL) return out;

  /* Open/create basic FITS output file*/
  fits_create_file(&out, outfile, ierr);
  if (*ierr>0) { /* error check */
    fprintf(stderr,"ERROR %d opening output FITS file %s \n", *ierr, outfile);
    table_print_error_stack();  /* Show any cfitsio error messages */
    return NULL;
  }/* end error test */

  /* create table_def for table */
  *table = make_cal_table_def(out, tabledef->date_obs, 0, ierr);
  if ((*ierr>0) || (*table==NULL)) return NULL;

  fits_write_key(out, TSTRING, "OBJECT", (void*)("MIDI simulation"),
		 "Fringe fitted simulated data", ierr);
  fits_write_key(out, TSTRING, "ORIGIN", (void*)("midi fringe"),
		 "Creator of FITS file", ierr);

  /* copy history */
  copy_history (inputfptr, out, ierr);

  /* add parameters in history */
  history(out, ierr);

  /* make table row objects */
  for (i=0; i<2; i++) {
    cal_rows[i] = make_cal_table_row(*table);
    cal_rows[i]->source = 0;
    cal_rows[i]->tele = i+1;
    cal_rows[i]->ref = 1;
    cal_rows[i]->weight = 10.0;
  }
  return out;
} /* end init_caltable */

void write_caltable(fitsfile *outfptr, cal_table_def* table, 
		    double time, float timei, 
		    cal_table_row *cal_rows[2], int *ierr)
/*----------------------------------------------------------------------- */
/*   Write calibration table entries to open FITS file                    */
/*   Inputs:                                                              */
/*      outfptr     cfitsio file pointer for output FITS file             */
/*      table       table_def for I/O                                     */
/*      time        reference time in days                                */
/*      timei       time span for solution                                */
/*      cal_rows    cal_table structures to be written                    */
/*   Outputs:                                                             */
/*      ierr        Error code: 0 => ok                                   */
/*----------------------------------------------------------------------- */
{
  /* are we writing output? */
  if (outfptr==NULL) return; /* nope */

  /* save time info */
  cal_rows[0]->time = time;
  cal_rows[1]->time = time;
  cal_rows[0]->timei = timei;
  cal_rows[1]->timei = timei;

  /* write 'em */
  write_cal_table(outfptr, table, cal_rows[0], (long)-1, ierr);
  write_cal_table(outfptr, table, cal_rows[1], (long)-1, ierr);
} /* end write_caltable */

void close_caltable(fitsfile *outfptr, cal_table_def* table,
		    cal_table_row *cal_rows[2], int *ierr)
/*----------------------------------------------------------------------- */
/*   Close calibration table in FITS file (closes file)                   */
/*   Also does other cleanup                                              */
/*   Inputs:                                                              */
/*      outfptr     cfitsio file pointer for output FITS file             */
/*      table       table_def for I/O                                     */
/*      cal_rows    cal_table structures to be deleted                    */
/*   Outputs:                                                             */
/*      ierr        Error code: 0 => ok                                   */
/*----------------------------------------------------------------------- */
{
  /* are we writing output? */
  if (outfptr==NULL) return; /* nope */

  /* Close FITS file - this will flush open table buffer */
   close_cal_table(outfptr, table, ierr);

  /* delete structures */
  if (cal_rows!=NULL) {
    kill_cal_table_row (cal_rows[0]);
    kill_cal_table_row (cal_rows[1]);
  }

} /* end close_caltable */

image_info* init_DRpix(int nd, int nr, float del_d, float del_r, 
		       float del_t, char *obs_date, int *ierr)
/*----------------------------------------------------------------------- */
/*   If global variable pixfile is set then it is used as the output FITS */
/*   file name, otherwise no file is written.                             */
/*   This file will contain a FITS image cube of the delay-rate functions */
/*   Copies old, writes new history                                       */
/*   Inputs:                                                              */
/*      nd          number of delay pixels                                */
/*      nr          number of rate pixels                                 */
/*      del_d       delay pixel increment                                 */
/*      del_r       rate pixel increment                                  */
/*      del_t       time pixel increment                                  */
/*      obs_date    observing date                                        */
/*   Outputs:                                                             */
/*      ierr        Error code: 0 => ok                                   */
/*   Returns                                                              */
/*      image_info pointer, NULL indicated no file to be written.         */
/*----------------------------------------------------------------------- */
{
  char *ctype[3] = {"DELAY","RATE","TIME"};
  float crota[3] = {0.0, 0.0, 0.0};
  double crval[3] = {0.0, 0.0, 0.0};
  int  bp, naxis[3];
  float crpix[3], crinc[3], bscale;
  long n;
  image_info *out=NULL;

  if (*ierr!=0) return NULL;     /* existing error conditions? */
  if (pixfile==NULL) return NULL;

  /* set values */
  bp = 8;  /* save as 8 bit integers */
  /* only half the delay plane */
  naxis[0] = nd/2; naxis[1] = nr; naxis[2] = 1; /* axis dimensions */
  crpix[0] = 1.0; /* reference pixels */
  crpix[1] = nr / 2 + 1.0;
  crpix[2] = 1.0;
  crinc[0] = del_d;
  crinc[1] = del_r;
  crinc[2] = del_t * nr / 2.0;

  fprintf (stdout, "Writing delay-rate images cube to %s\n",pixfile);

  /* make work array for image plane */
  n = nd * nr;
  DRplane = (float*)malloc(n*sizeof(float));
  if (DRplane==NULL) return NULL; /* check allocation */

  /* do it */
  out = make_image (pixfile, bp, naxis, ctype, crval, crpix, crinc, crota, 
		     obs_date, ierr);
  if (out==NULL) return out; /* failed */

  /* add amplitude info */
  bscale = 0.25;  /* units to sigma */
  fits_write_key(out->fptr, TFLOAT, "BSCALE", (void*)(&bscale),
		 "Pixel scaling to sigma", ierr);
  fits_write_key(out->fptr, TSTRING, "BUNIT", (void*)("RMS"),
		 "Pixel units are RMS in image", ierr);
  fits_write_key(out->fptr, TSTRING, "OBJECT", (void*)("MIDI simulation"),
		 "Fringe fitted simulated data", ierr);
  fits_write_key(out->fptr, TSTRING, "ORIGIN", (void*)("midi fringe"),
		 "Creator of FITS file", ierr);

  /* copy history */
  copy_history (inputfptr, out->fptr, ierr);

  /* add parameters in history */
  history(out->fptr, ierr);

  if (*ierr>0) { /* error check */
    fprintf(stderr,"ERROR %d creating output delay-rate image \n", *ierr);
    table_print_error_stack();  /* Show any cfitsio error messages */
    return NULL;
  }/* end error test */

  return out; /* done */
} /* end init_DRpix */

void write_DRpix(image_info *image, int nd, int nr, float **array, 
		 float rms, int *ierr)
/*----------------------------------------------------------------------- */
/*   Write delay-rate images to open FITS file                            */
/*   Output images are scaled into units  of the rms                  */
/*   Inputs:                                                              */
/*      image       image_info for I/O                                    */
/*      nd          number of delays                                      */
/*      nr          number of rates                                       */
/*      array       Complex array  as (real, imag)                        */
/*      rms         rms in image                                          */
/*   Outputs:                                                             */
/*      ierr        Error code: 0 => ok                                   */
/*----------------------------------------------------------------------- */
{
  long i, id, ir, iid, n;
  int  jerr;
  float scale, maxval=-1.0;
  
  if (image==NULL) return;  /* anybody home? */
  if (*ierr!=0) return;     /* existing error conditions? */

  i = 0;
  for (ir=0; ir<nr; ir++) {
    for (id=nd/2; id<nd; id++) {
      iid = id*2;
      DRplane[i++] = 
	sqrt (array[ir][iid]*array[ir][iid] + 
	      array[ir][iid+1]*array[ir][iid+1]);
      if (maxval<DRplane[i-1]) maxval = DRplane[i-1]; /* get max */
    }
  }

  /* scale units of rms, correct units for FFT etc. */
  n = (long)(nd/2) * (long)nr;
  /* scale = 255.0 / maxval;*/
  scale = 1.0 / (rms / (8.0 / (float)(n*2)));
  maxval = 255.0 / 4.0;  /* clipping level - related to BSCALE in header*/
  for (i=0; i<n; i++) {
    DRplane[i] *= scale;                       /* scale */
    if (DRplane[i]>maxval) DRplane[i] = maxval;  /* clip */
  }
  
  image_write(image, DRplane, ierr); /* write it out */
  if (*ierr>0) { /* error check */
    fprintf(stderr,"ERROR %d writing output delay-rate image \n", *ierr);
    table_print_error_stack();  /* Show any cfitsio error messages */
    jerr = 0;
    close_DRpix (image, &jerr); /* try to save what's there */
  }/* end error test */
} /* end write_DRpix */

void close_DRpix(image_info *image, int *ierr)
/*----------------------------------------------------------------------- */
/*   Close delay-rate image in FITS file                                  */
/*   Inputs:                                                              */
/*      image       image_info for I/O                                    */
/*   Outputs:                                                             */
/*      ierr        Error code: 0 => ok                                   */
/*----------------------------------------------------------------------- */
{
  if (image==NULL) return; /* anybody home? */
  if (*ierr!=0) return;     /* existing error conditions? */
  image_close(image, ierr);
  if (DRplane!=NULL) free(DRplane);
} /* end close_DRpix */

void history (fitsfile *fptr, int *ierr)
/*----------------------------------------------------------------------- */
/*   Write model parameters in FITS history                               */
/*   Inputs                                                               */
/*   Also writes to stdout                                                */
/*      fptr      cfitsio pointer to open FITS file                       */
/*   Output:                                                              */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*----------------------------------------------------------------------- */
{
  char hiline[FLEN_CARD], ttime[FLEN_CARD];
  sprintf(hiline,"*** VLTI Midi testbed fringe fitter ***");
  fits_write_history (fptr, hiline, ierr);
  fprintf(stdout, "%s in file %s\n", hiline, outfile);
  history_time(ttime); /* get current time/date */
  sprintf(hiline,"fringe run %s", ttime);
  fits_write_history (fptr, hiline, ierr);
  fprintf(stdout, "%s in file %s\n", hiline, outfile);
  sprintf(hiline,"calib input data file = %s",infile);
  sprintf(hiline,"fringe input file = %s",infile);
  fits_write_history (fptr, hiline, ierr);
  fprintf(stdout, "%s\n", hiline);
  sprintf(hiline,"fringe Solution interval = %d",SI);
  fits_write_history (fptr, hiline, ierr);
  fprintf(stdout, "%s\n", hiline);
  sprintf(hiline,"fringe narrow search = %d ",narrow);
  fits_write_history (fptr, hiline, ierr);
  fprintf(stdout, "%s\n", hiline);
} /* end history */
/* */
