/*
 * This file is part of MIA.
 *
 * MIA is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/* $Id: fringeDR.c,v 1.5 2002/06/04 09:29:29 jdejong Exp $ */
#define _POSIX_SOURCE_1
#include "vltPort.h"

/*-----------------------------------------------------------------------*/
/* Reads simulated midi-like data from a FITS binary table               */
/* Test Fringe finder for simulated Midi raw data                        */
/* Currently assumes a positive bias put into OPD delay                  */
/* This version can use a steady variation of the OPD.                   */
/*  Command line parameters:                                             */
/*   -in    Name of FITS file containing the data ,def simul.fit         */
/*   -out   Name of FITS output file with solutions, none=>no output     */
/*   -pix   Name of FITS image cube of delay-rate, none=>DRpix.fit       */
/*   -SI    Solution interval of fit in samples  (def. 10)               */
/*   -narrow use narrow search window(false)                             */
/*-----------------------------------------------------------------------*/
#include <math.h>
#include <malloc.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "imaging_data.h"
#include "cal_table.h"
#include "interpol.h"
#include "history.h"
#include "DRfit.h"

/* global variables, possibly read from command line */
char     *infile="simul.fit"; /* input FITS file name */
char     *pixfile="!DRpix.fit";     /* delay-rate image file */
char     *outfile="!xxx.fit";       /* output FITS file name */
int      SI = 10;             /* Solution interval in samples. */
int      narrow=0;            /* if true use narrow search window */

/* cfitsio pointer for input file */
fitsfile *inputfptr;

#define CLIGHT 2.997924562e8 /* speed of light */
#define TWOPI 6.283185308    /* two pi */
#define FRACT 0.35      /* Fraction of the photons that go into each cross correlation */
#define COHER 0.60      /* Instrumental coherence factor (poln mismatch)*/
#define COUNTS 2.564e4  /* counts per second per mJy */

/* internal prototypes */
void fringein (int argc, char **argv, int *ierr);
void Usage(void);
void read_data_table (fitsfile *fptr, imaging_data_table_def* tabledef, 
		       int *ierr);
fitsfile* init_caltable (imaging_data_table_def* tabledef,
			 cal_table_def **table, 
			 cal_table_row *cal_rows[2], int *ierr);
void write_caltable(fitsfile *outfptr, cal_table_def* table,
		    double time, float timei, 
		    cal_table_row *cal_rows[2], int *ierr);
void close_caltable(fitsfile *outfptr, cal_table_def* table,
		    cal_table_row *cal_rows[2], int *ierr);
void history (fitsfile *fptr, int *ierr);


int main ( int argc, char **argv )
/*----------------------------------------------------------------------- */
/*   Program to fringe fit simulated midi data                            */
/*----------------------------------------------------------------------- */
{
  int  ierr;
  imaging_data_table_def* tabledef;
  fitsfile *infptr=NULL;

  /* Startup - parse command line */
  ierr = 0;
  fringein (argc, argv, &ierr);
  if (ierr!=0) return ierr;

  /* Open FITS input file */
  fits_open_file(&infptr, infile, READONLY, &ierr);
  if (ierr>0) { /* error check */
    fprintf(stderr,"ERROR %d opening input FITS file %s \n", ierr, infile);
    table_print_error_stack();  /* Show any cfitsio error messages */
    return ierr;
  }/* end error test */

  /* Get basic table structures and read IMAGING_DETECTOR table */
  tabledef = open_imaging_data_table(infptr, &ierr);
  if ((tabledef==NULL) || (ierr>0)) { /* error test */
    fprintf (stderr,"error %d after table create\n",ierr);
    table_print_error_stack();  /* Show any cfitsio error messages */
    return ierr;
  } /* end error test */
  
  /* Read table  fitting data*/
  read_data_table(infptr, tabledef, &ierr);
  if (ierr>0) { /* error test */
    fprintf (stderr,"error %d after table read\n",ierr);
    table_print_error_stack();  /* Show any cfitsio error messages */
    return ierr;
  } /* end error test */
  
  /* Close FITS file */
  fits_close_file (infptr, &ierr);

   /* Show any cfitsio error messages */
  table_print_error_stack(); 

  return ierr;
} /* end of main */

void fringein (int argc, char **argv, int *ierr)
/*----------------------------------------------------------------------- */
/*  Parse control info from command line                                  */
/*   Input:                                                               */
/*      argc   Number of arguments from command line                      */
/*      argv   Array of strings from command line                         */
/*   Output:                                                              */
/*      ierr       Error code: 0 => ok                                    */
/*   Global values                                                        */
/*      infile;     input FITS file name                                  */
/*      SI          Solution interval in samples.                         */
/*----------------------------------------------------------------------- */
{
  int ax;
  char *arg;

  /*debug    fprintf(stderr, "number of arguments = %d\n",argc);*/
  if (argc<=1) Usage(); /* must have arguments */
  /* parse command line */
  for (ax=1; ax<argc; ax++) {
    arg = argv[ax];
    /*debug    fprintf(stderr, "arg = %s\n",arg);*/
    if (strcmp(arg, "-in") == 0) {
      infile = argv[++ax];
    } else if (strcmp(arg, "-out") == 0) {
      outfile = argv[++ax];
    } else if (strcmp(arg, "-pix") == 0) {
      pixfile = argv[++ax];
    } else if (strcmp(arg, "-SI") == 0){    /* solution interval */
      sscanf (argv[++ax],"%d",&SI);
    } else if (strcmp(arg, "-narrow") == 0){    /* narrow search */
      narrow = 1;
    } else { /* unknown argument */
      Usage();
    }
  }
  *ierr = 0;
} /* end fringein */

void Usage(void)
/*----------------------------------------------------------------------- */
/*   Tells about usage of program and bails out                           */
/*----------------------------------------------------------------------- */
{
    fprintf(stderr, "Usage: fringe -in file -out file -pix file -SI n -narrow\n");
    fprintf(stderr, "Simulated midi data as VLTI binary tables\n");
    fprintf(stderr, "Arguments:\n");
    fprintf(stderr, "  -in   Name of FITS file to contain the data, def simul.fit\n");
    fprintf(stderr, "  -out  Name of FITS output file with solutions, none=>no output\n");
    fprintf(stderr, "  -pix  Name of FITS image cube of delay-rate, none=>no output\n");
    fprintf(stderr, "  -SI   Solution interval in samples (def. 10)\n");
    fprintf(stderr, "  -narrow use narrow search window(def false)\n");
     /*/exit(1);  bail out */
  }/* end Usage */

void read_data_table (fitsfile *fptr, imaging_data_table_def* tabledef, 
		       int *ierr)
/*----------------------------------------------------------------------- */
/*   Loop reading table, fitting fringes, writing solutions if requested  */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*      tabledef  imaging_data table definition structure                 */
/*   Output:                                                              */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*----------------------------------------------------------------------- */
{
  long i, indx, j, irow, nreg, step, number, nx, integ;
  double  time, taudif;
  double sumtau=0.0, sumd=0.0, sumd2=0.0, cnttau=0.0, cntd2=0.0;
  int DRinit=0;
  imaging_data_row *row; 
  float **spectrum, **big, *OPD_off, *times, timei, fract_corr=-1.0;
  cal_table_def  *caltable=NULL;
  cal_table_row  *cal_rows[2] = {NULL, NULL};
  fitsfile *outfptr = NULL;

  if (*ierr>0) return; /* existing error condition? */
  if (tabledef==NULL) {*ierr=10000; return;} /* check pointer */

  inputfptr = fptr; /* save cfitsio pointer in global variable */

  /* set up output if necessary */
  outfptr = init_caltable(tabledef, &caltable, cal_rows, ierr);
  if (*ierr>0) { /* error test */
    fprintf(stderr,"ERROR %d writing CALIBRATION_TABLE\n", *ierr);
    return;
  }

  nx  = tabledef->detector->naxis[0][0]; /* number of pixels in Freq */
  if (nx<5) { /* error test */
    fprintf(stderr,"ERROR: too few (%ld) frequency channels\n", nx);
    return;
  }

  /* require at least 5 channels */

  /* allocate 2D data arrays */
  spectrum = (float**) malloc(SI*sizeof(float*));
  for (i=0; i<SI; i++)
    spectrum[i] = (float*) malloc(2*nx*sizeof(float));
  /* now one twice as large*/
  big = (float**) malloc(2*SI*sizeof(float*));
  for (i=0; i<2*SI; i++) {
    big[i] = (float*) malloc(4*nx*sizeof(float));
  }

  OPD_off = (float*) malloc(SI*sizeof(float));
  times = (float*) malloc(SI*sizeof(float));

  nreg = tabledef->nregion;        /* number of regions */

  row = make_imaging_data_row (tabledef);

  /* get size of table */
  number = 0;
  fits_get_num_rows (fptr, &number, ierr);

  /* looping over rows*/
  integ = 0; indx = 0; time = 0.0; timei = 0.0;
  for (irow = 1; irow<=number; irow++) {
  
    /* read next row */
     read_imaging_data_table(fptr, tabledef, row, -1, ierr);
     if (*ierr>0) { /* error test */
       fprintf(stderr,"ERROR %d reading IMAGING_DATA_TABLE\n", *ierr);
       return;
     }

     /* initialize fringe fitter if not done before. */
     /* Need to know exposure time for this.         */
     if (!DRinit) {
       setup_DRfringe_fit(pixfile, fptr, nx, SI, row->exp, 
			  tabledef->detector);
       DRinit = 1;
     }

     /* average delay passed by simul as row->opd */
     if (row->opd[0]!=0.0) {
       sumtau += row->opd[0];
       cnttau += 1.0;
     }

     /* correction for how the light is divided amoung spectra, etc */
     /* try to calibrate in mJy */
     if (fract_corr<0.0) 
       fract_corr = ((float)nx / (2.0 * FRACT * COHER)) /
	 (COUNTS * row->exp);
     
     /* copy data to work array */
     OPD_off[integ] = row->local_opd[0]; /* modulation delay */
     times[integ] = row->time;        /* integration time */
     indx = 0;
     for (i=0; i<nx; i++){
       /* real part - difference two spectra */
       spectrum[integ][indx++] = 
	 fract_corr * (row->data[0][i] - row->data[1][i]); 
       spectrum[integ][indx++] = 0.0;   /* imaginary part - not measured */
     }
     integ++;

     /* sum time for average */
     time  += row->time;
     timei += row->exp;

     if (integ>=SI) { /* time for fit? */
       /* debug */
       /*fprintf (stdout, " record %d ",irow);*/
       
       /* fit data */
       /* get times in seconds relative to reference */
       if (integ>0) time = time / integ; /* average time */
       for (i=0; i<integ; i++) times[i] = (times[i] - time) * 86400.0;

       cal_rows[1]->time = time;
       DRfringe_fit_data (spectrum, big, nx, integ, row->exp, OPD_off,
			  times, narrow, tabledef->detector,
			  cal_rows[1]);

       /* write solution if necessary */
       write_caltable(outfptr, caltable, time, timei, cal_rows, ierr);
       if (*ierr>0) { /* error test */
	 fprintf(stderr,"ERROR %d writing CALIBRATION_TABLE\n", *ierr);
	 return;
       }
       time = 0.0; timei = 0.0;
       integ = 0; indx = 0;

       /* Compare model passed by simul with fitted */
       if (cnttau>0.0) {
	 sumtau  = sumtau / cnttau;
	 taudif = sumtau - (cal_rows[1]->delay * 1.0E6 * CLIGHT);
	 /* ignore really bad fits */
	 if (abs(taudif)<15.0) {
	   sumd  += taudif;
	   sumd2 += taudif*taudif;
	   cntd2 += 1.0;
	 }
	 sumtau = 0.0; cnttau = 0.0;
       } /* end compare model */
     } /* end of do fit */
    
  } /* end loop over rows */
  
  /* fit any remaining data  */
  /* need to redo time FT structures for this orphan data */
  /*if (integ>0) 
    fringe_fit_data (spectrum, big, nx, integ, row->exp,  OPD_off,
    times, tabledef->detector);*/

  /* Give RMS delay error */
  if (cntd2>0.0) {
    sumd = sumd / cntd2;
    sumd2  = sqrt (sumd2 / cntd2);
    fprintf (stdout, "RMS delay error = %f microns, avg = %f microns\n",
	     sumd2, sumd);
  } /* end compare model */

  /* shutdown output caltable if necessary */
  close_caltable(outfptr, caltable, cal_rows, ierr);
  if (*ierr>0) { /* error test */
    fprintf(stderr,"ERROR %d closinging CALIBRATION_TABLE\n", *ierr);
    return;
  }

  /* shutdown output delay-rate fitting process */
  shutdown_DRfringe_fit(ierr);
  if (*ierr>0) { /* error test */
    fprintf(stderr,"ERROR %d closing delay-rate process\n", *ierr);
    return;
  }

  /* delete row structure */
  kill_imaging_data_row(row);

  /* delete temporary arrays */
  if (spectrum!=NULL) {
    for (i=0; i<SI; i++)
      if (spectrum[i]!=NULL) free(spectrum[i]);
    free(spectrum);
  }
  if (big!=NULL) {
    for (i=0; i<2*SI; i++)
      if (big[i]!=NULL) free(big[i]);
    free(big);
  }
} /* end read_data_table */

fitsfile* init_caltable (imaging_data_table_def* tabledef,
			 cal_table_def **table, cal_table_row *cal_rows[2], 
			  int *ierr)
/*----------------------------------------------------------------------- */
/*   If global variable outfile is set then it is used as the output FITS */
/*   file name, otherwise no file is written.                             */
/*      If output is desired, the file is created/opened and the          */
/*   cal_table row structures are initialized                             */
/*   Copies old, writes new history                                       */
/*   Inputs:                                                              */
/*      tabledef    imaging_data table definition                         */
/*   Outputs:                                                             */
/*      cal_rows    cal_table structures initialized for telescopes       */
/*                  1 and 2 with1 as the reference telescope              */
/*      table       table_def for I/O                                     */
/*      ierr        Error code: 0 => ok                                   */
/*   Returns                                                              */
/*      cfitsio file pointer, NULL indicated no file to be written.       */
/*----------------------------------------------------------------------- */
{
  int i;
  fitsfile *out=NULL;

  cal_rows[0] = NULL;
  cal_rows[1] = NULL;
  *table = NULL;

  /* don't write in no output file specified */
  if (outfile==NULL) return out;

  /* Open/create basic FITS output file*/
  fits_create_file(&out, outfile, ierr);
  if (*ierr>0) { /* error check */
    fprintf(stderr,"ERROR %d opening output FITS file %s \n", *ierr, outfile);
    table_print_error_stack();  /* Show any cfitsio error messages */
    return NULL;
  }/* end error test */

  /* create table_def for table */
  *table = make_cal_table_def(out, tabledef->date_obs, 0, ierr);
  if ((*ierr>0) || (*table==NULL)) return NULL;

  fits_write_key(out, TSTRING, "OBJECT", (void*)("MIDI simulation"),
		 "Fringe fitted simulated data", ierr);
  fits_write_key(out, TSTRING, "ORIGIN", (void*)("midi fringe"),
		 "Creator of FITS file", ierr);

  /* copy history */
  copy_history (inputfptr, out, ierr);

  /* add parameters in history */
  history(out, ierr);

  /* make table row objects */
  for (i=0; i<2; i++) {
    cal_rows[i] = make_cal_table_row(*table);
    cal_rows[i]->source = 0;
    cal_rows[i]->tele = i+1;
    cal_rows[i]->ref = 1;
    cal_rows[i]->weight = 10.0;
  }
  return out;
} /* end init_caltable */

void write_caltable(fitsfile *outfptr, cal_table_def* table, 
		    double time, float timei, 
		    cal_table_row *cal_rows[2], int *ierr)
/*----------------------------------------------------------------------- */
/*   Write calibration table entries to open FITS file                    */
/*   Inputs:                                                              */
/*      outfptr     cfitsio file pointer for output FITS file             */
/*      table       table_def for I/O                                     */
/*      time        reference time in days                                */
/*      timei       time span for solution                                */
/*      cal_rows    cal_table structures to be written                    */
/*   Outputs:                                                             */
/*      ierr        Error code: 0 => ok                                   */
/*----------------------------------------------------------------------- */
{
  /* are we writing output? */
  if (outfptr==NULL) return; /* nope */

  /* save time info */
  cal_rows[0]->time = time;
  cal_rows[1]->time = time;
  cal_rows[0]->timei = timei;
  cal_rows[1]->timei = timei;

  /* write 'em */
  write_cal_table(outfptr, table, cal_rows[0], (long)-1, ierr);
  write_cal_table(outfptr, table, cal_rows[1], (long)-1, ierr);
} /* end write_caltable */

void close_caltable(fitsfile *outfptr, cal_table_def* table,
		    cal_table_row *cal_rows[2], int *ierr)
/*----------------------------------------------------------------------- */
/*   Close calibration table in FITS file (closes file)                   */
/*   Also does other cleanup                                              */
/*   Inputs:                                                              */
/*      outfptr     cfitsio file pointer for output FITS file             */
/*      table       table_def for I/O                                     */
/*      cal_rows    cal_table structures to be deleted                    */
/*   Outputs:                                                             */
/*      ierr        Error code: 0 => ok                                   */
/*----------------------------------------------------------------------- */
{
  /* are we writing output? */
  if (outfptr==NULL) return; /* nope */

  /* Close FITS file - this will flush open table buffer */
   close_cal_table(outfptr, table, ierr);

  /* delete structures */
  if (cal_rows!=NULL) {
    kill_cal_table_row (cal_rows[0]);
    kill_cal_table_row (cal_rows[1]);
  }

} /* end close_caltable */

void history (fitsfile *fptr, int *ierr)
/*----------------------------------------------------------------------- */
/*   Write model parameters in FITS history                               */
/*   Inputs                                                               */
/*   Also writes to stdout                                                */
/*      fptr      cfitsio pointer to open FITS file                       */
/*   Output:                                                              */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*----------------------------------------------------------------------- */
{
  char hiline[FLEN_CARD], ttime[FLEN_CARD];
  sprintf(hiline,"*** VLTI Midi testbed fringe fitter ***");
  fits_write_history (fptr, hiline, ierr);
  fprintf(stdout, "%s in file %s\n", hiline, outfile);
  history_time(ttime); /* get current time/date */
  sprintf(hiline,"fringeDR run %s", ttime);
  fits_write_history (fptr, hiline, ierr);
  fprintf(stdout, "%s in file %s\n", hiline, outfile);
  sprintf(hiline,"calib input data file = %s",infile);
  sprintf(hiline,"fringeDR input file = %s",infile);
  fits_write_history (fptr, hiline, ierr);
  fprintf(stdout, "%s\n", hiline);
  sprintf(hiline,"fringeDR Solution interval = %d",SI);
  fits_write_history (fptr, hiline, ierr);
  fprintf(stdout, "%s\n", hiline);
  sprintf(hiline,"fringeDR narrow search = %d ",narrow);
  fits_write_history (fptr, hiline, ierr);
  fprintf(stdout, "%s\n", hiline);
} /* end history */
/* */
