/*
 * This file is part of MIA.
 *
 * MIA is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/* $Id: fringeFO.c,v 1.4 2002/06/04 09:29:29 jdejong Exp $ */
#define _POSIX_SOURCE_1
#include "vltPort.h"
/*-----------------------------------------------------------------------*/
/* Reads simulated MIDI data from a FITS binary table (Fourier Mode)     */
/* Test Fringe finder for simulated Midi raw data                        */
/* This version can use a steady variation of the OPD.                   */
/*  Command line parameters:                                             */
/*   -in    Name of FITS file containing the data ,def simul.fit         */
/*   -out   Name of FITS output file with solutions, none=>no output     */
/*   -pix   Name of FITS image cube of delay-rate, none=>DRpix.fit       */
/*   -SI    Solution interval of fit in samples  (def. 10)               */
/*   -narrow use narrow search window(false)                             */
/*-----------------------------------------------------------------------*/
#include <math.h>
#include <malloc.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "imaging_data.h"
#include "cal_table.h"
#include "interpol.h"
#include "history.h"
#include "FOfit.h"

/* global variables, possibly read from command line */
char     *infile="simul.fit";       /* input FITS file name */
char     *pixfile="!FOpix.fit";     /* delay-rate image file */
char     *outfile="!fringeFO.fit";  /* output FITS file name */
int      SI = 2;                    /* Solution interval in samples. */

/* cfitsio pointer for input file */
fitsfile *inputfptr;

#define CLIGHT 2.997924562e8 /* speed of light */
#define TWOPI 6.283185308    /* two pi */
#define FRACT 0.35      /* Fraction of the photons that go into each cross correlation */
#define COHER 0.60      /* Instrumental coherence factor (poln mismatch)*/
#define COUNTS 2.564e4  /* counts per second per mJy */

/* internal prototypes */
void fringein (int argc, char **argv, int *ierr);
void Usage(void);
void read_data_table (fitsfile *fptr, imaging_data_table_def* tabledef, 
		       int *ierr);
fitsfile* init_caltable (imaging_data_table_def* tabledef,
			 cal_table_def **table, 
			 cal_table_row *cal_rows[2], int *ierr);
void write_caltable(fitsfile *outfptr, cal_table_def* table,
		    double time, float timei, 
		    cal_table_row *cal_rows[2], int *ierr);
void close_caltable(fitsfile *outfptr, cal_table_def* table,
		    cal_table_row *cal_rows[2], int *ierr);
void history (fitsfile *fptr, int *ierr);


int main ( int argc, char **argv )
/*----------------------------------------------------------------------- */
/*   Program to fringe fit simulated Fourier mode MIDI data               */
/*----------------------------------------------------------------------- */
{
  int  ierr;
  imaging_data_table_def* tabledef;
  fitsfile *infptr=NULL;

  /* Startup - parse command line */
  ierr = 0;
  fringein (argc, argv, &ierr);
  if (ierr!=0) return ierr;

  /* Open FITS input file */
  fits_open_file(&infptr, infile, READONLY, &ierr);
  if (ierr>0) { /* error check */
    fprintf(stderr,"ERROR %d opening input FITS file %s \n", ierr, infile);
    table_print_error_stack();  /* Show any cfitsio error messages */
    return ierr;
  }/* end error test */

  /* Get basic table structures and read IMAGING_DETECTOR table */
  tabledef = open_imaging_data_table(infptr, &ierr);
  if ((tabledef==NULL) || (ierr>0)) { /* error test */
    fprintf (stderr,"error %d after table create\n",ierr);
    table_print_error_stack();  /* Show any cfitsio error messages */
    return ierr;
  } /* end error test */
  
  /* Read table  fitting data*/
  read_data_table(infptr, tabledef, &ierr);
  if (ierr>0) { /* error test */
    fprintf (stderr,"error %d after table read\n",ierr);
    table_print_error_stack();  /* Show any cfitsio error messages */
    return ierr;
  } /* end error test */
  
  /* Close FITS file */
  fits_close_file (infptr, &ierr);

   /* Show any cfitsio error messages */
  table_print_error_stack(); 

  return ierr;
} /* end of main */

void fringein (int argc, char **argv, int *ierr)
/*----------------------------------------------------------------------- */
/*  Parse control info from command line                                  */
/*   Input:                                                               */
/*      argc   Number of arguments from command line                      */
/*      argv   Array of strings from command line                         */
/*   Output:                                                              */
/*      ierr       Error code: 0 => ok                                    */
/*   Global values                                                        */
/*      infile;     input FITS file name                                  */
/*      SI          Solution interval in scans.                           */
/*----------------------------------------------------------------------- */
{
  int ax;
  char *arg;

  /*debug    fprintf(stderr, "number of arguments = %d\n",argc);*/
  if (argc<=1) Usage(); /* must have arguments */
  /* parse command line */
  for (ax=1; ax<argc; ax++) {
    arg = argv[ax];
    /*debug    fprintf(stderr, "arg = %s\n",arg);*/
    if (strcmp(arg, "-in") == 0) {
      infile = argv[++ax];
    } else if (strcmp(arg, "-out") == 0) {
      outfile = argv[++ax];
    } else if (strcmp(arg, "-pix") == 0) {
      pixfile = argv[++ax];
    } else if (strcmp(arg, "-SI") == 0){    /* solution interval */
      sscanf (argv[++ax],"%d",&SI);
    } else { /* unknown argument */
      Usage();
    }
  }
  *ierr = 0;
} /* end fringein */

void Usage(void)
/*----------------------------------------------------------------------- */
/*   Tells about usage of program and bails out                           */
/*----------------------------------------------------------------------- */
{
    fprintf(stderr, "Usage: fringe -in file -out file -pix file -SI n \n");
    fprintf(stderr, "Fringe fit MIDI Fourier mode data\n");
    fprintf(stderr, "Arguments:\n");
    fprintf(stderr, "  -in   Name of FITS file to contain the data, def simul.fit\n");
    fprintf(stderr, "  -out  Name of FITS output file with solutions, none=>no output\n");
    fprintf(stderr, "  -pix  Name of FITS image cube of delay-rate, none=>no output\n");
    fprintf(stderr, "  -SI   Solution interval in scans (def. 2)\n");
     /*/exit(1);  bail out */
  }/* end Usage */

void read_data_table (fitsfile *fptr, imaging_data_table_def* tabledef, 
		       int *ierr)
/*----------------------------------------------------------------------- */
/*   Loop reading table, fitting fringes, writing solutions if requested  */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*      tabledef  imaging_data table definition structure                 */
/*   Output:                                                              */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*----------------------------------------------------------------------- */
{
  long i, j, irow, nreg, step, number, nx, integ, maxstep, iscan;
  double  time, taudif;
  double sumtau=0.0, sumd=0.0, sumd2=0.0, cnttau=0.0, cntd2=0.0;
  int FOinit=0;
  imaging_data_row *row; 
  float **OPD_off=NULL, **times=NULL, timei, dOPD, fract_corr=-1.0;
  cal_table_def  *caltable=NULL;
  cal_table_row  *cal_rows[2] = {NULL, NULL};
  fitsfile *outfptr = NULL;
  float **IntChan=NULL;
  /* which columns are the interferometric data */
  int i1col=0, i2col=1;

  if (*ierr>0) return; /* existing error condition? */
  if (tabledef==NULL) {*ierr=10000; return;} /* check pointer */

  inputfptr = fptr; /* save cfitsio pointer in global variable */

  /* allocate storage array */
  maxstep = tabledef->maxstep;
  IntChan = (float**)malloc(SI*sizeof(float*));
  OPD_off = (float**) malloc(SI*sizeof(float*));
  times = (float**) malloc(SI*sizeof(float*));
  for (j=0; j<SI; j++) {
    IntChan[j] = (float*)malloc(2*maxstep*sizeof(float));
    for (i=0; i<2*maxstep; i++) IntChan[j][i] = 0.0;
    OPD_off[j] = (float*) malloc(maxstep*sizeof(float));
    times[j] = (float*) malloc(maxstep*sizeof(float));
    for (i=0; i<maxstep; i++) {
      OPD_off[j][i] = 0.0;
      times[j][i] = 0.0;
    }
  }
  /*fprintf (stdout,"maxstep = %ld \n",maxstep);*/

  /* set up output if necessary */
  outfptr = init_caltable(tabledef, &caltable, cal_rows, ierr);
  if (*ierr>0) { /* error test */
    fprintf(stderr,"ERROR %d writing CALIBRATION_TABLE\n", *ierr);
    return;
  }

  nx  = tabledef->detector->naxis[0][0]; /* number of pixels in Freq */
  if (nx>1) { /* error test */
    fprintf(stderr,"ERROR: too many (%ld) frequency channels\n", nx);
    return;
  }

  nreg = tabledef->nregion;        /* number of regions */

  row = make_imaging_data_row (tabledef);

  /* get size of table */
  number = 0;
  fits_get_num_rows (fptr, &number, ierr);

  /* looping over rows*/
  integ = 0; time = 0.0; timei = 0.0;
  iscan = 0;
  for (irow = 1; irow<=number; irow++) {
  
    /* read next row */
     read_imaging_data_table(fptr, tabledef, row, -1, ierr);
     if (*ierr>0) { /* error test */
       fprintf(stderr,"ERROR %d reading IMAGING_DATA_TABLE\n", *ierr);
       return;
     }

     if (integ>row->step) { /* time for next scan? */
       /* debug */
       /*fprintf (stdout, " record %d ",irow);*/
       iscan++;

       /* time for solution? */
       if (iscan>=SI) { 
       
	 /* fit data */
	 /* get times in seconds relative to reference */
	 if (integ>0) time = time / integ; /* average time */
	 for (j=0; j<SI; j++) {
	   for (i=0; i<integ; i++) times[j][i] = (times[j][i] - time) * 86400.0;
	 }
	 
	 /* initialize fringe fitter if not done before. */
	 /* Need to know exposure time, etc for this.         */
	 if (!FOinit) {
	   dOPD = (OPD_off[0][integ-1] - OPD_off[0][0]);
	   setup_FOfringe_fit(pixfile, fptr, maxstep, SI, row->exp, dOPD, 
			      tabledef->detector);
	   FOinit = 1;
	 }

	 cal_rows[1]->time = time;
	 FOfringe_fit_data (IntChan, integ, iscan, row->exp, OPD_off,
			    times, tabledef->detector, cal_rows[1]);
	 
	 /* write solution if necessary */
	 write_caltable(outfptr, caltable, time, timei, cal_rows, ierr);
	 if (*ierr>0) { /* error test */
	   fprintf(stderr,"ERROR %d writing CALIBRATION_TABLE\n", *ierr);
	   return;
	 }

	 iscan = 0;

	 /* Compare model passed by simul with fitted */
	 if (cnttau>0.0) {
	   sumtau  = sumtau / cnttau;
	   fprintf(stdout,"Average model OPD error was %f\n", sumtau);
	   taudif = sumtau - (cal_rows[1]->delay * 1.0E6 * CLIGHT);
	   /* ignore really bad fits */
	   if (abs(taudif)<15.0) {
	     sumd  += taudif;
	     sumd2 += taudif*taudif;
	     cntd2 += 1.0;
	   }
	   sumtau = 0.0; cnttau = 0.0;
	 } /* end compare model */
       } /* end of process data */
       time = 0.0; timei = 0.0;
       integ = 0; 
     } /* end of next scan */
    
     /* average delay passed by simul as row->opd */
     if (row->opd[0]!=0.0) {
       sumtau += row->opd[0];
       cnttau += 1.0;
     }

     /* correction for how the light is divided amoung spectra, etc */
     /* try to calibrate in mJy */
     if (fract_corr<0.0) 
       fract_corr = (1.0 / (2.0 * FRACT * COHER)) /
	 (COUNTS * row->exp);
     
     /* copy data to work array */
     OPD_off[iscan][integ] = row->local_opd[0]; /* modulation delay */
     times[iscan][integ] = row->time;        /* integration time */
     IntChan[iscan][2*integ]   = fract_corr * 
       (row->data[i1col][0] - row->data[i2col][0]);
     IntChan[iscan][2*integ+1] = 0.0;        /* zero imaginary part */
     integ++;

     /* sum time for average */
     time  += row->time;
     timei += row->exp;

  } /* end loop over rows */
  
  /* Give RMS delay error */
  if (cntd2>0.0) {
    sumd = sumd / cntd2;
    sumd2  = sqrt (sumd2 / cntd2);
    fprintf (stdout, "RMS delay error = %f microns, avg = %f microns\n",
	     sumd2, sumd);
  } /* end compare model */

  /* shutdown output caltable if necessary */
  close_caltable(outfptr, caltable, cal_rows, ierr);
  if (*ierr>0) { /* error test */
    fprintf(stderr,"ERROR %d closinging CALIBRATION_TABLE\n", *ierr);
    return;
  }

  /* shutdown output delay-rate fitting process */
  shutdown_FOfringe_fit(ierr);
  if (*ierr>0) { /* error test */
    fprintf(stderr,"ERROR %d closing delay-rate process\n", *ierr);
    return;
  }

  /* delete row structure */
  kill_imaging_data_row(row);

  /* deallocate work arrays */
  if (IntChan) {
    for (j=0; j<SI; j++) {
      if (IntChan[j]) free (IntChan[j]);
    }
    free (IntChan);
  }

  if (OPD_off) {
    for (j=0; j<SI; j++) {
      if (OPD_off[j]) free (OPD_off[j]);
    }
    free (OPD_off);
  }

  if (times) {
    for (j=0; j<SI; j++) {
      if (times[j]) free (times[j]);
    }
    free (times);
  }

} /* end read_data_table */

fitsfile* init_caltable (imaging_data_table_def* tabledef,
			 cal_table_def **table, cal_table_row *cal_rows[2], 
			  int *ierr)
/*----------------------------------------------------------------------- */
/*   If global variable outfile is set then it is used as the output FITS */
/*   file name, otherwise no file is written.                             */
/*      If output is desired, the file is created/opened and the          */
/*   cal_table row structures are initialized                             */
/*   Copies old, writes new history                                       */
/*   Inputs:                                                              */
/*      tabledef    imaging_data table definition                         */
/*   Outputs:                                                             */
/*      cal_rows    cal_table structures initialized for telescopes       */
/*                  1 and 2 with1 as the reference telescope              */
/*      table       table_def for I/O                                     */
/*      ierr        Error code: 0 => ok                                   */
/*   Returns                                                              */
/*      cfitsio file pointer, NULL indicated no file to be written.       */
/*----------------------------------------------------------------------- */
{
  int i;
  fitsfile *out=NULL;

  cal_rows[0] = NULL;
  cal_rows[1] = NULL;
  *table = NULL;

  /* don't write in no output file specified */
  if (outfile==NULL) return out;

  /* Open/create basic FITS output file*/
  fits_create_file(&out, outfile, ierr);
  if (*ierr>0) { /* error check */
    fprintf(stderr,"ERROR %d opening output FITS file %s \n", *ierr, outfile);
    table_print_error_stack();  /* Show any cfitsio error messages */
    return NULL;
  }/* end error test */

  /* create table_def for table */
  *table = make_cal_table_def(out, tabledef->date_obs, 0, ierr);
  if ((*ierr>0) || (*table==NULL)) return NULL;

  fits_write_key(out, TSTRING, "OBJECT", (void*)("MIDI simulation"),
		 "Fringe fitted simulated data", ierr);
  fits_write_key(out, TSTRING, "ORIGIN", (void*)("midi fringe"),
		 "Creator of FITS file", ierr);

  /* copy history */
  copy_history (inputfptr, out, ierr);

  /* add parameters in history */
  history(out, ierr);

  /* make table row objects */
  for (i=0; i<2; i++) {
    cal_rows[i] = make_cal_table_row(*table);
    cal_rows[i]->source = 0;
    cal_rows[i]->tele = i+1;
    cal_rows[i]->ref = 1;
    cal_rows[i]->weight = 10.0;
  }
  return out;
} /* end init_caltable */

void write_caltable(fitsfile *outfptr, cal_table_def* table, 
		    double time, float timei, 
		    cal_table_row *cal_rows[2], int *ierr)
/*----------------------------------------------------------------------- */
/*   Write calibration table entries to open FITS file                    */
/*   Inputs:                                                              */
/*      outfptr     cfitsio file pointer for output FITS file             */
/*      table       table_def for I/O                                     */
/*      time        reference time in days                                */
/*      timei       time span for solution                                */
/*      cal_rows    cal_table structures to be written                    */
/*   Outputs:                                                             */
/*      ierr        Error code: 0 => ok                                   */
/*----------------------------------------------------------------------- */
{
  /* are we writing output? */
  if (outfptr==NULL) return; /* nope */

  /* save time info */
  cal_rows[0]->time = time;
  cal_rows[1]->time = time;
  cal_rows[0]->timei = timei;
  cal_rows[1]->timei = timei;

  /* write 'em */
  write_cal_table(outfptr, table, cal_rows[0], (long)-1, ierr);
  write_cal_table(outfptr, table, cal_rows[1], (long)-1, ierr);
} /* end write_caltable */

void close_caltable(fitsfile *outfptr, cal_table_def* table,
		    cal_table_row *cal_rows[2], int *ierr)
/*----------------------------------------------------------------------- */
/*   Close calibration table in FITS file (closes file)                   */
/*   Also does other cleanup                                              */
/*   Inputs:                                                              */
/*      outfptr     cfitsio file pointer for output FITS file             */
/*      table       table_def for I/O                                     */
/*      cal_rows    cal_table structures to be deleted                    */
/*   Outputs:                                                             */
/*      ierr        Error code: 0 => ok                                   */
/*----------------------------------------------------------------------- */
{
  /* are we writing output? */
  if (outfptr==NULL) return; /* nope */

  /* Close FITS file - this will flush open table buffer */
   close_cal_table(outfptr, table, ierr);

  /* delete structures */
  if (cal_rows!=NULL) {
    kill_cal_table_row (cal_rows[0]);
    kill_cal_table_row (cal_rows[1]);
  }

} /* end close_caltable */

void history (fitsfile *fptr, int *ierr)
/*----------------------------------------------------------------------- */
/*   Write model parameters in FITS history                               */
/*   Inputs                                                               */
/*   Also writes to stdout                                                */
/*      fptr      cfitsio pointer to open FITS file                       */
/*   Output:                                                              */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*----------------------------------------------------------------------- */
{
  char hiline[FLEN_CARD], ttime[FLEN_CARD];
  sprintf(hiline,"*** VLTI Midi testbed fringe fitter ***");
  fits_write_history (fptr, hiline, ierr);
  fprintf(stdout, "%s in file %s\n", hiline, outfile);
  history_time(ttime); /* get current time/date */
  sprintf(hiline,"fringeFO run %s", ttime);
  fits_write_history (fptr, hiline, ierr);
  fprintf(stdout, "%s in file %s\n", hiline, outfile);
  sprintf(hiline,"calib input data file = %s",infile);
  sprintf(hiline,"fringeFO input file = %s",infile);
  fits_write_history (fptr, hiline, ierr);
  fprintf(stdout, "%s\n", hiline);
} /* end history */
/* */
