/*
 * This file is part of MIA.
 *
 * MIA is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/* $Id: select.c,v 1.4 2002/05/02 15:20:09 jdejong Exp $ */
#define _POSIX_SOURCE_1
#include "vltPort.h"

/*-----------------------------------------------------------------------*/
/*  Program to select data for a given source from a binary table        */
/*  ARRAY_GEOMETRY, SOURCE, and IMAGING_DETECTOR  tables are copied      */
/*  and then entries for the selected source in the IMAGING_DATA table   */
/*  are copied.                                                          */
/*-----------------------------------------------------------------------*/
#include <math.h>
#include <malloc.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "imaging_data.h"
#include "VLTIfileUtil.h"
#include "history.h"

/* global variables , possibly read from command line */
char  *infile="selectin.fit";   /* input FITS file name */
char  *outfile="!selectout.fit"; /* output FITS file name */
int   sid=1;                    /* source ID number */

/* internal prototypes */
void selectin (int argc, char **argv, int *ierr);
void Usage(void);
void copy_data_table (fitsfile *infptr, fitsfile *outfptr, int *ierr);

int main (int argc, char **argv)
/*----------------------------------------------------------------------- */
/*  Program to select data for a given source from a binary table         */
/*----------------------------------------------------------------------- */
{
  int  ierr, OK;
  fitsfile *infptr=NULL;
  fitsfile *outfptr=NULL;
  
  /*  Startup - parse command line arguments */
  ierr = 0;
  selectin (argc, argv, &ierr);
  if (ierr!=0) return ierr;
  
  /* Open input file*/
  fits_open_file(&infptr, infile, READONLY, &ierr);
  if (ierr>0) { /* error check */
    fprintf(stderr,"ERROR %d opening input FITS file %s \n", 
	    ierr, infile);
  }
  OK = (ierr==0);
  
  /* Open/create basic FITS output file*/
  fits_create_file(&outfptr, outfile, &ierr);
  if (ierr>0) { /* error check */
    fprintf(stderr,"ERROR %d opening output FITS file %s \n", 
	    ierr, outfile);
  }/* end error test */
  OK = (ierr==0);

  /* copy array geometry table */
  copy_table ("ARRAY_GEOMETRY", infptr, outfptr, &ierr);
  if (ierr>0) { /* error check */
    fprintf(stderr,"ERROR %d copying ARRAY_GEOMETRY table\n", ierr);
    if (OK) ierr = 0; /* ignore table copy problem */
  }/* end error test */

  /* copy optical_train table */
  copy_table ("OPTICAL_TRAIN", infptr, outfptr, &ierr);
  if (ierr>0) { /* error check */
    fprintf(stderr,"ERROR %d copying OPTICAL_TRAIN table\n", ierr);
    if (OK) ierr = 0; /* ignore table copy problem */
  }/* end error test */

  /* copy source table */
  copy_table ("SOURCE", infptr, outfptr, &ierr);
  if (ierr>0) { /* error check */
    fprintf(stderr,"ERROR %d copying SOURCE table\n", ierr);
    if (OK) ierr = 0; /* ignore table copy problem */
  } /* end error test */
  if (OK) ierr = 0; /* ignore table copy problem */
  
  /* copy data */
  copy_data_table (infptr, outfptr, &ierr);
  if (ierr>0) { /* error check */
    fprintf(stderr,"ERROR %d copying selected data\n", ierr);
  } /* end error test */

  /* Close FITS files */
  fits_close_file (outfptr, &ierr);
  fits_close_file (infptr, &ierr);

  /* Show any cfitsio error messages */
  table_print_error_stack(); 
    
  return ierr;
} /* end of main */

void selectin (int argc, char **argv, int *ierr)
/*----------------------------------------------------------------------- */
/*  Parse control info from command line */
/*   Input: */
/*      argc   Number of arguments from command line */
/*      argv   Array of strings from command line */
/*   Output: */
/*      ierr       Error code: 0 => ok */
/*----------------------------------------------------------------------- */
{
  int ax;
  char *arg;

  if (argc<=1) Usage(); /* must have arguments */
/* parse command line */
  for (ax=1; ax<argc; ax++) {
    arg = argv[ax];
    if (strcmp(arg, "-in") == 0) {
      infile = argv[++ax];
    } else if (strcmp(arg, "-out") == 0) {
      outfile = argv[++ax];
    } else if (strcmp(arg, "-sid") == 0){ 
      sscanf (argv[++ax],"%d",&sid);
    } else { /* unknown argument */
      Usage();
    }
  }

  /* must specify files */
  /*  if (!infile) Usage();*/
  if (!infile) infile = "simul.fit";
  *ierr = 0;
} /* end selectin */

void Usage(void)
/*----------------------------------------------------------------------- */
/*   Tells about usage of program and bails out */
/*----------------------------------------------------------------------- */
{
    fprintf(stderr, "Usage: select  -in file -out file -sid  sid \n");
    fprintf(stderr, "Copy selected data from a FITS binary table\n");
    fprintf(stderr, "Arguments:\n");
    fprintf(stderr, "  -in   Name of input FITS file, def selectin.fit\n");
    fprintf(stderr, "  -out  Name of output FITS output file, def selectout.fit\n");
    fprintf(stderr, "  -sid  Source id number (def. 1)\n");
    /* exit(1);  bail out */
  }/* end Usage */
void copy_data_table (fitsfile *infptr, fitsfile *outfptr, int *ierr)
/*----------------------------------------------------------------------- */
/*   Read and print input IMAGING_DATA table                              */
/*   Selects source using global sid value                                */
/*   Inputs:                                                              */
/*      infptr    cfitsio pointer to open FITS input file                 */
/*      outfptr   cfitsio pointer to open FITS output file                */
/*   Output:                                                              */
/*      ierr      Error code: 0 => ok                                     */
/*----------------------------------------------------------------------- */
{
  imaging_data_table_def *intabledef, *outtabledef;
  imaging_data_row *row; 
  long irow, orow, nrow;

  if (*ierr>0) return;        /* existing error condition? */
  if (infptr==NULL) return;   /* sanity check */
  if (outfptr==NULL) return;  /* sanity check */

  /* print some info */
  fprintf (stdout, "Copying IMAGING_DATA for source id = %d\n",sid);

  /* Now open input Data table - this reads the IMAGING_DETECTOR table 
     and opens the IMAGING_DATA table */
  intabledef = open_imaging_data_table(infptr, ierr);
  if ((!intabledef) || (*ierr>0)) { /* error check */
    fprintf (stderr,"error %d opening input data table\n", *ierr);
    return;
  } /* end of error check */

  /* Now create output Data table - this writes the IMAGING_DETECTOR table 
     and opens the IMAGING_DATA table */
  outtabledef = 
    make_imaging_data_table_def(outfptr, intabledef->detector, 
			    intabledef->maxstep, ierr);
  if ((!outtabledef) || (*ierr>0)) { /* error check */
    fprintf (stderr,"error %d opening output data table\n", *ierr);
    return;
  } /* end of error check */

  /* create row structure */
  row =  make_imaging_data_row(intabledef);
  if (!row) { /* allocation error? */
    fprintf (stderr, "copy_data: ERROR creating row structure.\n");
    *ierr = 10000;
    kill_imaging_data_table_def(intabledef,1);  /* cleanup */
    outtabledef->detector = NULL; /* same structure */
    kill_imaging_data_table_def(outtabledef,1); 
    return;
  }

  /* copy history */
  copy_history (infptr, outfptr, ierr);

  /* get size of table */
  fits_get_num_rows (infptr, &nrow, ierr);

  /* looping over rows copying */
  orow = 0;
  for (irow = 1; irow<=nrow; irow++) {
    read_imaging_data_table(infptr, intabledef, row, irow, ierr); /* read */
    if (*ierr!=0) { /* error? */
      fprintf (stderr, "copy_table: ERROR %d reading input data row %ld\n",
	       *ierr, irow);
      kill_imaging_data_table_def(intabledef,1);  /* cleanup */
      outtabledef->detector = NULL; /* same structure */
      kill_imaging_data_table_def(outtabledef,1); 
      kill_imaging_data_row(row); 
      return;
    }
    /* write if this one is wanted (source for first region) */
    if (row->source[0]==sid) {
      orow++; /* output row number */
      write_imaging_data_table (outfptr, outtabledef, row, orow, ierr); /* write */
      if (*ierr!=0) { /* error? */
	fprintf (stderr, "copy_table: ERROR %d writing output data row %ld\n",
		 *ierr, orow);
	kill_imaging_data_table_def(intabledef,1);  /* cleanup */
	outtabledef->detector = NULL; /* same structure */
	kill_imaging_data_table_def(outtabledef,1); 
	kill_imaging_data_row(row); 
	return;
      }
    } /* end of write output */
  } /* end loop copying rows */
  
  /* print out much copied */
  fprintf (stdout, "Copied = %ld/%ld rows\n",orow,nrow);
  
  /* cleanup */
  kill_imaging_data_table_def(intabledef,1);
  outtabledef->detector = NULL; /* same structure */
  kill_imaging_data_table_def(outtabledef,1);
  kill_imaging_data_row(row); 
  
} /* end copy_data_table */
/* */
