/*
 * This file is part of MIA.
 *
 * MIA is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/* $Id: simul.c,v 1.4 2002/06/04 09:29:29 jdejong Exp $ */
#define _POSIX_SOURCE_1
#include "vltPort.h"

/*-----------------------------------------------------------------------*/
/* Writes simulated midi-like data to a FITS binary table                */
/* Simulates two spectra (two regions) with phase different by 180 deg.  */ 
/*  Simulator for Midi raw data                                          */
/* Command line arguments:                                               */
/* -input  the name of the input parameter file, def. demo/simul.in      */
/*                                                                       */
/* N=5 is 390 mJy                                                        */
/*  30% of photons go to photometric channel and Fract% to each spectrum */
/*  Input_file parameters:                                               */
/*   out   Name of FITS file to contain the data ,def simul.fit          */
/*   amp   Amplitude of correlated signal in mJy  def 390                */
/*   rms   RMS OPD noise in microns, def 0                               */
/*   delay Delay to add (microns)  def 200                               */
/*   rate  Delay rate to add (microns/sec)  def. 0                       */
/*   number Number of frames, def 100                                    */
/*   integ Integration time in msec, def 25                              */
/*   nstep number of OPD delay steps, def 40                             */
/*   step  OPD delay step in microns, def 2.0                            */
/*   OPDfile  file containing OPD offsets                                */
/*   tau0  Desired atmospheric coherence time (sec) def no change        */
/*Processing parameters are written in HISTORY records in the FITS header*/
/*-----------------------------------------------------------------------*/
#define SIMULATEMAIN /* this contains the main for a data simulator */
#include "simulate.h"
#include "parser.h"

/* global variables, possibly read from command line */
char     *input_file="demo/simul.in";  /* input parameter file name */
/* global variables possibly read from the input_file */
char     *outfile="!simul.fit";     /* output FITS file name */
long     number=100;        /* number of samples */
int      integ=25;          /* integration time in msec */
long     amplitude = 390;   /* Amplitude of correlated signal in mJy */
float    delay = 200.0;     /* Delay to add (microns) */
float    rate = 0.0;        /* Delay rate to add (microns/sec) */
float    rmsOPD = 0.0;      /* RMS OPD noise */
int      nstep = 40;        /* number of OPD delay steps */
float    step = 2.0;        /* OPD delay step in microns*/
char     *OPDfile=NULL;     /* OPD error file */
double   tau0 = -1.0;       /* Desired atm. coherence time (sec) */
double   detgain=0.01;      /* detector gain */

/* internal prototypes */
void simulin (int argc, char **argv, int *ierr);
void Usage(void);
void write_data_table (fitsfile *fptr, imaging_data_table_def* tabledef, 
		       int *ierr);
void history (fitsfile *fptr, int *ierr);
/* create parser list */
parse_list* make_parse_list();
/* copy parse_list to global variables */
void parse_list_to_global (parse_list* list);

int main ( int argc, char **argv )
/*----------------------------------------------------------------------- */
/*   Program to write a FITS binary table                                 */
/*----------------------------------------------------------------------- */
{
  int  ierr, nx, ny, nreg;
  imaging_data_table_def* tabledef;
  fitsfile *outfptr=NULL;

  /* Startup - parse command line */
  ierr = 0;
  simulin (argc, argv, &ierr);
  if (ierr!=0) return ierr;

  /* Open/create basic FITS output file*/
  fits_create_file(&outfptr, outfile, &ierr);
  if (ierr>0) { /* error check */
    fprintf(stderr,"ERROR %d opening output FITS file %s \n", ierr, outfile);
    table_print_error_stack();  /* Show any cfitsio error messages */
    return ierr;
  }/* end error test */

  /* Create basic table structures and write IMAGING_DETECTOR table */
  /* set basic paramaters */
  nx = Nchan;
  ny = 1;
  nreg = 2;
  tabledef = make_data_table(outfptr, nx, ny, nreg, nstep, &ierr);
  if ((tabledef==NULL) || (ierr>0)) { /* error test */
    fprintf (stderr,"error %d after table create\n",ierr);
    table_print_error_stack();  /* Show any cfitsio error messages */
    return ierr;
  } /* end error test */
  
  /* Write table */
  write_data_table(outfptr, tabledef, &ierr);
  if (ierr>0) { /* error test */
    fprintf (stderr,"error %d after table write\n",ierr);
    table_print_error_stack();  /* Show any cfitsio error messages */
    return ierr;
  } /* end error test */
  
  /* Close FITS file - this will flush open table buffer */
  fits_close_file (outfptr, &ierr);

   /* Show any cfitsio error messages */
  table_print_error_stack(); 

  return ierr;
} /* end of main */

void simulin (int argc, char **argv, int *ierr)
/*----------------------------------------------------------------------- */
/*  Parse control info from command line                                  */
/*   Input:                                                               */
/*      argc   Number of arguments from command line                      */
/*      argv   Array of strings from command line                         */
/*   Output:                                                              */
/*      ierr       Error code: 0 => ok                                    */
/*   Global values                                                        */
/*      outfile;    output FITS file name                                 */
/*      number      number of samples                                     */
/*      integ       integration time in msec                              */
/*      amplitude   Amplitude of correlated signal in counts              */
/*      delay       Delay to add (microns)                                */
/*      rate        Delay rate to add (microns/sec)                       */
/*      nstep       Number of OPD modulation steps                        */
/*      step        Size in microns of OPD modulation steps               */
/*      OPDfile     OPD error file                                        */
/*      tau0        Desired atm. coherence time (sec)                     */
/*----------------------------------------------------------------------- */
{
  int ax;
  char *arg;
  parse_list* list;

  /* command line arguments */
  /*debug    fprintf(stderr, "number of arguments = %d\n",argc);*/
  if (argc<=1) Usage(); /* must have arguments */
  /* parse command line */
  for (ax=1; ax<argc; ax++) {
    arg = argv[ax];
    /*debug    fprintf(stderr, "arg = %s\n",arg);*/
    if (strcmp(arg, "-input") == 0){ /* input parameters */
      input_file = argv[++ax];
    } else { /* unknown argument */
      Usage();
    }
  }
  
  /* must specify file */
  if (!input_file) {Usage(); exit(1);}

  /* parse input file */
  /* create/init parse_list */
  list = make_parse_list();

  /* parse input file, bail out on error */
  if (parse_input_file(input_file, list)) exit(1);

  /* copy to input_info */
  parse_list_to_global (list);

  /* debug 
     return;*/

  /* delete parser list */
  kill_parse_list(list);

  *ierr = 0;
} /* end simulin */

void Usage(void)
/*----------------------------------------------------------------------- */
/*   Tells about usage of program and bails out                           */
/*----------------------------------------------------------------------- */
{
    fprintf(stderr, "Usage: simul -input file\n");
    fprintf(stderr, "Simulated midi data as VLTI binary tables\n");
    fprintf(stderr, "Arguments:\n");
    fprintf(stderr, "  -input input parameter file, def demo/simul.in\n");
    
    /*/exit(1);  bail out */
  }/* end Usage */

void write_data_table (fitsfile *fptr, imaging_data_table_def* tabledef, 
		       int *ierr)
/*----------------------------------------------------------------------- */
/*   Loop writing table                                                   */
/*   Inputs:                                                            : */
/*      fptr      cfitsio pointer to open FITS file                       */
/*      tabledef  imaging_data table definition structure                 */
/*   Output:                                                              */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*----------------------------------------------------------------------- */
{
  long i, j, irow, ireg, nreg, istep;
  imaging_data_row *row; 
  float step_sec, *bg, *gain, amp, OPD_off = 0.0;
  FILE *hFile = NULL;
  int nf;

  /* if using OPD_model */
  if (OPDfile!=NULL) hFile = OPDfile_open(OPDfile, tau0);

  if (*ierr>0) return; /* existing error condition? */
  if (tabledef==NULL) {*ierr=10000; return;} /* check pointer */

  /* initialize randum number generator */
  amp = GaussNoise (-1.0);

  /* frequency dependent gain factors 1% fluctuations */
  nf = tabledef->detector->naxis[0][0]; /* number of pixels in Freq */
  gain = (float*)malloc(nf*sizeof(float));
  for (i=0; i<nf; i++) gain[i] = 1.0 + GaussNoise(0.01);

  /* 30% frequency dependent background */
  bg =  (float*)malloc(nf*sizeof(float));
  for (i=0; i<nf; i++) 
    bg[i] = gain[i] * (1.0 + ((0.3 * (i-nf/2)) / (float)nf));;

  /* number of regions */
  nreg = tabledef->nregion;       

  row = make_imaging_data_row (tabledef);
  row->frame = 0;

  /* set correlations types */
  for (i=0; i<nreg; i++)
    row->targtype[i] = 2; /* data type */

  /* modulation step size in seconds */
   step_sec = 1.0e-6 * step / CLIGHT;

  /* set constant values */
  row->exp = 0.001 * integ; /* exposure in seconds */
  row->offset[0] = 0.0;
  row->offset[1] = 0.0;
  row->rotation = 0.0;
  row->reference = 0;
  row->opd = 0.0;
  for (ireg=0; ireg<row->nregion; ireg++) {
    row->targtype[ireg] = 1;
    row->source[ireg] = 1;
  }

  /* looping over rows*/
  istep = 0; /* OPD modulation step */
  for (irow = 1; irow<=number; irow++) {
    /* fill in data record */
    row->time = irow * (row->exp / 86400.0); /* time in days */
    row->frame++;
    row->step = istep + 1; 
    row->local_opd = (istep-(nstep/2)) * step_sec; /* add modulation delay */
    istep++;
    if (istep>=nstep) istep = 0; /* cycle done? */

    /* next OPD model if necessary */
    if (OPDfile!=NULL) {
      OPD_off = OPDfile_read (hFile, 86400.0*row->time);
      row->opd = OPD_off+delay; /* save model value in record */
    }  else { /* only delay/rate added */
      row->opd = delay+rate*86400.0*row->time; /* save model value in record */
    }

    /* rest of model */
    simulate_data (row, OPD_off, amplitude, integ, delay, 
		   rate, rmsOPD, gain, bg, tabledef->detector);
    
    /* write next row */
     write_imaging_data_table(fptr, tabledef, row, -1, ierr);
     if (*ierr>0) { /* error test */
       fprintf(stderr,"ERROR %d writing IMAGING_DATA_TABLE\n", *ierr);
       return;
     }

  } /* end loop over rows */
  
  /* delete row structure */
  kill_imaging_data_row(row);

  /* close OPD model file */
  if (OPDfile!=NULL) OPDfile_close(hFile);

} /* end write_data_table */

void history (fitsfile *fptr, int *ierr)
/*----------------------------------------------------------------------- */
/*   Write model parameters in FITS history                               */
/*   Inputs                                                               */
/*   Also writes to stdout                                                */
/*      fptr      cfitsio pointer to open FITS file                       */
/*   Output:                                                              */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*----------------------------------------------------------------------- */
{
  char hiline[FLEN_CARD], ttime[FLEN_CARD];

  sprintf(hiline,"*** VLTI Midi data simulator ***");
  fits_write_history (fptr, hiline, ierr);
  fprintf(stdout, "%s in file %s\n", hiline,outfile);
  history_time(ttime); /* get current time/date */
  sprintf(hiline,"simul run %s", ttime);
  fits_write_history (fptr, hiline, ierr);
  fprintf(stdout, "%s in file %s\n", hiline, outfile);
  sprintf(hiline,"simul Amplitude = %ld",amplitude);
  fits_write_history (fptr, hiline, ierr);
  fprintf(stdout, "%s\n", hiline);
  sprintf(hiline,"simul Integration = %d msec",integ);
  fits_write_history (fptr, hiline, ierr);
  fprintf(stdout, "%s\n", hiline);
  sprintf(hiline,"simul Delay offset = %f micron",delay);
  fits_write_history (fptr, hiline, ierr);
  fprintf(stdout, "%s\n", hiline);
  sprintf(hiline,"simul Delay rate = %f micron/s",rate);
  fits_write_history (fptr, hiline, ierr);
  fprintf(stdout, "%s\n", hiline);
  sprintf(hiline,"simul RMS OPD noise = %f micron",rmsOPD);
  fits_write_history (fptr, hiline, ierr);
  fprintf(stdout, "%s\n", hiline);
  sprintf(hiline,"simul Readout noise = %ld",(long)RON);
  fits_write_history (fptr, hiline, ierr);
  fprintf(stdout, "%s\n", hiline);
  sprintf(hiline,"simul Sky brightness per ms = %ld",(long)Sky);
  fits_write_history (fptr, hiline, ierr);
  fprintf(stdout, "%s\n", hiline);
  fprintf(stdout, "simul Number of samples is %ld\n",number);
  sprintf(hiline,"simul number of OPD steps = %d",nstep);
  fits_write_history (fptr, hiline, ierr);
  fprintf(stdout, "%s\n", hiline);
  sprintf(hiline,"simul OPD step size (microns) = %f",step);
  fits_write_history (fptr, hiline, ierr);
  fprintf(stdout, "%s\n", hiline);
  if (OPDfile) {
    sprintf(hiline,"simul OPD model file = %s",OPDfile);
    fits_write_history (fptr, hiline, ierr);
    fprintf(stdout, "%s\n", hiline);
     sprintf(hiline,"simul Tau_0 = %lf",tau0);
    fits_write_history (fptr, hiline, ierr);
    fprintf(stdout, "%s\n", hiline);
  }
} /* end history */

/* create parser list for simul */
  parse_list* make_parse_list() {
  parse_list *out = create_parse_list();

  /* add parser items */
  /* output FITS file name */
  out->list[(out->NumItem)++] = 
    make_parse_item ("outfile", PSTRING, 0.0, 0, "simul.fit");

  /* detector electron to ADU factor */
  out->list[(out->NumItem)++] = 
    make_parse_item ("gain", PFLOAT, 0.01, 0, NULL);

  /* number of samples */
  out->list[(out->NumItem)++] = 
    make_parse_item ("number", PINTEGER, 0.0, 100, NULL);

  /* integration time in msec */
  out->list[(out->NumItem)++] = 
    make_parse_item ("integ", PINTEGER, 0.0, 25, NULL);

  /* Amplitude of correlated signal in mJy */
  out->list[(out->NumItem)++] = 
    make_parse_item ("amplitude", PINTEGER, 0.0, 390, NULL);

  /* Delay to add (microns) */
  out->list[(out->NumItem)++] = 
    make_parse_item ("delay", PFLOAT, 200.0, 0, NULL);

  /* Delay rate to add (microns/sec) */
  out->list[(out->NumItem)++] = 
    make_parse_item ("rate", PFLOAT, 0.0, 0, NULL);

  /* RMS OPD noise */
  out->list[(out->NumItem)++] = 
    make_parse_item ("rmsOPD", PFLOAT, 0.0, 0, NULL);

  /* number of OPD delay steps */
  out->list[(out->NumItem)++] = 
    make_parse_item ("nstep", PINTEGER, 0.0, 40, NULL);

  /* OPD delay step in microns */
  out->list[(out->NumItem)++] = 
    make_parse_item ("step", PFLOAT, 2.0, 0, NULL);

  /* OPD error file */
  out->list[(out->NumItem)++] = 
    make_parse_item ("OPDfile", PSTRING, 0.0, 0, NULL);

  /* Desired atm. coherence time (sec) */
  out->list[(out->NumItem)++] = 
    make_parse_item ("tau0", PFLOAT, -1.0, 0, NULL);

  /* Simulation control parameters */
  /* RON:   readout noise of detector */
  out->list[(out->NumItem)++] = 
    make_parse_item ("RON", PINTEGER, 0.0, 2000, NULL);

  /* Sky:   sky (+telescope) brightness per msec */
  out->list[(out->NumItem)++] = 
    make_parse_item ("Sky", PFLOAT, 6.0E7, 0, NULL);

  /* Coher:  Instrumental coherence factor (poln mismatch) */
  out->list[(out->NumItem)++] = 
    make_parse_item ("Coher", PFLOAT, 0.60, 0, NULL);

  /* Wavelength:  center wavelength in microns */
  out->list[(out->NumItem)++] = 
    make_parse_item ("Wavelength", PFLOAT, 10.0, 0, NULL);

  /* DW: total range of wavelength in microns */
  out->list[(out->NumItem)++] = 
    make_parse_item ("DW", PFLOAT, 7.0, 0, NULL);

  /* Nchan: Number of frequency channels, assumed independent */
  out->list[(out->NumItem)++] = 
    make_parse_item ("Nchan", PINTEGER, 0.0, 320, NULL);

  /* Fract:  Fraction of the photons that go into each 
     cross correlation */
  out->list[(out->NumItem)++] = 
    make_parse_item ("Fract", PFLOAT, 0.35, 0, NULL);

  /* Counts:  counts per second per mJy */
  out->list[(out->NumItem)++] = 
    make_parse_item ("Counts", PFLOAT, 2.564e4, 0, NULL);

  return out;
} /* end make_parse_list */

/* copy parse_list to global variables */
void parse_list_to_global (parse_list* list) {
  long indx;

  /* output FITS file name */
  indx = find_parse_entry("outfile", list);
  if (indx>=0) {
    if (list->list[indx]->cvalue) { /* valid string? */
      outfile = malloc (strlen(list->list[indx]->cvalue)+1);
      strcpy (outfile,list->list[indx]->cvalue);
    }
  }

  /* detector electron to ADU factor */
  indx = find_parse_entry("gain", list);
  if (indx>=0) DetGain = list->list[indx]->fvalue;

  /* number of samples */
  indx = find_parse_entry("number", list);
  if (indx>=0) number = list->list[indx]->ivalue;

  /* integration time in msec */
  indx = find_parse_entry("integ", list);
  if (indx>=0) integ = list->list[indx]->ivalue;

  /* Amplitude of correlated signal in mJy */
  indx = find_parse_entry("amplitude", list);
  if (indx>=0) amplitude = list->list[indx]->ivalue;

  /* Delay to add (microns) */
  indx = find_parse_entry("delay", list);
  if (indx>=0) delay = list->list[indx]->fvalue;

  /* Delay rate to add (microns/sec) */
  indx = find_parse_entry("rate", list);
  if (indx>=0) rate = list->list[indx]->fvalue;

  /* RMS OPD noise */
  indx = find_parse_entry("rmsOPD", list);
  if (indx>=0) rmsOPD = list->list[indx]->fvalue;

  /* number of OPD delay steps */
  indx = find_parse_entry("nstep", list);
  if (indx>=0) nstep = list->list[indx]->ivalue;

  /* OPD delay step in microns */
  indx = find_parse_entry("step", list);
  if (indx>=0) step = list->list[indx]->fvalue;

  /* OPD error file */
  indx = find_parse_entry("OPDfile", list);
  if (indx>=0) {
    if (list->list[indx]->cvalue) { /* valid string? */
      OPDfile = malloc (strlen(list->list[indx]->cvalue)+1);
      strcpy (OPDfile,list->list[indx]->cvalue);
    }
  }

  /* Desired atm. coherence time (sec) */
  indx = find_parse_entry("tau0", list);
  if (indx>=0) tau0 = list->list[indx]->fvalue;

  /* Simulation control parameters */
  /* RON:   readout noise of detector */
  indx = find_parse_entry("RON", list);
  if (indx>=0) RON = list->list[indx]->ivalue;

  /* Sky:   sky (+telescope) brightness per msec */
  indx = find_parse_entry("Sky", list);
  if (indx>=0) Sky = list->list[indx]->fvalue;

  /* Coher:  Instrumental coherence factor (poln mismatch) */
  indx = find_parse_entry("Coher", list);
  if (indx>=0) Coher = list->list[indx]->fvalue;

  /* Wavelength:  center wavelength in microns */
  indx = find_parse_entry("Wavelength", list);
  if (indx>=0) Wavelength = list->list[indx]->fvalue;

  /* DW: total range of wavelength in microns */
  indx = find_parse_entry("DW", list);
  if (indx>=0) DW = list->list[indx]->fvalue;

  /* Nchan: Number of frequency channels, assumed independent */
  indx = find_parse_entry("Nchan", list);
  if (indx>=0) Nchan  = list->list[indx]->ivalue;

  /* Fract:  Fraction of the photons that go into each 
     cross correlation */
  indx = find_parse_entry("Fract", list);
  if (indx>=0) Fract = list->list[indx]->fvalue;

  /* Counts:  counts per second per mJy */
  indx = find_parse_entry("Counts", list);
  if (indx>=0) Counts = list->list[indx]->fvalue;
  
} /* end  parse_list_to_global */
/* */
/* */
