/*
 * This file is part of MIA.
 *
 * MIA is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/* $Id$
 * $Log$
 *
 * midiFTscans.c
 *
 * Created:     Sun Aug  3 13:18:57 2003 by Koehler@rommie
 * Last change: Tue Aug 19 14:35:51 2003
 *
 * Based on oirCompressData.c by Walter Jaffe
 *
 * reads a MIDI-fringe-search or -track data-set (with PRISM),
 * fourier-transforms each scan pixel-by-pixel,
 * picks frequency where fringes should be in this pixel,
 * and writes resulting frames to another MIDI-datafile
 * (one frame per scan in input file)
 *
 * USAGE: midiFTscans "midi-files" outfile
 */

#define DEBUG 1

#define MAX_REGIONS 4	/** max. number of data-regions we can handle **/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <float.h>
#include <math.h>
#include "midiMultifile.h"
#ifdef DOUBLE_FFTW
# include "drfftw.h"
#else
# include "rfftw.h"
#endif



/********************************* MAIN *********************************/

int main( int argc, char **argv)
{
   struct multifile *MFile;
   float *buffy[MAX_REGIONS];

   int nRegion, iRegion;
   int maxstep;
   int i,ierr;
   long iScan, iRow, oRow;
   int nx, ny;
   long npix, ipix;
   int iKey;

   fitsfile *inFile;
   fitsfile *outFile = NULL;
   imaging_data_table_def *idata;
   imaging_fdata_table_def *odata;
   imaging_detector_table_def *outDetectorDef;
   imaging_data_row *inData;
   imaging_fdata_row *outData;

	rfftw_plan rfft_plan;
	fftw_real *ExpBuf= NULL;	/** this shouldn't be double, but that wouldn't hurt too much **/
	fftw_real *FFTBuf= NULL;
	long detwin1strx;

   if( argc<3 )
   {  printf("USAGE: %s inputfile outputfile\n",argv[0]);  return 10; }

   ierr = 0;
   if ((MFile = multifile_open(argv[1], &ierr)) == NULL)
   {	fits_report_error(stdout,ierr);  return 11;  }

  /*
   *    open first file and get detector info; then close file
   */
  fits_open_file (&inFile, MFile->filelist[0], READONLY, &ierr);
  if (ierr != 0)
  {  fits_report_error(stdout,ierr);  return 12;  }

  fits_read_key(inFile,TLONG,"HIERARCH ESO DET WIN1 STRX", &detwin1strx,NULL,&ierr);
#ifdef DEBUG
  fprintf(stderr,"detwin1strx is %ld\n",detwin1strx);
#endif

  outDetectorDef = read_imaging_detector_table(inFile, &ierr);
  idata          = open_imaging_data_table(inFile, &ierr);
  inData         = make_imaging_data_row(idata);
  nRegion        = idata->nregion;
  iKey = table_find_key(idata->def, "MAXSTEP");
  if (iKey>0)
	table_get_keyword(idata->def, iKey, (void*)&maxstep, &ierr);

  fits_close_file(inFile, &ierr);
  kill_imaging_data_table_def(idata,1);

#ifdef DEBUG
   fprintf(stderr,"got detector info from 1.file\n");
   printf("scanlength is %d\n",maxstep);
#endif
  /*
   *    create output data table
   */
   fits_create_file (&outFile, argv[2], &ierr);
   if (ierr !=0)
   {  fits_report_error(stderr, ierr);  return 13; }

   odata= make_imaging_fdata_table_def(outFile, outDetectorDef, maxstep, &ierr);
   if (ierr !=0)
   {  fits_report_error(stderr, ierr);  return 14; }

   outData = make_imaging_fdata_row(odata);
   oRow = 1;

#ifdef DEBUG
   fprintf(stderr,"output data table created.\n");
#endif
  /*
   *  set data and nonData info to some defined state
   */
   outData->time = -1.;
   outData->opd[0] = -1.;
   outData->opd[1] = -1.;
   outData->local_opd[1] = -1.;
   outData->local_opd[0] = -1.;
   outData->exp = -1.;
   outData->offset[0] = -1.;
   outData->offset[1] = -1.;
   outData->rotation  = -1.;
   outData->frame     = 0;
   outData->reference = 0;
   for (iRegion=0; iRegion<nRegion; iRegion++)
   {
     npix = (long) outDetectorDef->naxis[iRegion][0]
	  * (long) outDetectorDef->naxis[iRegion][1]
	  * maxstep;
     //fprintf(stderr,"init region %d - %ld pixels\n",iRegion,npix);

     if ((buffy[iRegion]= malloc(sizeof(float) * npix)) == NULL)
     {
	multifile_close(MFile,&ierr);
	fprintf(stderr,"Out of memory chips\n");
	return 15;
     }
     outData->targtype[iRegion] = 'U';
   }
#ifdef DEBUG
   fprintf(stderr,"output data table initialized\n");
#endif
   /***** prepare FFT **********************/

  if( !(rfft_plan= rfftw_create_plan(maxstep,
			FFTW_REAL_TO_COMPLEX, FFTW_ESTIMATE|FFTW_OUT_OF_PLACE)))
  {
	multifile_close(MFile,&ierr);
	fprintf(stderr,"Strange: can't create fftw-plan!\a\n");
	return 16;
  }
  /*** NOTE: we should try FFTW_MEASURE some time... ****/

  if( !(ExpBuf= malloc(maxstep*sizeof(fftw_real)))
   || !(FFTBuf= malloc(maxstep*sizeof(fftw_real))))
  {
    multifile_close(MFile,&ierr);
    fprintf(stderr,"Out of memory for FFT buffers!\a\n");
    return 17;
  }
 /***********************************************************************
  *     loop over all scans
  ***********************************************************************/
  iRow = 1;	/** rows in fbts are 1-based! **/

  for( iScan=0;  iScan < MFile->totframes/maxstep;  iScan++)
  {
    double minopd= 1e99, maxopd=-1e99, opd;

#ifdef DEBUG
    //if( (iRow&63) == 0)
    fprintf(stderr,"Scan%4ld, first row %ld\r",iScan,iRow);
#endif
    for( i=0;  i<maxstep;  ++i)
    {
	if (multifile_rd_img_data_tbl( MFile, inData, iRow, &ierr) != 0)
	{	fits_report_error(stderr,ierr);
		multifile_close(MFile, &ierr);
		return 20;
	}
	opd = inData->local_opd[1]-inData->local_opd[0] + inData->opd[1]-inData->opd[0];
	if( minopd > opd)  minopd= opd;
	if( maxopd < opd)  maxopd= opd;

	for (iRegion=0; iRegion<nRegion; iRegion++)
	{
	  short *inPix;
	  float *outPix;

	  nx = MFile->inDetectorDef->naxis[iRegion][0];
          ny = MFile->inDetectorDef->naxis[iRegion][1];
          npix = nx * (long)ny;
          inPix = (short *)inData->data[iRegion];
	  outPix= buffy[iRegion] + i*npix;

          for (ipix=0; ipix<npix; ipix++)
		*(outPix++) = 32768. + (float) *(inPix++);
	}
	++iRow;
    } /** end of loop over rows in scan **/

    opd = (maxopd - minopd) * 1e6;	/** total opd in this scan **/

    for (iRegion=0; iRegion<nRegion; iRegion++)
    {
	float *inPix;
	double p, freq, wfreql, wfreqh;
	fftw_real P1,P2;
	int freqi;

	nx = MFile->inDetectorDef->naxis[iRegion][0];
        ny = MFile->inDetectorDef->naxis[iRegion][1];
        npix = nx * (long)ny;

        for( ipix=0; ipix<npix;  ++ipix)  /** foreach pix in region **/
	{
	  inPix = buffy[iRegion] + ipix;
	  for( i=0;  i<maxstep;  ++i)	  /** foreach frame in scan **/
		ExpBuf[i] = (fftw_real) inPix[i*npix];

	  rfftw_one(rfft_plan, ExpBuf, FFTBuf);

	  /* old: freq  = opd / (-0.0519 * ((ipix%nx)+detwin1strx) + 21.1844); */
	  p= (double)((ipix%nx) + detwin1strx);
	  freq  = opd / (-0.0001539 * p*p + 0.009491 * p + 15.451905);
	  wfreql= ceil(freq) - freq;
	  wfreqh= freq - floor(freq);
	  freqi = (int)floor(freq);

	  P1 = sqrt(FFTBuf[freqi] * FFTBuf[freqi]
		    + FFTBuf[maxstep-freqi] *  FFTBuf[maxstep-freqi]);
	  ++freqi;
	  P2 = sqrt(FFTBuf[freqi] * FFTBuf[freqi]
		    + FFTBuf[maxstep-freqi] *  FFTBuf[maxstep-freqi]);

	  outData->data[iRegion][ipix] = P1 * wfreql + P2 * wfreqh;
	}
    }
    opd = (maxopd + minopd) / 2.;	/** average total opd **/
    outData->opd[0] = opd;
    outData->opd[1] = opd;
    outData->local_opd[0] = minopd-opd;	/** diff to min opd (<0) **/
    outData->local_opd[1] = maxopd-opd;	/** diff to max opd (>0) **/

    /** write row with FT(scan)-data **/
    write_imaging_fdata_table(outFile, odata, outData, oRow++, &ierr);
  }
  if( close_imaging_fdata_table(outFile, odata, &ierr))
	fits_report_error(stderr,ierr);

  for (iRegion=0; iRegion<nRegion; iRegion++)  free(buffy[iRegion]);
  free(ExpBuf);
  free(FFTBuf);
  rfftw_destroy_plan(rfft_plan);
  multifile_close(MFile,&ierr);
  kill_imaging_fdata_row(outData);
  kill_imaging_data_row(inData);

#if 0
   for (iRegion=0; iRegion<nRegion; iRegion++)
   {
     nx = MFile->inDetectorDef->naxis[iRegion][0];
     ny = MFile->inDetectorDef->naxis[iRegion][1];
     npix = nx * (long)ny;

     for( ipix=0;  ipix<npix;  ipix++)
	outData->data[iRegion][ipix] /= totalRows;
   }
#endif

  return 0;
}
