/*
 * This file is part of MIA.
 *
 * MIA is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/* $Id: oirMeanRMS.c,v 1.3 2003/12/01 17:13:55 koehler Exp $
 * $Log: oirMeanRMS.c,v $
 * Revision 1.3  2003/12/01 17:13:55  koehler
 * small bugfix
 *
 * Revision 1.2  2003/06/17 07:12:34  koehler
 * many small improvements
 *
 *
 * oirmeanRMS.c
 *
 * Created:     Thu Apr  3 10:53:59 2003 by Koehler@sun47
 * Last change: Sun Aug  3 14:29:18 2003
 *
 * Based on oirCompressData.c by Walter Jaffe
 *
 * reads one or more midi-datafiles, computes mean & rms of the data
 * regions, and writes another midi-datafile
 * (row 0 = mean, row 1 = rms)
 *
 * USAGE: oirMeanRMS "midi-files" outfile
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <float.h>
#include <math.h>
#include "fitsio.h"
#include "imaging_data.h"
#include "imaging_fdata.h"
#include "minrtsFrameData.h"

char** strSplit (char *string, int *nString) {
   char *blank = " ";
   char *oldString;
   int   nS, iS ;
   char *tok;
   char *stringCopy;
   char **output;

   *nString = 0;
   stringCopy = strdup(string);
   oldString = stringCopy;
   tok = strtok(oldString, blank);
   if (tok == NULL) return NULL;
   oldString = NULL;
   for (nS = 0; tok != NULL; nS++) {
      tok = strtok(oldString, blank);
   }
   *nString = nS;
   output = (char **)malloc(nS * sizeof(char *));
   strcpy (stringCopy, string);
   oldString = stringCopy;
   blank = " ";
   for (iS=0; iS<nS; iS++) {
      tok = strtok(oldString, blank);
      oldString = NULL;
      output[iS]= strdup(tok);
   }
   free(oldString);
   return output;
}

void killFileList(char **fileList, int nFiles) {
   int iFile;
   for (iFile=0; iFile<nFiles;iFile++) {free(fileList[iFile]);}
   free(fileList);
   return;
} /* end of killFileList */

/********************************* MAIN *********************************/

int main( int argc, char **argv)
{
   char **fileList;
   int nFiles, iFile;
   int nRegion, iRegion;
   int maxstep;
   int i,ierr;
   long nRows, iRow, oRow;
   float totalRows;	/** we count in float, because
			    we want to divide by it **/
   int nx, ny;
   long npix, ipix;
   short *inPix;
   int iKey;

   fitsfile *inFile;
   fitsfile *outFile = NULL;
   imaging_data_table_def *idata;
   imaging_fdata_table_def *odata;
   imaging_detector_table_def *inDetectorDef;
   imaging_detector_table_def *outDetectorDef;
   imaging_data_row *inData;
   imaging_fdata_row *outData;

   double **sum2Buf;


   if( argc<3 ) {
     printf("USAGE: %s inputfile outputfile\n",argv[0]);
     return 10;
   }
   fileList = strSplit(argv[1], &nFiles);

   for (i=0; i<nFiles; i++) {printf("%s \n", fileList[i]);}
   ierr = 0;

  /*
   *    open first file and get detector info; then close file
   */
   fits_open_file (&inFile, fileList[0], READONLY, &ierr);
   if (ierr != 0) {
      fits_report_error(stdout,ierr);
      return 11;
   }
   outDetectorDef = read_imaging_detector_table(inFile, &ierr);
   idata          = open_imaging_data_table(inFile, &ierr);
   inData         = make_imaging_data_row(idata);
   nRegion        = idata->nregion;
   iKey = table_find_key(idata->def, "MAXSTEP");
   if (iKey>0)
	table_get_keyword (idata->def, iKey, (void*)&maxstep, &ierr);
   i = fits_close_file(inFile, &ierr);

   //fprintf(stderr,"got detector info from 1.file\n");

  /*
   *	allocate buffer for sum^2 of regions
   */
   if(	!(sum2Buf = malloc( nRegion * sizeof(double *))))
   {
     printf("Error - out of memory!\a\n");
     return 10;
   }
   for( iRegion=0;  iRegion<nRegion;  ++iRegion)
   {
     npix = (long) outDetectorDef->naxis[iRegion][0]
	  * (long) outDetectorDef->naxis[iRegion][1];

     if( !(sum2Buf[iRegion] = malloc( npix * sizeof(double))) )
     {
       printf("Error - not enough memory for region buffers!\a\n");
       return 15;
     }
   }
   //fprintf(stderr,"alloced sum2 buffers\n");

  /*
   *    create output data table
   */
   fits_create_file (&outFile, argv[2], &ierr);
   if (ierr !=0) {
      fits_report_error(stderr, ierr);
      return 1;
      }
   odata = make_imaging_fdata_table_def (outFile,
      outDetectorDef, maxstep, &ierr);
   if (ierr !=0) {
      fits_report_error(stderr, ierr);
      return 1;
      }
   outData = make_imaging_fdata_row(odata);
   oRow = 1;

   //fprintf(stderr,"output data table created.\n");

  /*
   *  set data and nonData info to some defined state
   */
   outData->time = -1.;
   outData->opd[0] = -1.;
   outData->opd[1] = -1.;
   outData->local_opd[1] = -1.;
   outData->local_opd[0] = -1.;
   outData->exp = -1.;
   outData->offset[0] = -1.;
   outData->offset[1] = -1.;
   outData->rotation  = -1.;
   outData->frame     = 0;
   outData->reference = 0;
   for (iRegion=0; iRegion<nRegion; iRegion++)
   {
     npix = (long) outDetectorDef->naxis[iRegion][0]
	  * (long) outDetectorDef->naxis[iRegion][1];
     //fprintf(stderr,"init region %d - %ld pixels\n",iRegion,npix);

     outData->targtype[iRegion] = 'U';
     for( ipix=0;  ipix<npix;  ++ipix)
     {
	outData->data[iRegion][ipix]= 0.0;
	sum2Buf[iRegion][ipix]= 0.0;
     }
   }
   //fprintf(stderr,"output data table initialized\n");
  /*
   *  loop over input files to compute mean
   */
   printf("===== computing mean\n");
   totalRows = 0;

   for (iFile=0; iFile<nFiles; iFile++) {
      fits_open_file (&inFile, fileList[iFile], READONLY, &ierr);
      if (ierr != 0) {
         fits_report_error(stderr,ierr);
         return 1;
      }
      inDetectorDef = read_imaging_detector_table(inFile, &ierr);
      idata = open_imaging_data_table(inFile, &ierr);
      i = fits_get_num_rows (inFile,  &nRows, &ierr);
      printf("reading %s\n%ld rows\n",fileList[iFile],nRows);

   /* suppress autoscaling to avoid overflow */
      for (i=0; i<idata->nregion; i++)
         fits_set_tscale(inFile, 1+idata->data_col[i], 1.0, 0.0, &ierr);
   /*
    *     loop over input rows
    */
      for (iRow=1; iRow<=nRows; iRow++) {
         i = read_imaging_data_table(inFile, idata, inData, iRow, &ierr);
	 if( (iRow&63) == 0)  fprintf(stderr," %ld \r",iRow);

         for (iRegion=0; iRegion<nRegion; iRegion++)  {
            nx = inDetectorDef->naxis[iRegion][0];
            ny = inDetectorDef->naxis[iRegion][1];
            npix = nx * (long)ny;
            inPix = (short *)inData->data[iRegion];
            for (ipix=0; ipix<npix; ipix++)
	    {
		float pixval = 32768. + (float) *(inPix++);
		outData->data[iRegion][ipix] += pixval;
	    }
	 }
	 totalRows += 1;
      }
      i = fits_close_file(inFile, &ierr);
   }
   for (iRegion=0; iRegion<nRegion; iRegion++)
   {
     nx = inDetectorDef->naxis[iRegion][0];
     ny = inDetectorDef->naxis[iRegion][1];
     npix = nx * (long)ny;

     for( ipix=0;  ipix<npix;  ipix++)
	outData->data[iRegion][ipix] /= totalRows;
   }
  /*
   *  write row with mean data
   */
   i = write_imaging_fdata_table(outFile, odata, outData, oRow++, &ierr);

  /*
   *  loop over input files to compute rms
   */
   printf("===== computing rms\n");
   totalRows = 0;

   for (iFile=0; iFile<nFiles; iFile++) {
      fits_open_file (&inFile, fileList[iFile], READONLY, &ierr);
      if (ierr != 0) {
         fits_report_error(stderr,ierr);
         return 1;
      }
      inDetectorDef = read_imaging_detector_table(inFile, &ierr);
      idata = open_imaging_data_table(inFile, &ierr);
      i = fits_get_num_rows (inFile,  &nRows, &ierr);
      printf("reading %s\n%ld rows\n",fileList[iFile],nRows);

   /* suppress autoscaling to avoid overflow */
      for (i=0; i<idata->nregion; i++)
         fits_set_tscale(inFile, 1+idata->data_col[i], 1.0, 0.0, &ierr);
   /*
    *     loop over input rows
    */
      for (iRow=1; iRow<=nRows; iRow++) {
         i = read_imaging_data_table(inFile, idata, inData, iRow, &ierr);
	 if( (iRow&63) == 0)  fprintf(stderr," %ld \r",iRow);

         for (iRegion=0; iRegion<nRegion; iRegion++)  {
            nx = inDetectorDef->naxis[iRegion][0];
            ny = inDetectorDef->naxis[iRegion][1];
            npix = nx * (long)ny;
            inPix = (short *)inData->data[iRegion];
            for (ipix=0; ipix<npix; ipix++)
	    {
		float pixval= (32768. + *(inPix++)) - outData->data[iRegion][ipix];
		sum2Buf[iRegion][ipix] += (double)(pixval * pixval);
	    }
	 }
	 totalRows += 1;
      }
      i = fits_close_file(inFile, &ierr);
   }
   for (iRegion=0; iRegion<nRegion; iRegion++)
   {
     nx = inDetectorDef->naxis[iRegion][0];
     ny = inDetectorDef->naxis[iRegion][1];
     npix = nx * (long)ny;

     for( ipix=0;  ipix<npix;  ipix++)
       outData->data[iRegion][ipix]= sqrt(sum2Buf[iRegion][ipix] / totalRows);
   }
  /*
   *  write row with rms data
   */
   i = write_imaging_fdata_table(outFile, odata, outData, oRow++, &ierr);

   i = close_imaging_fdata_table(outFile, odata, &ierr);

  kill_imaging_fdata_row(outData);
  kill_imaging_data_row(inData);
  kill_imaging_data_table_def (idata,1);
  return 0;
}
